using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections;
using System.Threading;
using System.Runtime.InteropServices;
using ClubDev.Utilities;

namespace ClubDev.CaptionButton
{
    // Main class that one would create in their application
    public partial class CaptionButtonManager : Form
    {

        // Class which stores an application's information, and all of the icons that
        // are associated with that application
        private class IconGroup : IComparable
        {
            public ArrayList icons;                 // icons that appear on the title bar
            public int appHandle;                   // window handle
            public string collabHost;               // the current hostname of the machine
            public string collabName;               // the current username of the user
            public Win32Api.Rect appDimensions;     // application window dimensions
            public bool hide_always;

            // Constructor
            //
            // Parameters:
            //      - int hwnd : window handle for an application
            //
            // NOTE:  Currently only one icon is added
            public IconGroup(int hwnd, string collabHost, string collabName)
            {
                appHandle = hwnd;
                this.collabHost = collabHost;
                this.collabName = collabName;
                this.hide_always = false;

                icons = new ArrayList();
            
                IntPtr appWnd = new IntPtr(hwnd);
                appDimensions = new Win32Api.Rect();
                if (!Win32Api.GetWindowRect(appWnd, ref appDimensions))
                {
                    throw new Exception("Error getting window's rectangle, window does not exist");
                    
                }

                icons.Add(new CaptionButton(hwnd, new Point(appDimensions.right - 170, appDimensions.top), this.collabHost, this.collabName));
                
            }


            // Call before removing an IconGroup.  This function cleans up after all of the icons
            // in this group
            public void destroy()
            {
                foreach (CaptionButton icon in icons)
                {
                    icon.Hide();
                    icon.Close();
                    icon.Dispose();
                }
            }

            // This function is called to update the position of the icons in the group so that they
            // are located at the correct place on the title bar
            public void updateIconPosition()
            {
                int delta_x, delta_y;

                Win32Api.Rect newdims = new Win32Api.Rect();
                IntPtr appWnd = new IntPtr(appHandle);
                if(!Win32Api.GetWindowRect(appWnd, ref newdims))
                    return;

                delta_x = newdims.right - appDimensions.right;
                delta_y = newdims.top - appDimensions.top;


                // Only move the icons if we need to
                if (delta_x != 0 || delta_y != 0)
                {
                    foreach (CaptionButton ic in icons)
                    {
                        ic.Location = new Point(newdims.right - 170, newdims.top);
                    }
                    
                }
                appDimensions = newdims;
            }


            // Function that checks if icons need to be hidden because they are covered
            // by another window.
            //
            // Parameters:
            //
            // place_in_z order = This icon groups ordering relative to the other apps.
            //                      0 = on top of the pile
            // icongroups = Group of IconGroups that this IconGroup is checked against for
            //                  visibility
            public void updateIconVisibility(int place_in_z_order, ArrayList icongroups)
            {
                if (hide_always)
                    return;

                foreach (CaptionButton ic in icons)
                {
                    ic.hidden = false;
                }

                //walk up the pile, check if we are covered
                for (int i = place_in_z_order -1; i > -1; i--)
                {
                    foreach (CaptionButton ic in icons)
                    {
                 
                        if (((IconGroup)icongroups[i]).appDimensions.left > ic.Location.X + 25)
                        {
                            continue;
                        }

                        if (((IconGroup)icongroups[i]).appDimensions.bottom < ic.Location.Y + 2)
                        {
                            continue;
                        }

                        if (((IconGroup)icongroups[i]).appDimensions.right < ic.Location.X + 8)
                        {
                            continue;
                        }

                        if (((IconGroup)icongroups[i]).appDimensions.top > ic.Location.Y + 25)
                        {
                            continue;
                        }

                        ic.hidden = true;
                        
                    }
                }

                                
                   
                // If an icon should be hidden, hide it.  Otherwise, show the icon.
                foreach (CaptionButton ic in icons)
                {
                    try
                    {
                        if (ic.hidden && place_in_z_order != 0)
                        {
                            ic.Hide();
                        }
                        else
                        {
                            ic.Show();
                        }
                    }
                    catch { }
                }
                 
            }


            // Function that allows for comparing an IconGroup to other objects.
            // The compare is based upon the application window handle
            public int CompareTo(object obj)
            {
                if (obj is IconGroup)
                {
                    IconGroup ig = (IconGroup) obj;
                    return appHandle.CompareTo(ig.appHandle);
                }
                if (obj is int)
                {
                    int hwnd = (int)obj;
                    return appHandle.CompareTo(hwnd);
                }
                throw new ArgumentException("object is not an IconGroup or an int");
            }
        }


        // TitleIconsManager members

        private ArrayList icongroups;       // All of the IconGroups we are managing
        private System.Windows.Forms.Timer timer;
        private String[] image_locations; //locations of images to use for icons
        private MouseEventHandler[] on_mouse_up_handlers; //handlers for on mouse up

        private string collabHost;
        private string collabName;

        public CaptionButtonManager(string collabHost, string collabName)
        {
            this.collabHost = collabHost;
            this.collabName = collabName;

            icongroups = new ArrayList();
            InitializeComponent();
            
            this.timer = new System.Windows.Forms.Timer();
            this.timer.Interval = 250;
            this.timer.Tick += new EventHandler(timer_Tick);
            this.timer.Start();
        }




        // NOT FINISHED !!! DO NOT CALL!!!
        /*
        public CaptionButtonManager(String[] image_locations, MouseEventHandler[] on_mouse_up_handlers)
        {
            icongroups = new ArrayList();
            this.image_locations = image_locations;
            this.on_mouse_up_handlers = on_mouse_up_handlers;

            InitializeComponent();
            this.timer = new System.Windows.Forms.Timer();
            this.timer.Interval = 10;
            this.timer.Tick += new EventHandler(timer_Tick);
            this.timer.Start();

        }
        */

        // Called ever timer tick
        void timer_Tick(object sender, EventArgs e)
        {
            run();
        }

        

        // Worker function ... called once per timer tick
        public void run()
        {
            
               
            // Get a list of all active parent windows
            ActiveWindowArray active_windows = new ActiveWindowArray();
            ArrayList parent_windows = new ArrayList();

            foreach (int hwnd in active_windows)
            {
                if (!Win32Api.IsWindowVisible(new IntPtr(hwnd)))
                    continue;

                if (Win32Api.GetParent(new IntPtr(hwnd)) != IntPtr.Zero)
                    continue;

                StringBuilder title = new StringBuilder(256);
                Win32Api.GetWindowText(new IntPtr(hwnd), title, title.Capacity);
                if (title.ToString() == "" ||
                    title.ToString() == "Icon" ||
                    title.ToString() == "TitleIconsManager" ||
                    title.ToString() == "Battery Meter" ||
                    title.ToString() == "GDI+ Window" ||
                    title.ToString() == "Program Manager" ||
                    title.ToString() == "HiddenFaxWindow" ||
                    title.ToString() == "SMax4PNP" ||
                    title.ToString() == "MCI command handling window" ||
                    title.ToString() == "CalcMsgPumpWnd" ||
                    title.ToString() == "TaskEng - Task Scheduler Engine Process" ||
                    title.ToString() == "MS_WebcheckMonitor" ||
                    title.ToString() == "AEBalloonTip" ||
                    title.ToString() == "MMDEVAPI Device Window" ||
                    title.ToString() == "BluetoothNotificationAreaIconWindowClass" ||
                    title.ToString() == "Windows Defender" ||
                    title.ToString() == "DWM Notification Window" ||
                    title.ToString() == "F" ||
                    title.ToString() == "Start Menu" ||
                    title.ToString() == "ClubDev" ||
                    title.ToString() == "Icon (Not Responding)" ||
                    title.ToString() == "AutoSuggest Drop-Down" ||
                    title.ToString() == "Add Progress" ||
                    title.ToString() == "Checking out" ||
                    title.ToString() == "Check Out" ||
                    title.ToString() == "Get Progress" ||
                    title.ToString() == "Check In Progress" ||
                    title.ToString() == "Replicated Window Server" ||
                    title.ToString() == "Screen Saver" ||
                    title.ToString() == "SysTrayIcon" ||
                    title.ToString() == "Announce to Group..." ||
                    title.ToString() == "Exception Helper Window")
                    continue;

                // need to check for hover dialogs
                StringBuilder class_name = new StringBuilder(256);
                Win32Api.GetClassName(new IntPtr(hwnd), class_name, class_name.Capacity);

                if (class_name.ToString() == "OfficeTooltip" 
                    || class_name.ToString() == "VBBubble"
                    || class_name.ToString() == "VSDebugger CDataTipWnd"
                    || class_name.ToString() == "SysFader"
                    || class_name.ToString() == "Auto-Suggest Dropdown"
                    || class_name.ToString() == "WindowsScreenSaverClass"
                    || class_name.ToString() == "#32770"
                    || class_name.ToString() == "SysDragImage"
                    || class_name.ToString() == "MsoCommandBarPopup"
                    || class_name.ToString() == "ComboLBox"
                    || class_name.ToString() == "VBFloatingPalette"
                    || class_name.ToString() == "SideBar_HTMLHostWindow"
                    || class_name.ToString() == "#32768"
                    || class_name.ToString() == "IMWindowClass"
                    || class_name.ToString() == "MSBLWindowClass")
                    continue;

                /*Win32Api.WindowInfo wininfo = new Win32Api.WindowInfo();
                wininfo = new Win32Api.WindowInfo();
                wininfo.cbSize = (uint)Marshal.SizeOf(wininfo);
                Win32Api.GetWindowInfo(new IntPtr(hwnd), out wininfo);
                if ((wininfo.dwExStyle & Win32Api.WS_EX_APPWINDOW) != Win32Api.WS_EX_APPWINDOW)
                {
                    if(class_name.ToString()!="MozillaUIWindowClass")
                        continue;
                }*/

                Win32Api.Rect newdims = new Win32Api.Rect();
                Win32Api.GetWindowRect(new IntPtr(hwnd), ref newdims);
                if (newdims.Width <= 0 || newdims.Height <= 0)
                    continue;
                if (newdims.bottom < 0 && newdims.left < 0 && newdims.right < 0 && newdims.top < 0)
                    continue;

                parent_windows.Add(hwnd);
                
            }

            

            // Remove any icons that are from windows no longer active
            ArrayList remove_group = new ArrayList();
            
            //      Get a list of the icons to remove
            foreach (IconGroup icon in icongroups)
            {
                if (!active_windows.Contains(icon.appHandle))
                {
                    Console.WriteLine("Schedule remove for " + icon.appHandle.ToString());
                    remove_group.Add(icon);
                }
            }

            //      Use the list to remove the icons
            foreach (IconGroup removeme in remove_group)
            {
                Console.WriteLine("Removing icon for " + removeme.appHandle.ToString());
             
                removeme.destroy();
                icongroups.Remove(removeme);

                ClubDev.Database.ClubDevDataSet clubDevDataSet = new ClubDev.Database.ClubDevDataSet();
                ClubDev.Database.ClubDevDataSetTableAdapters.SharedArtifactsTableAdapter sharedArtifactsTableAdapter = new ClubDev.Database.ClubDevDataSetTableAdapters.SharedArtifactsTableAdapter();

                sharedArtifactsTableAdapter.DeleteByHWNDandGroup(removeme.appHandle, removeme.collabHost);
            }

            remove_group.Clear();

            


            // Add any new active windows to our list
            // Sorting is necessary for BinarySearch to function
            icongroups.Sort();
            foreach (int hwnd in parent_windows)
            {
                if (icongroups.BinarySearch(hwnd) < 0)
                {
                    //add a new icon group for this window
                    IconGroup new_group = new IconGroup(hwnd, this.collabHost, this.collabName);
                    StringBuilder title = new StringBuilder(256);
                    Win32Api.GetWindowText(new IntPtr(hwnd), title, title.Capacity);
                    if (title.ToString().Contains("Replicated"))
                    {
                        foreach (CaptionButton icon in new_group.icons)
                        {
                            icon.hidden = true;
                            icon.Hide();
                        }
                        new_group.hide_always = true;
                    }
                    Console.WriteLine("Adding icon for " + new_group.appHandle.ToString());
                    icongroups.Add(new_group);
                    StringBuilder class_name = new StringBuilder(256);
                    Win32Api.GetClassName(new IntPtr(hwnd), class_name, class_name.Capacity);
                    Console.WriteLine("   Title = " + title.ToString());
                    Console.WriteLine("   Class = " + class_name.ToString());
                }
            }
            icongroups.Sort();
            
            // Order the Icon groups from front to back
            ArrayList depth_sorted_icons = new ArrayList();

            

            //Get the top most window
            IntPtr temp_ptr = Win32Api.GetTopWindow(IntPtr.Zero);
            int index = icongroups.BinarySearch((int)temp_ptr);

            if (index >= 0)
            {
                depth_sorted_icons.Add(icongroups[index]);
                icongroups.Remove(icongroups[index]);
            }


            // As we walk through all visible windows from front to back, find the
            // corresponding window in icongroups and move it to depth_sorted_icons.
            // This way, depth_sorted_icons will consist of IconGroups that are
            // depth sorted from front to back
            while (icongroups.Count > 0  && temp_ptr!=IntPtr.Zero)
            {
                temp_ptr = Win32Api.GetWindow(temp_ptr, Win32Api.GW_HWNDNEXT);
                index = icongroups.BinarySearch((int)temp_ptr);
                if (index >= 0)
                {
                    depth_sorted_icons.Add(icongroups[index]);
                    icongroups.Remove(icongroups[index]);
                }
            }

            
            

            // Update icon locations
            foreach (IconGroup ig in depth_sorted_icons)
                ig.updateIconPosition();

            
            // Update icon visibility
            for (int i = 0; i < depth_sorted_icons.Count; i++)
            {
                ((IconGroup)depth_sorted_icons[i]).updateIconVisibility(i, depth_sorted_icons);
            }

            foreach (IconGroup leftover in icongroups)
            {
                //Console.WriteLine("leftover found: apphandle = " + leftover.appHandle);
                leftover.destroy();
            }

            icongroups = depth_sorted_icons;
            
            

        }

    }
}