// Copyright (C) Josh Smith - January 2007
using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Documents;
using System.Windows.Media;
using System.Windows.Input;
using System.Windows.Shapes;


namespace ClubDev.SupportProviders
{
	#region CanvasDragDropManager

    public class CanvasDragDropManager<ItemType> where ItemType : class
	{
		#region Data

		bool				            canInitiateDrag;
		DragAdorner			            dragAdorner;
		double				            dragAdornerOpacity;
		bool				            isDragInProgress;
		UIElement			            itemUnderDragCursor;
		Point				            ptMouseDown;
		bool				            showDragAdorner;
        Canvas                          canvas;
        UIElement                       root;

        Utilities.GlobalActivityHook    globalMouseHook;

        object                            dropOnWorkspaceObject;

		#endregion // Data

		#region Constructors

		/// <summary>
		/// Initializes a new instance of ListViewDragManager.
		/// </summary>
		public CanvasDragDropManager()
		{
			this.canInitiateDrag = false;
			this.dragAdornerOpacity = 0.7;
			this.showDragAdorner = true;
		}

		/// <summary>
		/// Initializes a new instance of ListViewDragManager.
		/// </summary>
		/// <param name="listView"></param>
        public CanvasDragDropManager(Canvas canvas, UIElement root) : this()
		{
            this.Canvas = canvas;
            this.root = root;
		}

		/// <summary>
		/// Initializes a new instance of ListViewDragManager.
		/// </summary>
		/// <param name="listView"></param>
		/// <param name="dragAdornerOpacity"></param>
        public CanvasDragDropManager(Canvas canvas, UIElement root, double dragAdornerOpacity) : this(canvas, root)
		{
			this.DragAdornerOpacity = dragAdornerOpacity;
		}

		/// <summary>
		/// Initializes a new instance of ListViewDragManager.
		/// </summary>
		/// <param name="listView"></param>
		/// <param name="showDragAdorner"></param>
        public CanvasDragDropManager(Canvas canvas, UIElement root, bool showDragAdorner) : this(canvas, root)
		{
			this.ShowDragAdorner = showDragAdorner;
		}

    	#endregion // Constructors

		#region Public Interface

		#region DragAdornerOpacity

		/// <summary>
		/// Gets/sets the opacity of the drag adorner.  This property has no
		/// effect if ShowDragAdorner is false. The default value is 0.7
		/// </summary>
		public double DragAdornerOpacity
		{
			get { return this.dragAdornerOpacity; }
			set
			{
				if( this.IsDragInProgress )
					throw new InvalidOperationException( "Cannot set the DragAdornerOpacity property during a drag operation." );

				if( value < 0.0 || value > 1.0 )
					throw new ArgumentOutOfRangeException( "DragAdornerOpacity", value, "Must be between 0 and 1." );

				this.dragAdornerOpacity = value;
			}
		}

			#endregion // DragAdornerOpacity

		#region IsDragInProgress

		/// <summary>
		/// Returns true if there is currently a drag operation being managed.
		/// </summary>
		public bool IsDragInProgress
		{
			get { return this.isDragInProgress; }
			private set { this.isDragInProgress = value; }
		}

			#endregion // IsDragInProgress
		
        #region Canvas

        /// <summary>
        /// Gets/sets the Canvas whose dragging is managed.  This property
        /// can be set to null, to prevent drag management from occuring.  If
        /// the Canvas's AllowDrop property is false, it will be set to true.
        /// </summary>
        public Canvas Canvas
        {
            get { return canvas; }
            set
            {
                if (this.IsDragInProgress)
                    throw new InvalidOperationException("Cannot set the Canvas property during a drag operation.");

                if (this.canvas != null)
                {
                    #region Unhook Events

                    canvas.PreviewMouseLeftButtonDown -= PreviewMouseLeftButtonDown;
                    canvas.PreviewMouseMove -= PreviewMouseMove;
                    canvas.Drop -= Drop;

                    #endregion // Unhook Events
                }

                this.canvas = value;

                if (this.canvas != null)
                {
                    if (!this.canvas.AllowDrop)
                        this.canvas.AllowDrop = true;

                    #region Hook Events

                    canvas.PreviewMouseLeftButtonDown += PreviewMouseLeftButtonDown;
                    canvas.PreviewMouseMove += PreviewMouseMove;
                    canvas.Drop += Drop;

                    #endregion // Hook Events
                }
            }
        }

        #endregion // Canvas	

		#region ShowDragAdorner

		/// <summary>
		/// Gets/sets whether a visual representation of the ListViewItem being dragged
		/// follows the mouse cursor during a drag operation.  The default value is true.
		/// </summary>
		public bool ShowDragAdorner
		{
			get { return this.showDragAdorner; }
			set
			{
				if( this.IsDragInProgress )
					throw new InvalidOperationException( "Cannot set the ShowDragAdorner property during a drag operation." );

				this.showDragAdorner = value;
			}
		}

			#endregion // ShowDragAdorner

		#endregion // Public Interface

        #region Event Handling Methods

        #region PreviewMouseLeftButtonDown       UPDATED - CANVAS SAFE

        void PreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {

            Console.WriteLine("down Button");

            if (this.IsMouseOverScrollbar)
                return;

            UIElement dragItem = null;
            for (int i = 0; i < this.canvas.Children.Count; ++i)
            {
                UIElement item = this.canvas.Children[i];
                if (this.IsMouseOver(item))
                {
                    dragItem = item;
                    break;
                }
            }

            if (dragItem != null)
            {
                this.canInitiateDrag = true;
                this.ItemUnderDragCursor = dragItem;
            }

            if (this.canInitiateDrag)
            {
                // Remember the location and index of the ListViewItem the user clicked on for later.
                this.ptMouseDown = Utilities.MouseUtilities.GetMousePosition(dragItem);
            }
            else {
                this.ptMouseDown = new Point(-10000, -10000);
            }

            this.dropOnWorkspaceObject = ((FrameworkElement)sender).Parent;


        }

        #endregion // PreviewMouseLeftButtonDown

        #region GlobalMouseHookActivity          UPDATED - CANVAS SAFE

        void GlobalMouseHookActivity(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            if (this.ShowDragAdornerResolved)
                this.UpdateDragAdornerLocation();
        }

        #endregion

        #region PreviewMouseMove                 UPDATED - CANVAS SAFE

        void PreviewMouseMove(object sender, MouseEventArgs e)
        {
            if (!this.CanStartDragOperation)
                return;

            if (ItemUnderDragCursor == null)
                return;

            AdornerLayer adornerLayer = this.ShowDragAdornerResolved ? this.InitializeAdornerLayer(ItemUnderDragCursor) : null;

            globalMouseHook = new Utilities.GlobalActivityHook(true, false);
            globalMouseHook.OnMouseActivity += new System.Windows.Forms.MouseEventHandler(GlobalMouseHookActivity);

            this.InitializeDragOperation(ItemUnderDragCursor);
            this.PerformDragOperation(ItemUnderDragCursor);
            this.FinishDragOperation(ItemUnderDragCursor, adornerLayer);

            globalMouseHook.OnMouseActivity -= GlobalMouseHookActivity;
            globalMouseHook.Stop();
            globalMouseHook = null;


        }

        #endregion // PreviewMouseMove

        #region Drop                             UPDATED - CANVAS SAFE

        void Drop(object sender, DragEventArgs e)
        {
            e.Effects = DragDropEffects.Move;

            

            

            // Check to make sure data is in the event container
            if (!e.Data.GetDataPresent(typeof(ItemType)))
                return;

            // Get the data object which was dropped.
            ItemType data = e.Data.GetData(typeof(ItemType)) as ItemType;
            if (data == null)
                return;


            Canvas dataCanvas = data as Canvas;

            Type parentType = ((FrameworkElement)dataCanvas.Parent).Parent.GetType();
            Type destinationType = ((FrameworkElement)sender).Parent.GetType();

            if (parentType.Name.Equals("SharedScreenItemsSummary"))
            {
                SharedScreenItemsSummary sourceSharedScreenItems = ((FrameworkElement)dataCanvas.Parent).Parent as SharedScreenItemsSummary;
                SharedScreenItemsSummary destinationSharedScreenItems = ((FrameworkElement)sender).Parent as SharedScreenItemsSummary;

                if (sourceSharedScreenItems.Equals(destinationSharedScreenItems))
                    sourceSharedScreenItems.RepositionItem(dataCanvas, Utilities.MouseUtilities.GetMousePosition((FrameworkElement)sender));

                else
                    Console.WriteLine("DO MOVE OF APPLICATION");
                
                    
                    //sourceCollaboratorItems.RemoveItem(dataCanvas);

                //destinationCollaboratorItems.AddItem(dataCanvas, Utilities.MouseUtilities.GetMousePosition((FrameworkElement)sender));


                //Don't allow the drag to be handled elsewhere
                this.dropOnWorkspaceObject = null;
            }

            else if(parentType == typeof(CollaboratorItemsGrid))
            {
                Console.WriteLine("collaborator");


                if (((FrameworkElement)dataCanvas.Parent).Parent.Equals(((FrameworkElement)sender).Parent))
                {
                    CollaboratorItemsGrid sourceCollabotatorItemsGrid = ((FrameworkElement)dataCanvas.Parent).Parent as CollaboratorItemsGrid;
                    sourceCollabotatorItemsGrid.RepositionItem(dataCanvas, Utilities.MouseUtilities.GetMousePosition((FrameworkElement)sender));
                }


                //Don't allow the drag to be handled elsewhere
                this.dropOnWorkspaceObject = null;
            }

            else 
            {
                Console.WriteLine("UN HANDLED DRAG EVENT");
            }
        }

        #endregion // grid_Drop

        #endregion // Event Handling Methods

        #region Private Helpers                  UPDATED

        #region CanStartDragOperation            UPDATED

        bool CanStartDragOperation
        {
            get
            {
                if (Mouse.LeftButton != MouseButtonState.Pressed)
                    return false;

                if (!this.canInitiateDrag)
                    return false;

                //if (this.columnToSelect == -1)
                //    return false;

                //if (!this.HasCursorLeftDragThreshold)
                //    return false;

                return true;
            }
        }

        #endregion // CanStartDragOperation

        #region FinishDragOperation              UPDATED - CANVAS SAFE

        void FinishDragOperation(UIElement draggedItem, AdornerLayer adornerLayer)
        {
            //operation was  between sources; it was dragged across items, or into local workspace
            if (this.dropOnWorkspaceObject != null)
            {
                bool isOverItem = false;

                Canvas root = ((Canvas)this.root);
                for (int i = 0; i < root.Children.Count; i++)
                {
                    if (IsMouseOver(root.Children[i]))
                    {


                        if (root.Children[i].GetType() == typeof(SharedScreen))
                        {
                            isOverItem = true;

                            if (this.canvas.Parent.GetType() == typeof(CollaboratorItemsGrid))
                            {
                                Console.WriteLine("source is a collaborator");

                                CollaboratorItemsGrid sourceCollaborator = (CollaboratorItemsGrid)this.canvas.Parent;
                                SharedScreen targetSharedScreen = (SharedScreen)root.Children[i];

                                targetSharedScreen.InitiateShare(sourceCollaborator, ((FrameworkElement)draggedItem).Name);

                                break;
                            }

                            else if (this.canvas.Parent.GetType() == typeof(SharedScreenItemsFullView))
                            {
                                Console.WriteLine("source is shared screen full view");

                                SharedScreenItemsFullView sourceFullView = (SharedScreenItemsFullView)this.canvas.Parent;
                                SharedScreen targetSharedScreen = (SharedScreen)root.Children[i];

                                if (sourceFullView.Parent == targetSharedScreen)
                                {
                                    Console.WriteLine("moving in local application");

                                    
                                            sourceFullView.RepositionLocalArtifact(draggedItem);
                                }
                            }
                        }

                        else if (root.Children[i].GetType() == typeof(Grid))
                        {
                            Console.WriteLine("OVER GRID");
                            Grid collaboratorGrid = (Grid)root.Children[i];
                            for (int j = 0; j < collaboratorGrid.Children.Count; j++)
                            {
                                if (IsMouseOver(collaboratorGrid.Children[j]))
                                    if (collaboratorGrid.Children[j].GetType() == typeof(Collaborator))
                                    {
                                        Collaborator currentCollaborator = (Collaborator)collaboratorGrid.Children[j];

                                        for (int k = 0; k < currentCollaborator.Children.Count; k++)
                                        {
                                            if ((IsMouseOver(currentCollaborator.Children[k])) && currentCollaborator.Children[k].GetType() == typeof(CollaboratorSummary))
                                            {
                                                CollaboratorSummary currentCollaboratorSummary = (CollaboratorSummary)currentCollaborator.Children[k];

                                                for (int m = 0; m < currentCollaboratorSummary.Children.Count; m++)
                                                {
                                                    if ((IsMouseOver(currentCollaboratorSummary.Children[m])) && currentCollaboratorSummary.Children[m].GetType() == typeof(CollaboratorItemsGrid))
                                                    {
                                                        Console.WriteLine("OVER COLLABORATOR SUMMARY");
                                                        isOverItem = true;

                                                        if (this.canvas.Parent.GetType() == typeof(CollaboratorItemsGrid))
                                                        {
                                                            Console.WriteLine("source is collaborator");

                                                            CollaboratorItemsGrid sourceCollaborator = (CollaboratorItemsGrid)this.canvas.Parent;
                                                            CollaboratorItemsGrid targetCollaborator = (CollaboratorItemsGrid)currentCollaboratorSummary.Children[m];

                                                            targetCollaborator.InitiateRelocation(sourceCollaborator, ((FrameworkElement)draggedItem).Name);

                                                        }

                                                        break;
                                                    }
                                                }
                                                break;
                                            }
                                        }
                                        break;
                                    }
                            }

                        }
                        else
                            Console.WriteLine("Over something else: " + root.Children[i].GetType().ToString());
                    
                    
                    }
                }

                if (!isOverItem)
                {   // The item being dragged is not over anything, so this means it was dragged into the local workspace
                
                    Console.WriteLine("dragged in to local workspace");

                    if (dropOnWorkspaceObject.GetType() == typeof(CollaboratorItemsGrid))
                        ((CollaboratorItemsGrid)dropOnWorkspaceObject).InitiateRelocationToLocalWorkspace(((FrameworkElement)draggedItem).Name);

                    if (dropOnWorkspaceObject.GetType() == typeof(SharedScreenItemsSummary))
                        ((SharedScreenItemsSummary)dropOnWorkspaceObject).InitiateRelocationToLocalWorkspace(((FrameworkElement)draggedItem).Name);
                }

            }

            this.dropOnWorkspaceObject = null;


            // Let the ListViewItem know that it is not being dragged anymore.
            GridItemDragState.SetIsBeingDragged(draggedItem, false);

            this.IsDragInProgress = false;

            if (this.ItemUnderDragCursor != null)
                this.ItemUnderDragCursor = null;

            // Remove the drag adorner from the adorner layer.
            if (adornerLayer != null)
            {
                adornerLayer.Remove(this.dragAdorner);
                this.dragAdorner = null;
            }
        }

        #endregion // FinishDragOperation

        #region HasCursorLeftDragThreshold       UPDATED

        bool HasCursorLeftDragThreshold
        {
            get
            {
                if (this.ItemUnderDragCursor == null)
                    return false;

                UIElement item = this.ItemUnderDragCursor;
                Rect bounds = VisualTreeHelper.GetDescendantBounds(item);
                Point ptInItem = this.canvas.TranslatePoint(this.ptMouseDown, item);

                // In case the cursor is at the very top or bottom of the item
                // we want to make the vertical threshold very small so that dragging
                // over an adjacent item does not select it.
                double topOffset = Math.Abs(ptInItem.Y);
                double btmOffset = Math.Abs(bounds.Height - ptInItem.Y);
                double vertOffset = Math.Min(topOffset, btmOffset);

                double width = SystemParameters.MinimumHorizontalDragDistance * 2;
                double height = Math.Min(SystemParameters.MinimumVerticalDragDistance, vertOffset) * 2;
                Size szThreshold = new Size(width, height);

                Rect rect = new Rect(this.ptMouseDown, szThreshold);
                rect.Offset(szThreshold.Width / -2, szThreshold.Height / -2);
                Point ptInListView = Utilities.MouseUtilities.GetMousePosition(this.canvas);
                return !rect.Contains(ptInListView);
            }
        }

        #endregion // HasCursorLeftDragThreshold

        #region InitializeAdornerLayer           UPDATED - CANVAS SAFE

        AdornerLayer InitializeAdornerLayer(UIElement itemToDrag)
        {
            // Create a brush which will paint the ListViewItem onto
            // a visual in the adorner layer.
            VisualBrush brush = new VisualBrush(itemToDrag);

            // Create an element which displays the source item while it is dragged.
            this.dragAdorner = new DragAdorner(itemToDrag, itemToDrag.RenderSize, brush);

            // Set the drag adorner's opacity.		
            this.dragAdorner.Opacity = this.DragAdornerOpacity;

            AdornerLayer layer = AdornerLayer.GetAdornerLayer(itemToDrag);
            layer.Add(dragAdorner);

            // Save the location of the cursor when the left mouse button was pressed.
            this.ptMouseDown = Utilities.MouseUtilities.GetMousePosition(itemToDrag);

            return layer;
        }

        #endregion // InitializeAdornerLayer

        #region InitializeDragOperation          UPDATED - CANVAS SAFE

        void InitializeDragOperation(UIElement itemToDrag)
        {
            // Set some flags used during the drag operation.
            this.IsDragInProgress = true;
            this.canInitiateDrag = false;

            // Let the ListViewItem know that it is being dragged.
            GridItemDragState.SetIsBeingDragged(itemToDrag, true);
        }

        #endregion // InitializeDragOperation

        #region IsMouseOver                      UPDATED

        bool IsMouseOver(Visual target)
        {
            // We need to use MouseUtilities to figure out the cursor
            // coordinates because, during a drag-drop operation, the WPF
            // mechanisms for getting the coordinates behave strangely.

            Rect bounds = VisualTreeHelper.GetDescendantBounds(target);
            Point mousePos = Utilities.MouseUtilities.GetMousePosition(target);

            Console.WriteLine("IsMouseOver: bounds: " + bounds.ToString());
            Console.WriteLine("IsMouseOver: mousePos: " + mousePos.ToString());

            return bounds.Contains(mousePos);

        }

        #endregion // IsMouseOver

        #region IsMouseOverScrollbar             UPDATED

        /// <summary>
        /// Returns true if the mouse cursor is over a scrollbar in the ListView.
        /// </summary>
        bool IsMouseOverScrollbar
        {
            get
            {
                Point ptMouse = Utilities.MouseUtilities.GetMousePosition(this.canvas);
                HitTestResult res = VisualTreeHelper.HitTest(this.canvas, ptMouse);
                if (res == null)
                    return false;

                DependencyObject depObj = res.VisualHit;
                while (depObj != null)
                {
                    if (depObj is ScrollBar)
                        return true;

                    // VisualTreeHelper works with objects of type Visual or Visual3D.
                    // If the current object is not derived from Visual or Visual3D,
                    // then use the LogicalTreeHelper to find the parent element.
                    if (depObj is Visual || depObj is System.Windows.Media.Media3D.Visual3D)
                        depObj = VisualTreeHelper.GetParent(depObj);
                    else
                        depObj = LogicalTreeHelper.GetParent(depObj);
                }

                return false;
            }
        }

        #endregion // IsMouseOverScrollbar

        #region ItemUnderDragCursor              UPDATED

        UIElement ItemUnderDragCursor
        {
            get { return this.itemUnderDragCursor; }
            set
            {
                if (this.itemUnderDragCursor == value)
                    return;

                        this.itemUnderDragCursor = value;

                        if (this.itemUnderDragCursor != null)
                        {
                            GridItemDragState.SetIsUnderDragCursor(this.itemUnderDragCursor, true);
                        }
                    
                
            }
        }

        #endregion // ItemUnderDragCursor

        #region PerformDragOperation             UPDATED - CANVAS SAFE

        void PerformDragOperation(UIElement itemToDrag)
        {
            ItemType dragItem = itemToDrag as ItemType;
            DragDropEffects allowedEffects = DragDropEffects.Move;
            DragDrop.DoDragDrop(this.root, dragItem, allowedEffects);
        }

        #endregion // PerformDragOperation

        #region ShowDragAdornerResolved          UPDATED - CANVAS SAFE

        bool ShowDragAdornerResolved
        {
            get { return this.ShowDragAdorner && this.DragAdornerOpacity > 0.0; }
        }

        #endregion // ShowDragAdornerResolved

        #region UpdateDragAdornerLocation        UPDATED - CANVAS SAFE

        void UpdateDragAdornerLocation()
        {
            if (this.dragAdorner != null)
            {
                Point ptCursor = Utilities.MouseUtilities.GetMousePosition(this.ItemUnderDragCursor);

                this.dragAdorner.SetOffsets(
                   ptCursor.X - this.ptMouseDown.X,
                   ptCursor.Y - this.ptMouseDown.Y);
            }
        }

        #endregion // UpdateDragAdornerLocation

        #endregion // Private Helpers


    #endregion //GridDragDropManager
    }


    #region GridItemDragState                    UPDATED

    /// <summary>
    /// Exposes attached properties used in conjunction with the ListViewDragDropManager class.
    /// Those properties can be used to allow triggers to modify the appearance of ListViewItems
    /// in a ListView during a drag-drop operation.
    /// </summary>
    /// <remarks>
    /// Documentation: http://www.codeproject.com/useritems/ListViewDragDropManager.asp
    /// </remarks>
    public static class GridItemDragState
    {
        #region IsBeingDragged                   UPDATED

        /// <summary>
        /// Identifies the GridItemDragState's IsBeingDragged attached property.  
        /// This field is read-only.
        /// </summary>
        public static readonly DependencyProperty IsBeingDraggedProperty =
            DependencyProperty.RegisterAttached(
                "IsBeingDragged",
                typeof(bool),
                typeof(GridItemDragState),
                new UIPropertyMetadata(false));

        /// <summary>
        /// Returns true if the specified item is being dragged, else false.
        /// </summary>
        /// <param name="item">The item to check.</param>
        public static bool GetIsBeingDragged(UIElement item)
        {
            return (bool)item.GetValue(IsBeingDraggedProperty);
        }

        /// <summary>
        /// Sets the IsBeingDragged attached property for the specified item.
        /// </summary>
        /// <param name="item">The item to set the property on.</param>
        /// <param name="value">Pass true if the element is being dragged, else false.</param>
        internal static void SetIsBeingDragged(UIElement item, bool value)
        {
            item.SetValue(IsBeingDraggedProperty, value);
        }

        #endregion // IsBeingDragged

        #region IsUnderDragCursor                UPDATED

        /// <summary>
        /// Identifies the GridItemDragState's IsUnderDragCursor attached property.  
        /// This field is read-only.
        /// </summary>
        public static readonly DependencyProperty IsUnderDragCursorProperty =
            DependencyProperty.RegisterAttached(
                "IsUnderDragCursor",
                typeof(bool),
                typeof(GridItemDragState),
                new UIPropertyMetadata(false));

        /// <summary>
        /// Returns true if the specified item is currently underneath the cursor 
        /// during a drag-drop operation, else false.
        /// </summary>
        /// <param name="item">The item to check.</param>
        public static bool GetIsUnderDragCursor(UIElement item)
        {
            return (bool)item.GetValue(IsUnderDragCursorProperty);
        }

        /// <summary>
        /// Sets the IsUnderDragCursor attached property for the specified item.
        /// </summary>
        /// <param name="item">The item to set the property on.</param>
        /// <param name="value">Pass true if the element is underneath the drag cursor, else false.</param>
        internal static void SetIsUnderDragCursor(UIElement item, bool value)
        {
            item.SetValue(IsUnderDragCursorProperty, value);
        }

        #endregion // IsUnderDragCursor
    }

    #endregion // GridItemDragState


}