//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//                      Cursor Redirection Toolkit (CuRT)                       //
//               Written by Andy McGovern, Orchid Research Group                //
//                           andymcgovern@gmail.com                             //
//                          http://orchid.cs.uiuc.edu                           //
//  Department of Computer Science, University of Illinois at Urbana-Champaign  //
//                                July 4, 2006                                  //
//      Copyright 2006, The Board of Trustees of the University of Illinois     //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Net;
using System.Text;
using System.Net.Sockets;
using System.Threading;
using System.Windows.Forms;
using System.Diagnostics;
using System.Drawing;

namespace UIUC.DCS.ORCHID.CuRTClientLibrary
{

    //! This enumerates the possible key modifiers that can be chosen as part of the hotkey
    public enum KeyModifiers
    {
        Alt = 1,
        Control = 2,
        Windows = 8
    }

    //! This enumerates the possible colors that can be chosen for a virtual cursor
    public enum CursorColors
    {
        None = -1,
        Green = 0,
        Blue = 1,
        Red = 2,
        Yellow = 3,
        Magenta = 4,
        Gray = 5,
        Purple = 6,
        Aqua = 7,
        Brown = 8,
        White = 9
    }

    //! This enumerates the possible screen edges, for use when edge detection events are fired
    public enum ScreenEdges
    {
        Top = 0,
        Bottom = 1,
        Left = 2,
        Right = 3
    }

    //! CuRTClient class
    /*!
     *  The CuRTClient class represents one client connected to a CuRTServer.  The CuRTClient has two main
     *  modes of operation, represented by the two different CuRTClient constructors.  One is the standalone
     *  application mode.  In this mode, the user manually specifies the connection parameters, and a GUI allows
     *  the user to send his cursor/keyboard input to other machines in the system and then return his
     *  cursor/keyboard input back to his own machine.  To execute the CuRTClient in this mode, just run this compiled
     *  code (the main method of this class starts the CuRTClient in the standalone application mode).
     *  The other mode is to use the CuRTClient as an API.  In this mode, a developer specifies all connection parameters
     *  in the CuRTClient constructor.  Then, method calls in this class can be called to manipulate the client and to
     *  interact with other clients.
     */
    public class CuRTClient
    {
        //! this is a reference to the CuRTGUI object that allows the user to interact with the system
        protected internal CuRTGUI GUI;

        //! this is a reference to the CursorModule object that controls all cursor and keyboard interactions
        protected internal CursorModule cursorModule;

        //! this is a reference to the SocketModule object that manages all CuRT network communications
        protected internal SocketModule socketModule;

        //! this integer represents the index (0-9) of this local client
        protected internal int localIndex;

        //! this String array contains the usernames of all clients currently connected to the CuRT system
        protected internal String[] clientUsernameList;

        //! this is the username of the local client, to be sent to the server during initialization
        private String localClientUsername;

        //! this is the server's IP address
        private String IP;

        //! the hotkey's key modifier (ALT, CTRL, or WINKEY)
        private KeyModifiers keyModifier;

        //! the hotkey used with a key modifier to return the local cursor to the local machine
        private Keys hotkey;

        //! cursor color scheme
        /*!
         *  This is a 10-item array of enumerated CursorColors.  Each index cooresponds to a potential client connected
         *  to the system, and this array is used to determine the color of the cursors in the system, from the
         *  perspective of this local system.
         */
        private CursorColors[] colorScheme;

        //! cursorMovementResolution
        /*!
         *  At times, the CuRT system can experience lag when several cursors are being remoted to the same machine
         *  simultaneously.  To help allevaite this problem, the cursorMovementResolution parameter can be adjusted.
         *  When a cursor is being remoted, this parameter represents the number of cursor movements that are ignored 
         *  on the local machine before a movement is sent to the remote machine.  For example, a value of 1 means 
         *  that all cursor movements are sent from the local machine to the remot machine, and a value of 3 means
         *  that only every third cursor movement is sent from the local machine to the remote machine.  Note that 
         *  coordinates are sent relatively, not absolutely, but when the cursorMovementResolution parameter has 
         *  a value greater than 1, the values that are sent represent the movement since the last time
         *  that the coordiantes were sent.  Also note that in the standalone version of CuRT system, this parameter
         *  has a default value of 3.
         */
        private int cursorMovementResolution;


        //! this encapsulates the set of initialization parameters needed when using the CuRTClient as an API 
        private EdgeDetectionParams edgeDetectionParams;

        //!  CuRTClient
        /*!
         * This is one of two CuRTClient constructors.  This one is used in standalone application mode.  This one
         * simply starts the CuRTGUI, which then starts the ConnectionBox.  The ConnectionBox allows the user to
         * specify the connection parameters manually.
        */
        public CuRTClient()
        {
            localIndex = -1;
            ExecuteGUI();
        }

        //!  CuRTClient
        /*!
         * This is one of two CuRTClient constructors.  This one is used the API mode.  This one takes in the
         * same parameters that the user specifies manually in the standalone application version.  The GUI is started
         * here as well, but the ConnectionBox is bypassed.
        */
        public CuRTClient(String newClientName, String newIP, KeyModifiers newModkey, Keys newHotkey,
            CursorColors[] newColorScheme, int newCursorMovementResolution, EdgeDetectionParams newEdgeDetectionParams)
        {
            this.localClientUsername = newClientName;
            this.IP = newIP;
            this.keyModifier = newModkey;
            this.hotkey = newHotkey;
            this.colorScheme = newColorScheme;
            this.cursorMovementResolution = newCursorMovementResolution;
            edgeDetectionParams = newEdgeDetectionParams;



            localIndex = -1;

            Thread ExecuteGUIWithParams = new Thread(new ThreadStart(ExecuteGUIWithParameters));
            ExecuteGUIWithParams.Name = "ExecuteGUIWithParameters";
            ExecuteGUIWithParams.IsBackground = true;
            ExecuteGUIWithParams.Start();

            //ExecuteGUIWithParameters();
        }

        public void Disconnect()
        {
            this.socketModule.SendUDPMessage("EXIT");
            Thread.Sleep(500);
            this.socketModule.Dispose();
            this.socketModule = null;

            GUI.Close();

            Application.Exit();
        }

        //!  ExecuteGUI
        /*!
         * This method is used by the standalone application version of the CuRTClient constructor to
         * start the CuRTGUI.
        */
        protected internal void ExecuteGUI()
        {
            GUI = new CuRTGUI(this);
            GUI.Show();
            Application.Run(GUI);
        }

        //!  ExecuteGUI
        /*!
         * This method is used by the API version of the CuRTClient constructor to start the CuRTGUI.
         * The appropriate parameters (stored as CuRTClient class member variables) are passed to the CuRTGUI.
         */
        [STAThread]
        protected internal void ExecuteGUIWithParameters()
        {
            GUI = new CuRTGUI(this, localClientUsername, IP, keyModifier, hotkey, colorScheme, cursorMovementResolution,
                edgeDetectionParams);
            GUI.Show();
            Application.Run(GUI);
        }

        //!  AnotherInstanceRunning
        /*!
         * This method counts the number of processes running with the same name as the current one.  It then returns
         * true if this result is 1 (meaning that this process is the only one running), or false otherwise.  This is
         * used to prevent two instances of the CuRTClient from running at the same time.
         */
        public static bool AnotherInstanceRunning()
        {
            Process thisProcess = Process.GetCurrentProcess();
            Process[] processes = Process.GetProcessesByName(thisProcess.ProcessName);

            if (processes.Length == 1)
                return false;
            else
                return true;
        }

        //!  InitializeSocketModule
        /*!
         * This method initializes the SocketModule and starts its associated network listeners in a separate thread.
         */
        protected internal bool InitializeSocketModule(String serverIP, String username)
        {
            socketModule = new SocketModule(serverIP, this);
            bool successfulConnect = socketModule.ConnectToServer(username);
            if (!successfulConnect)
                return false;

            Thread SocketListeners = new Thread(new ThreadStart(socketModule.InitializeListeners));
            SocketListeners.Name = "SocketListeners";
            SocketListeners.IsBackground = true;
            SocketListeners.Start();

            return true;
        }

        //!  InitializeCursorModule
        /*!
         * This is one of two overloaded methods to initialize the CursorModule.  This one is used in the
         * standalone application version of the CuRT system.
         */
        protected internal bool InitializeCursorModule()
        {

            cursorModule = new CursorModule(this);
            cursorModule.InitializeCursors();
            return true;
        }

        //!  InitializeCursorModule
        /*!
         * This is one of two overloaded methods to initialize the CursorModule.  This one is used in the
         * API version of the CuRT system.
         */
        protected internal bool InitializeCursorModule(CursorColors[] newColorScheme,
            int newCursorMovementResolution, EdgeDetectionParams newEdgeDetectionParams)
        {

            cursorModule = new CursorModule(this, newColorScheme,
            newCursorMovementResolution, newEdgeDetectionParams);
            cursorModule.InitializeCursors();
            return true;
        }


        //!  GetClientUsername
        /*!
         * This is an API method that returns the username of the client at the specified index.  If there is no
         * client at that specified index, the String "[EMPTY]" is returned.
         */
        public String GetClientUsername(int index)
        {
            return clientUsernameList[index];
        }

        //!  GetClientIndex
        /*!
         * This is an API method that returns the index of the client with the specified username.  If there is no
         * client with the specified username, -1 is returned.  Note: in the case of multiple clients with the same
         * username, the first matching username is returned.
         */
        public int GetClientIndex(String name)
        {
            for (int i = 0; i < 10; i++)
                if (name == clientUsernameList[i])
                    return i;

            return -1;
        }

        //!  GetLocalClientUsername
        /*!
         * This is an API method that returns username of the local client.
         */
        public String GetLocalClientUsername()
        {
            return localClientUsername;
        }

        //!  GetLocalClientIndex
        /*!
         * This is an API method that returns index of the local client.
         */
        public int GetLocalClientIndex()
        {
            return localIndex;
        }

        //!  GetClientNameList
        /*!
         * This is an API method that returns a String array containing the usernames of all clients currently connected
         * to the CuRT system.  For every index where no client is connected, the String "[EMPTY]" is returned.
         */
        public String[] GetClientNameList()
        {
            return clientUsernameList;
        }

        //!  GetClientNamesCurrentlyRemotedToLocalMachine
        /*!
         * This is an API method that a String array containing the usernames of all clients whose cursors
         * are currently being remoted to the local machine.  For every index where the client is either
         * not being remoted to the local machine or is not connected to the system, the String "[EMPTY]" is returned.
         */
        public String[] GetClientNamesCurrentlyRemotedToLocalMachine()
        {
            String[] newList = new String[10];
            for (int i = 0; i < 10; i++)
                if (cursorModule.cursors[i].Visible || cursorModule.blockedCursors[i].Visible)
                    newList[i] = clientUsernameList[i];
                else
                    newList[i] = "[EMPTY]";

            return newList;
        }

        //!  GetClientIndicesCurrentlyRemotedToLocalMachine
        /*!
         * This is an API method that returns a boolean array indicating which clients are currently remoting 
         * their cursors to the local machine.  A value of true at a given index indicates that the client at that
         * index is currently remoting to the local machine, and a value of false indicates that the client at that
         * index is either not currently remoting to the local machine, or is not connected to the system.
         */
        public bool[] GetClientIndicesCurrentlyRemotedToLocalMachine()
        {
            bool[] newList = new bool[10];
            for (int i = 0; i < 10; i++)
                if (cursorModule.cursors[i].Visible || cursorModule.blockedCursors[i].Visible)
                    newList[i] = true;
                else
                    newList[i] = false;

            return newList;
        }

        //!  IsLocalCursorCurrentlyRemote
        /*!
         * This is an API method that returns a boolean indicating whether the local machine's cursor is currently
         * being remoted to another machine in the CuRT system.
         */
        public bool IsLocalCursorCurrentlyRemote()
        {
            return cursorModule.localCursorIsRemote;
        }

        //!  GetLocalCursorRemoteIndexLocation
        /*!
         * This is an API method that returns an integer indicating the index of the client where the local cursor
         * is currently being remoted.  If the local cursor is not currently being remoted, a value of -1 is returned.
         */
        public int GetLocalCursorRemoteIndexLocation()
        {
            return cursorModule.localCursorRemoteIndex;
        }

        //!  GetLocalCursorRemoteIndexName
        /*!
         * This is an API method that returns a String representing the username of the of the client where the local cursor
         * is currently being remoted.  If the local cursor is not currently being remoted, the String "[Empty]" is returned.
         */
        public String GetLocalCursorRemoteUsernameLocation()
        {
            if (cursorModule.localCursorRemoteIndex == -1)
                return "[Empty]";

            return GetClientUsername(cursorModule.localCursorRemoteIndex);
        }

        //!  GetLocalCursorColor
        /*!
         * This is an API method that returns an enumerated CursorColors value indicating the local cursor's color.
         */
        public CursorColors GetLocalCursorColor()
        {
            return cursorModule.cursors[localIndex].currColor;
        }

        //!  GetCursorColor
        /*!
         * This is an API method that returns an enumerated CursorColors value indicating the cursor's color at
         * the specified index.
         */
        public CursorColors GetCursorColor(int index)
        {
            return cursorModule.cursors[index].currColor;
        }

        //!  GetCursorColor
        /*!
         * This is an API method that returns an enumerated CursorColors value indicating the color of the cursor
         * with the specified client name.
         */
        public CursorColors GetCursorColor(String name)
        {
            return cursorModule.cursors[GetClientIndex(name)].currColor;
        }

        //!  GetLocalCursorPosition
        /*!
         * This is an API method that returns a Point value indicating the local cursor's coordinate position
         * on the local machine.  If the local cursor is currently being remoted, an empty Point object is returned.
         */
        public Point GetLocalCursorPosition()
        {
            if (cursorModule.localCursorIsRemote)
                return Point.Empty;
            else
            {
                Point currPos = new Point();
                CursorModule.GetCursorPos(ref currPos);
                return currPos;
            }
        }

        //!  GetCursorPosition
        /*!
         * This is an API method that returns a Point value indicating the coordinate position on the local machine
         * of the cursor with the speficied client index.  If the client at the specified index is not currently
         * being remoted to the local machine, or is not connected to the system at all, an empty Point object is 
         * returned.
         */
        public Point GetCursorPosition(int index)
        {
            if (index == localIndex)
                return GetLocalCursorPosition();

            if (cursorModule.cursors[index].Visible || cursorModule.blockedCursors[index].Visible)
                return new Point(cursorModule.cursors[index].Bounds.X, cursorModule.cursors[index].Bounds.Y);
            else
                return Point.Empty;

        }

        //!  GetCursorPosition
        /*!
         * This is an API method that returns a Point value indicating the coordinate position on the local machine
         * of the cursor with the speficied client username.  If the client with the specified username is not 
         * currently being remoted to the local machine, or there is no such client with the speficied username,
         * an empty Point object is returned.
         */
        public Point GetCursorPosition(String name)
        {
            int index = GetClientIndex(name);

            if (index == localIndex)
                return GetLocalCursorPosition();

            if (index == -1)
                return Point.Empty;

            if (cursorModule.cursors[index].Visible || cursorModule.blockedCursors[index].Visible)
                return new Point(cursorModule.cursors[index].Bounds.X, cursorModule.cursors[index].Bounds.Y);
            else
                return Point.Empty;
        }

        //!  GetLocalCursorScreen
        /*!
         * This is an API method that returns a Screen object that represents the screen (for use in multi-screen
         * setups) where the local cursor is currently located.  If the local cursor is currently being remoted, 
         * a null object is returned.
         */
        public Screen GetLocalCursorScreen()
        {
            Point currPos = new Point();
            CursorModule.GetCursorPos(ref currPos);

            if (cursorModule.localCursorIsRemote)
                return null;
            else
                return Screen.FromPoint(currPos);

        }


        //!  GetCursorScreen
        /*!
         * This is an API method that returns a Screen object that represents the screen (for use in multi-screen
         * setups) where the cursor with the specified index is currently located.  If there is no cursor at that
         * index, or if that cursors is not currently being remoted to the local machine, a null object is returned.
         */
        public Screen GetCursorScreen(int index)
        {

            if (cursorModule.cursors[index].Visible || cursorModule.blockedCursors[index].Visible)
                return Screen.FromControl(cursorModule.cursors[index]);
            else
                return null;

        }


        //!  GetCursorScreen
        /*!
         * This is an API method that returns a Screen object that represents the screen (for use in multi-screen
         * setups) where the cursor with whose client has the specified username is currently located.  If there 
         * is no client with that username, or if the client with that username is not currently remoting its
         * cursor to the local machine, a null object is returned.
         */
        public Screen GetCursorScreen(String name)
        {
            int index = GetClientIndex(name);

            if (index == -1 || (!cursorModule.cursors[index].Visible && !cursorModule.blockedCursors[index].Visible))
                return null;
            else
                return Screen.FromControl(cursorModule.cursors[index]);

        }


        //!  RemoveLocalCursorFromRemote
        /*!
         * This is an API method that attempts to remove the local cursor from a remote client.  If the local cursor
         * is currently being remoted on a remote client, then it is returned to the local machine and is moved
         * to the coordinates of the newLocation parameter, and true is returned. If it is not currently beind remoted,
         * then false is returned.  If the point parameter is an empty point object, the cursor's new location is 
         * set to (0,0).
         */
        public bool RemoveLocalCursorFromRemote(Point newLocation)
        {
            if (newLocation == Point.Empty)
                newLocation = new Point(0, 0);

            if (cursorModule.localCursorIsRemote)
            {
                GUI.CleanUpToReleaseLocalCursor();
                cursorModule.localCursorIsRemote = false;
                cursorModule.localCursorRemoteIndex = -1;
                cursorModule.MoveLocalCursor(newLocation.X, newLocation.Y);
                return socketModule.RemoveLocalCursorFromRemote();
            }
            return false;

        }

        //!  SendLocalCursorToRemoteMachine
        /*!
         * This is an API method that sends the local cursor to a remote machine, at the specified X and Y
         * coordinates.  The current location of the local cursor on the local machine is returned as a Point
         * object.  If the local cursor is already being remoted, it is first removed from the remote
         * machine, and then sent to the new remote machine (an Empty point is returned in that case).  If the
         * index parameter is the same as the local index, an Empty point is returned.
         */
        public Point SendLocalCursorToRemoteMachine(int index, Point startingLocation)
        {
            Point returnPoint = Point.Empty;

            if (index == localIndex)
                return returnPoint;

            if (cursorModule.localCursorIsRemote)
            {
                RemoveLocalCursorFromRemote(Point.Empty);
            }
            else
            {
                returnPoint = GetLocalCursorPosition();
            }

            GUI.PrepareToSendLocalCursor();

            cursorModule.localCursorIsRemote = true;
            cursorModule.localCursorRemoteIndex = index;
            socketModule.SendCursorToRemoteMachine(localIndex, index, startingLocation.X, startingLocation.Y);

            return returnPoint;
        }

        //!  SendLocalCursorToRemoteMachine
        /*!
         * This is an API method that sends the local cursor to a remote machine, at the specified X and Y
         * coordinates.  The current location of the local cursor on the local machine is returned as a Point
         * object.  If the local cursor is already being remoted, it is first removed from the remote
         * machine, and then sent to the new remote machine (an Empty point is returned in that case).  If the
         * index parameter is the same as the local index, an Empty point is returned.
         */
        public Point SendLocalCursorToRemoteMachine(String username, Point startingLocation)
        {
            Point returnPoint = Point.Empty;

            if (username == localClientUsername)
                return returnPoint;

            int index = GetClientIndex(username);

            if (cursorModule.localCursorIsRemote)
            {
                RemoveLocalCursorFromRemote(Point.Empty);
            }
            else
            {
                returnPoint = GetLocalCursorPosition();
            }

            GUI.PrepareToSendLocalCursor();

            cursorModule.localCursorIsRemote = true;
            cursorModule.localCursorRemoteIndex = index;
            socketModule.SendCursorToRemoteMachine(localIndex, index, startingLocation.X, startingLocation.Y);

            return returnPoint;
        }

        //!  SendRemoteCursorToRemoteMachine
        /*!
         * This is an API method that attempts to send a virtual cursor from the local machine to a remote
         * machine.  If there is no virtual cursor on the local machine with the specified index, or if there 
         * is no client connected at the indicated destination index, or if the specified cursor index is the 
         * local index, or if the specified destination machine index is the local index, an empty point object 
         * is returned.  Otherwise, the virtual cursor is passed to the indicated remote 
         * machine, and a point object with the virtual cursor's final position on the local machine is returned.
         * If the point parameter is an empty point object, the cursor's new location is set to (0,0).
         */
        public Point SendRemoteCursorToRemoteMachine(int cursorIndex, int machineIndex, Point newLocation)
        {
            if (cursorIndex == localIndex || machineIndex == localIndex)
                return Point.Empty;

            if (cursorModule.cursors[cursorIndex].Visible || cursorModule.blockedCursors[cursorIndex].Visible)
            {

                Point prevPos = new Point(cursorModule.cursors[cursorIndex].Bounds.X,
                    cursorModule.cursors[cursorIndex].Bounds.Y);
                if (newLocation == Point.Empty)
                    newLocation = new Point(0, 0);
                socketModule.SendCursorToRemoteMachine(cursorIndex, machineIndex, newLocation.X, newLocation.Y);
                return prevPos;
            }

            return Point.Empty;

        }


        //!  SendRemoteCursorToRemoteMachine
        /*!
         * This is an API method that attempts to send a virtual cursor from the local machine to a remote
         * machine.  If there is no virtual cursor on the local machine with the specified index, or if there 
         * is no client connected at the indicated destination index, or if the specified cursor index is the 
         * local index, or if the specified destination machine index is the local index, an empty point object 
         * is returned.  Otherwise, the virtual cursor is passed to the indicated remote 
         * machine, and a point object with the virtual cursor's final position on the local machine is returned.
         * If the point parameter is an empty point object, the cursor's new location is set to (0,0).
         */
        public Point SendRemoteCursorToRemoteMachine(String cursorName, int machineIndex, Point newLocation)
        {
            int cursorIndex = GetClientIndex(cursorName);

            if (cursorIndex == -1)
                return Point.Empty;

            if (cursorIndex == localIndex || machineIndex == localIndex)
                return Point.Empty;

            if (cursorModule.cursors[cursorIndex].Visible || cursorModule.blockedCursors[cursorIndex].Visible)
            {

                Point prevPos = new Point(cursorModule.cursors[cursorIndex].Bounds.X,
                    cursorModule.cursors[cursorIndex].Bounds.Y);
                if (newLocation == Point.Empty)
                    newLocation = new Point(0, 0);
                socketModule.SendCursorToRemoteMachine(cursorIndex, machineIndex, newLocation.X, newLocation.Y);
                return prevPos;
            }

            return Point.Empty;
        }

        //!  SendRemoteCursorToRemoteMachine
        /*!
         * This is an API method that attempts to send a virtual cursor from the local machine to a remote
         * machine.  If there is no virtual cursor on the local machine with the specified index, or if there 
         * is no client connected at the indicated destination index, or if the specified cursor index is the 
         * local index, or if the specified destination machine index is the local index, an empty point object 
         * is returned.  Otherwise, the virtual cursor is passed to the indicated remote 
         * machine, and a point object with the virtual cursor's final position on the local machine is returned.
         * If the point parameter is an empty point object, the cursor's new location is set to (0,0).
         */
        public Point SendRemoteCursorToRemoteMachine(int cursorIndex, String machineName, Point newLocation)
        {
            int machineIndex = GetClientIndex(machineName);

            if (machineIndex == -1)
                return Point.Empty;

            if (cursorIndex == localIndex || machineIndex == localIndex)
                return Point.Empty;

            if (cursorModule.cursors[cursorIndex].Visible || cursorModule.blockedCursors[cursorIndex].Visible)
            {

                Point prevPos = new Point(cursorModule.cursors[cursorIndex].Bounds.X,
                    cursorModule.cursors[cursorIndex].Bounds.Y);
                if (newLocation == Point.Empty)
                    newLocation = new Point(0, 0);
                socketModule.SendCursorToRemoteMachine(cursorIndex, machineIndex, newLocation.X, newLocation.Y);
                return prevPos;
            }

            return Point.Empty;
        }

        //!  SendRemoteCursorToRemoteMachine
        /*!
         * This is an API method that attempts to send a virtual cursor from the local machine to a remote
         * machine.  If there is no virtual cursor on the local machine with the specified index, or if there 
         * is no client connected at the indicated destination index, or if the specified cursor index is the 
         * local index, or if the specified destination machine index is the local index, an empty point object 
         * is returned.  Otherwise, the virtual cursor is passed to the indicated remote 
         * machine, and a point object with the virtual cursor's final position on the local machine is returned.
         * If the point parameter is an empty point object, the cursor's new location is set to (0,0).
         */
        public Point SendRemoteCursorToRemoteMachine(String cursorName, String machineName, Point newLocation)
        {
            int cursorIndex = GetClientIndex(cursorName);

            int machineIndex = GetClientIndex(machineName);

            if (cursorIndex == -1 || machineIndex == -1)
                return Point.Empty;

            if (cursorIndex == localIndex || machineIndex == localIndex)
                return Point.Empty;

            if (cursorModule.cursors[cursorIndex].Visible || cursorModule.blockedCursors[cursorIndex].Visible)
            {

                Point prevPos = new Point(cursorModule.cursors[cursorIndex].Bounds.X,
                    cursorModule.cursors[cursorIndex].Bounds.Y);
                if (newLocation == Point.Empty)
                    newLocation = new Point(0, 0);
                socketModule.SendCursorToRemoteMachine(cursorIndex, machineIndex, newLocation.X, newLocation.Y);
                return prevPos;
            }

            return Point.Empty;
        }

        //!  SendRemoteCursorToOwnMachine
        /*!
         * This is an API method that attempts to send a virtual cursor from the local machine "back home"
         * to its local machine.  If the virtual cursor at the specified index is not currently remoted to the
         * local machine, or if the client at the specified index is not currently connected to the CuRT system,
         * an empty point object is returned. Otherwise, the virtual cursor is sent back to its own machine, 
         * and a point object with the virtual cursor's final position on the local machine is returned.  
         * If the point parameter is an empty point object, the cursor's new location is set to (0,0).
         */
        public Point SendRemoteCursorToOwnMachine(int cursorIndex, Point newLocation)
        {
            return SendRemoteCursorToRemoteMachine(cursorIndex, cursorIndex, newLocation);
        }

        //!  SendRemoteCursorToOwnMachine
        /*!
         * This is an API method that attempts to send a virtual cursor from the local machine "back home"
         * to its local machine.  If the virtual cursor with the specified username is not currently remoted to the
         * local machine, or if no client with the specified username exists in the CuRT system,
         * an empty point object is returned. Otherwise, the virtual cursor is sent back to its own machine, 
         * and a point object with the virtual cursor's final position on the local machine is returned.  
         * If the point parameter is an empty point object, the cursor's new location is set to (0,0).
         */
        public Point SendRemoteCursorToOwnMachine(String cursorName, Point newLocation)
        {
            return SendRemoteCursorToRemoteMachine(cursorName, cursorName, newLocation);
        }

        //!  CursorOnEdgeListener
        /*!
         * This is a method that is called whenever the edge detection event is fired.  This method currently just
         * displays all pertinent information to the console window, but it can be redefined to take more
         * interesting actions.
         */
        public class CursorOnEdgeListener
        {
            public void AnnounceCursorOnEdge(CursorOnEdgeEventArgs e)
            {
                try
                {
                    Console.WriteLine("CURSOR ON EDGE!");
                    Console.WriteLine("index: " + e.cursorIndex);
                    Console.WriteLine("location: " + e.cursorLocation.X + ", " + e.cursorLocation.Y);
                    Console.WriteLine("name: " + e.cursorName);
                    Console.WriteLine("direction: " + e.edge);
                    Console.WriteLine("source: " + e.source);
                    Console.WriteLine("screen: " + e.screen);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message.ToString());
                }
            }

        }

        //!  CuRTClient Main method
        /*!
         * This is the CuRTClient Main method.  It simply starts the CuRTClient in standalone application mode.
         * To use the CuRTClient in the API mode, do not execute this Main method.  Rather, create an instance
         * of a CuRTClient object and pass it the appropriate parameters.
         */
        public static void Main()
        {

            CuRTClient newClient;
            if (!AnotherInstanceRunning())
            {
                newClient = new CuRTClient();
            }

        }
    }
}
