//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//                      Cursor Redirection Toolkit (CuRT)                       //
//               Written by Andy McGovern, Orchid Research Group                //
//                           andymcgovern@gmail.com                             //
//                          http://orchid.cs.uiuc.edu                           //
//  Department of Computer Science, University of Illinois at Urbana-Champaign  //
//                                July 4, 2006                                  //
//      Copyright 2006, The Board of Trustees of the University of Illinois     //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Net;
using System.Text;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using System.Drawing;
using System.Reflection;
using System.Threading;

namespace UIUC.DCS.ORCHID.CuRTClientLibrary
{


    //! CursorOnEdgeHandler delegate
    /*!
     *  This is the delegate used for both edge detection systems (Speed Detection and Loiter Detection)
     *  in an attempt to determine when the user is intending to send a cursor to another machine by moving it
     *  to the edge of the screen.
     */
    public delegate void CursorOnEdgeHandler(CursorOnEdgeEventArgs e);

    //! CursorOnEdgeEventArgs
    /*!
     *  This class encapsulates the data members that are set when an edge detection event is fired.  This information
     *  is returned to the member that receives the event so that an appropriate action (such as moving the cursor
     *  to another screen) can be taken.
     */
    public class CursorOnEdgeEventArgs : EventArgs
    {
        //! the index (0-9) of the machine that is controlling the cursor
        public readonly int cursorIndex;

        //! the user-specified name of the machine that is controlling the cursor
        public readonly String cursorName;

        //! a point that contains the coordinates (X,Y) of the cursor at the moment when the event is fired
        public readonly Point cursorLocation;

        //! an enumerated ScreenEdges object representing the particular edge (top, bottom, left, or right) where the edge detection occurred  
        public readonly ScreenEdges edge;

        //! a Screen object representing the particular screen where the edge detection occurred (for use with multi-screen setups)
        public readonly Screen screen;

        //! an integer representing the type of edge detection that occurred (0 = Speed Detection; 1 = Loiter Detection)
        public readonly int source;

        //! CursorOnEdgeEventArgs constructor
        /*!
         *  This constructor just takes the abovementioned six values as parameters and assigns their values
         *  to the appropriate member variables.
         */
        public CursorOnEdgeEventArgs(int cursorIndex, String cursorName,
            Point cursorLocation, ScreenEdges edge, Screen screen, int source)
        {
            this.cursorIndex = cursorIndex;
            this.cursorName = cursorName;
            this.cursorLocation = cursorLocation;
            this.edge = edge;
            this.screen = screen;
            this.source = source;
        }

    }

    //! CursorModule class
    /*!
     *  This class manages the functionality of the following components of the CuRTClient: local cursor/keyboard event
     *  listening and blocking/unblocking; virtual cursor display, movement, and blocking/unblocking; and local/virtual 
     *  cursor edge detection.
     */
    public class CursorModule
    {
        //!  ScreenBounds
        /*!
         * The ScreenBounds class is used to store the screen bounds of one individual screen (of potentially
         * many) on the local machine.
         */
        public class ScreenBounds
        {
            public int xMin;
            public int xMax;
            public int yMin;
            public int yMax;
        }

        //! array of virtual cursors
        /*!
         *  This is an array of VirtualCursors, which are simply Windows forms that encapsulate some extra data members.
         *  There are 10 total VirtualCursors in this array, and all are fully initialized when the CuRTClient starts.
         *  The index numbers in this array (0-9) represent the index numbers of the CuRTClients that are currently connected
         *  to the system.  This array is actually comprised of subclasses of the VirtualCursor object, where each subclass 
         *  represents a different possible color that a cursor can have.  These cursors are set to invisible by default,
         *  and they are only made visible when a client decides to send its cursor remotely to this machine.  At that point,
         *  the corresponding cursor on this machine is made visible.
         */
        protected internal VirtualCursor[] cursors;

        //! array of blocked virtual cursors
        /*!
         *  This array corresponds to the array mentioned above.  The only difference is that this array contains the
         *  "blocked" representations of the same cursors--that is, these cursors' images have an X through them to
         *  represent that the cursor is blocked.  These cursors are all initialized when the CuRTClient starts, but they
         *  are invisible by default.  When a cursor needs to be represented as blocked, the corresponding cursor in this
         *  array is set to visible, and then it is set back to invisible when the cursor is to be unblocked.
         */
        protected internal VirtualCursor[] blockedCursors;

        //! local blocked cursor
        /*!
         *  This is one single VirtualCursor that represents the local machine's cursor.  When the local machine's cursor
         *  needs to be blocked, this VirtualCursor is set to visible, and hten when the local machine's cursor needs to 
         *  be unblocked, this VirtualCursor is set to invisible.
         */
        private VirtualCursor localBlockedCursor;

        //! edge detection parameters
        /*!
         *  This object encapsulates the data members needed for the two edge detection systems, Speed Detection and
         *  Loiter Detection.  In the case of the CuRTClient standalone application, these two systems are disabled, as
         *  there is no way to listen for an event from a standalone application.  In the case of the CuRTClient API,
         *  however, the edgeDetectionParams object is passed to the CursorModule constructor, and the parameter object
         *  is copied to this member variable.  The parameters are then used throughout the execution of the CuRTClient
         *  to perform appropriate edge detection.
         */
        private EdgeDetectionParams edgeDetectionParams;

        //! a reference to the CuRTClient object
        /*!
         *  This is simply a reference to the CuRTClient object that is occasionally used by this CursorModule
         *  to refer back to the client object.
         */
        private CuRTClient client;

        //! cursorMovementResolution
        /*!
         *  At times, the CuRT system can experience lag when several cursors are being remoted to the same machine
         *  simultaneously.  To help allevaite this problem, the cursorMovementResolution parameter can be adjusted.
         *  When a cursor is being remoted, this parameter represents the number of cursor movements that are ignored 
         *  on the local machine before a movement is sent to the remote machine.  For example, a value of 1 means 
         *  that all cursor movements are sent from the local machine to the remot machine, and a value of 3 means
         *  that only every third cursor movement is sent from the local machine to the remote machine.  Note that 
         *  coordinates are sent relatively, not absolutely, but when the cursorMovementResolution parameter has 
         *  a value greater than 1, the values that are sent represent the movement since the last time
         *  that the coordiantes were sent.  Also note that in the standalone version of CuRT system, this parameter
         *  has a default value of 3.
         */
        protected internal int cursorMovementResolution;

        //! currCursorMovementCount
        /*!
         *  This value is used in conjunection with the cursorMovementResolution value.  This value represents the
         *  total number of cursor movements that have occurred.  When this value equals the cursorMovementResolution,
         *  this value is reset to zero, and the cursor movement is sent to the appropriate client machine.
         */
        protected internal int currCursorMovementCount;

        /*! This is an array of ScreenBounds for all screens in use on the local machine, for use in determining
         * whether the cursor is within the greater boundaries of the machine's multiple screens.
         */
        protected internal ScreenBounds[] screenBounds;

        //! localCursorRemoteIndex
        /*!
         *  This integer represents the index of the client machine where the local machine's cursor is currently
         *  being remoted.  If the local machine's cursor is not currently being remoted, this parameter has a value
         *  of -1 (this is also its default value).
         */
        protected internal int localCursorRemoteIndex;

        //! localX: this represents the local machine cursor's X coordinate value 
        private int localX;

        //! localY: this represents the local machine cursor's Y coordinate value 
        private int localY;

        //! screenWidth: this represents the width (in pixels) of the local machine's primary screen
        private int screenWidth;

        //! screenWidth: this represents the height (in pixels) of the local machine's primary screen
        private int screenHeight;

        //! cursor color scheme
        /*!
         *  This is a 10-item array of enumerated CursorColors.  Each index cooresponds to a potential client connected
         *  to the system, and this array is used to determine the color of the cursors in the system, from the
         *  perspective of this local system.
         */
        private CursorColors[] colorScheme;

        //! localCursorIsRemote
        /*!
         *  This is a boolean that represents whether the local machine's cursor is currently being remoted to another machine
         */
        public bool localCursorIsRemote;

        //! dragFromLocal
        /*!
         *  This is a boolean that represents whether the local machine's cursor is currently enacting a drag
         */
        public bool dragFromLocal;

        //! dragFromRemote
        /*!
         *  This is a boolean that represents whether a virtual cursor on the local machine is currently enacting a drag
         */
        public bool dragFromVirtualCursor;

        //! draggingCursorIndex
        /*!
         *  This is an integer that represents the client index of the machine that is currently enacting a drag (-1 if none)
         */
        public int draggingCursorIndex;

        //! cursor on edge event handler
        /*!
         *  This is the event handler used by both edge detection systems to determine when a cursor is on the edge of the screen.
         */
        public static event CursorOnEdgeHandler CursorOnEdge;

        //! flag that determines whether Speed Detection is active
        protected internal bool speedDetectionActive;

        //! flag that determines whether Loiter Detection is active
        protected internal bool loiterDetectionActive;

        #region Mouse Event Enumerations

        //! mouse event enumerations
        /*!
         *  These are mouse event enumerations used for both mouse event listening and mouse event simulation.
         *  Only MOUSEEVENTF_LEFTDOWN, MOUSEEVENTF_LEFTUP, MOUSEEVENTF_RIGHTDOWN, and MOUSEEVENTF_RIGHTUP are used in
         *  this class, but the rest are left as reference for possible additions to this code.  Note that these enumerations
         *  are also defined in the UserActivityHook class, and more of them are used there.
         */

        //! mouse move
        const int MOUSEEVENTF_MOVE = 0x0001;
        //! left button down
        const int MOUSEEVENTF_LEFTDOWN = 0x0002;
        //! left button up
        const int MOUSEEVENTF_LEFTUP = 0x0004;
        //! right button down
        const int MOUSEEVENTF_RIGHTDOWN = 0x0008;
        //! right button up
        const int MOUSEEVENTF_RIGHTUP = 0x0010;
        //! middle button down
        const int MOUSEEVENTF_MIDDLEDOWN = 0x0020;
        //! middle button up
        const int MOUSEEVENTF_MIDDLEUP = 0x0040;
        //! x button down
        const int MOUSEEVENTF_XDOWN = 0x0080;
        //! x button up
        const int MOUSEEVENTF_XUP = 0x0100;
        //! wheel button rolled
        const int MOUSEEVENTF_WHEEL = 0x0800;
        //! map to entire virtual desktop
        const int MOUSEEVENTF_VIRTUALDESK = 0x4000;
        //! absolute move
        const int MOUSEEVENTF_ABSOLUTE = 0x8000;

        #endregion

        #region Imported DLL functions

        //! GetCursorPos imported DLL function
        /*!
         *  This returns a Point that represents the local machine's current cursor position on the screen.
         */
        [DllImport("user32.dll")]
        protected internal static extern bool GetCursorPos(ref Point lpPoint);

        //! SetCursorPos imported DLL function
        /*!
         *  This function provides the ability to send the local machine's cursor to a specific point on the screen.
         */
        [DllImport("user32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool SetCursorPos(int X, int Y);

        //! mouse_event imported DLL function
        /*!
         *  This function provides the ability to simulate mouse actions such as clicks.
         */
        [DllImport("user32.dll")]
        private static extern void mouse_event(
            int dwFlags, // motion and click options
            int dx, // horizontal position or change
            int dy, // vertical position or change
            int dwData, // wheel movement
            int dwExtraInfo // application-defined information
            );

        //! BlockInput imported DLL function
        /*!
         *  This function provides the ability to temporarily block all cursor and keyboard input on the local machine.
         */
        [DllImport("user32.dll")]
        static extern bool BlockInput(bool fBlockIt);

        #endregion

        //! CursorModule constructor--standalone application
        /*!
         *  This is one of two CursorModule constructors.  This constructor is called when the CuRT system is to be used
         *  as a standalone application.  It sets the color scheme and the edge detection to default values, and it also
         *  sets some other initial values to be used by the CursorModule.
         */
        protected internal CursorModule(CuRTClient client)
        {
            //the color scheme for virtual cursors is set to default values
            colorScheme = new CursorColors[10];
            colorScheme[0] = CursorColors.Green;
            colorScheme[1] = CursorColors.Blue;
            colorScheme[2] = CursorColors.Red;
            colorScheme[3] = CursorColors.Yellow;
            colorScheme[4] = CursorColors.Magenta;
            colorScheme[5] = CursorColors.Gray;
            colorScheme[6] = CursorColors.Purple;
            colorScheme[7] = CursorColors.Aqua;
            colorScheme[8] = CursorColors.Brown;
            colorScheme[9] = CursorColors.White;

            screenWidth = Screen.PrimaryScreen.Bounds.Width;
            screenHeight = Screen.PrimaryScreen.Bounds.Height;

            //both edge detection systems are deactived in the standalone application version of the system
            speedDetectionActive = false;
            loiterDetectionActive = false;

            //cursorMovementResolution is set to 3 by default in the standalone application
            cursorMovementResolution = 3;

            //normal default values are set for several member variables
            currCursorMovementCount = 0;
            this.client = client;
            localCursorIsRemote = false;
            dragFromLocal = false;
            dragFromVirtualCursor = false;
            draggingCursorIndex = -1;
            localCursorRemoteIndex = -1;

            Screen[] allScreens = Screen.AllScreens;
            screenBounds = new ScreenBounds[allScreens.Length];
            Rectangle bounds;

            for (int i = 0; i < screenBounds.Length; i++)
            {
                screenBounds[i] = new ScreenBounds();
                bounds = allScreens[i].Bounds;
                screenBounds[i].xMin = bounds.X;
                screenBounds[i].xMax = screenBounds[i].xMin + bounds.Width;
                screenBounds[i].yMin = bounds.Y;
                screenBounds[i].yMax = screenBounds[i].yMin + bounds.Height;
            }

        }

        //! CursorModule constructor--API
        /*!
         *  This is one of two CursorModule constructors.  This constructor is called when the CuRT system is to be used
         *  as an API within a larger application.  If the newColorScheme parameter is null, the color scheme is set to its
         *  default value.  Otherwise, the color scheme is set to the newColorScheme parameter.  If the newEdgeDetectionParams
         *  parameter is set to null, the edge detection system is disabled.  Otherwise, the newEdgeDetectionParams member variable
         *  values are assigned to the EdgeDetectionParams member variable in the CursorModule object.
         *  
         */
        protected internal CursorModule(CuRTClient client, CursorColors[] newColorScheme, int newCursorMovementResolution,
            EdgeDetectionParams newEdgeDetectionParams)
        {
            //The color scheme for virtual cursors is set to the newColorScheme parameter, unless newColorScheme is null.
            //If it is null, then the color scheme is set to its default values.
            if (newColorScheme != null)
                colorScheme = newColorScheme;
            else
            {
                colorScheme = new CursorColors[10];
                colorScheme[0] = CursorColors.Green;
                colorScheme[1] = CursorColors.Blue;
                colorScheme[2] = CursorColors.Red;
                colorScheme[3] = CursorColors.Yellow;
                colorScheme[4] = CursorColors.Magenta;
                colorScheme[5] = CursorColors.Gray;
                colorScheme[6] = CursorColors.Purple;
                colorScheme[7] = CursorColors.Aqua;
                colorScheme[8] = CursorColors.Brown;
                colorScheme[9] = CursorColors.White;
            }

            screenWidth = Screen.PrimaryScreen.Bounds.Width;
            screenHeight = Screen.PrimaryScreen.Bounds.Height;

            //if the newEdgeDetectionParams parameter is null, this means that the edge detection system 
            //should be disabled
            if (newEdgeDetectionParams == null)
            {
                speedDetectionActive = false;
                loiterDetectionActive = false;
            }
            else
                edgeDetectionParams = newEdgeDetectionParams;

            //if any of the speed detection parameter values is less than 1, this means that the speed 
            //detection system should be disabled
            if (edgeDetectionParams.miniJumpsNeeded < 1 ||
                edgeDetectionParams.miniJumpSize < 1 ||
                edgeDetectionParams.zoneSize < 1)
                speedDetectionActive = false;
            else
                speedDetectionActive = true;

            //if the loiter detection parameter value is less than 1, this means that the loiter 
            //detection system should be disabled
            if (edgeDetectionParams.loiterMovementsNeeded < 1)
                loiterDetectionActive = false;
            else
                loiterDetectionActive = true;

            //the cursorMovementResolution variable in this class is set to the value of the parameter passed into the
            //constructor
            cursorMovementResolution = newCursorMovementResolution;

            //normal default values are set for several member variables
            currCursorMovementCount = 0;
            this.client = client;
            localCursorIsRemote = false;
            dragFromLocal = false;
            dragFromVirtualCursor = false;
            draggingCursorIndex = -1;
            localCursorRemoteIndex = -1;

            Screen[] allScreens = Screen.AllScreens;
            screenBounds = new ScreenBounds[allScreens.Length];
            Rectangle bounds;

            for (int i = 0; i < screenBounds.Length; i++)
            {
                screenBounds[i] = new ScreenBounds();
                bounds = allScreens[i].Bounds;
                screenBounds[i].xMin = bounds.X;
                screenBounds[i].xMax = screenBounds[i].xMin + bounds.Width;
                screenBounds[i].yMin = bounds.Y;
                screenBounds[i].yMax = screenBounds[i].yMin + bounds.Height;
            }

        }

        //! OnCursorOnEdgeEvent
        /*!
         *  This method calls the CursorOnEdge event.
         */
        public static void OnCursorOnEdgeEvent(CursorOnEdgeEventArgs e)
        {
            if (CursorOnEdge != null)
                CursorOnEdge(e);
        }

        //! InitializeCursors
        /*!
         * This method is called once to create all the VirtualCursors, both normal and blocked, to fill the two arrays
         * of cursors.  It also initializes the local blocked cursor.
         */
        protected internal void InitializeCursors()
        {

            cursors = new VirtualCursor[10];
            blockedCursors = new VirtualCursor[10];

            //For each index 0-9, a cursor and a blocked cursor are created.  The switch statement determines
            //the correct color for each cursor/blocked cursor pair.
            for (int i = 0; i < 10; i++)
            {
                switch (colorScheme[i])
                {
                    case CursorColors.Green:
                        cursors[i] = new VirtualCursorGreen();
                        blockedCursors[i] = new VirtualCursorGreenBlocked();
                        break;
                    case CursorColors.Blue:
                        cursors[i] = new VirtualCursorBlue();
                        blockedCursors[i] = new VirtualCursorBlueBlocked();
                        break;
                    case CursorColors.Red:
                        cursors[i] = new VirtualCursorRed();
                        blockedCursors[i] = new VirtualCursorRedBlocked();
                        break;
                    case CursorColors.Yellow:
                        cursors[i] = new VirtualCursorYellow();
                        blockedCursors[i] = new VirtualCursorYellowBlocked();
                        break;
                    case CursorColors.Magenta:
                        cursors[i] = new VirtualCursorMagenta();
                        blockedCursors[i] = new VirtualCursorMagentaBlocked();
                        break;
                    case CursorColors.Gray:
                        cursors[i] = new VirtualCursorGray();
                        blockedCursors[i] = new VirtualCursorGrayBlocked();
                        break;
                    case CursorColors.Purple:
                        cursors[i] = new VirtualCursorPurple();
                        blockedCursors[i] = new VirtualCursorPurpleBlocked();
                        break;
                    case CursorColors.Aqua:
                        cursors[i] = new VirtualCursorAqua();
                        blockedCursors[i] = new VirtualCursorAquaBlocked();
                        break;
                    case CursorColors.Brown:
                        cursors[i] = new VirtualCursorBrown();
                        blockedCursors[i] = new VirtualCursorBrownBlocked();
                        break;
                    case CursorColors.White:
                        cursors[i] = new VirtualCursorWhite();
                        blockedCursors[i] = new VirtualCursorWhiteBlocked();
                        break;
                    default:
                        cursors[i] = new VirtualCursorWhite();
                        blockedCursors[i] = new VirtualCursorWhiteBlocked();
                        break;

                }

                //each cursor and blocked cursor is initially set to be topmost and hidden 
                cursors[i].TopMost = true;
                cursors[i].Hide();
                blockedCursors[i].TopMost = true;
                blockedCursors[i].Hide();
                blockedCursors[i].SetDesktopLocation(100, 100);

            }

            //the local blocked cursor is also initialized
            localBlockedCursor = new VirtualCursorWhiteBlocked();
            localBlockedCursor.TopMost = true;
            localBlockedCursor.Visible = false;

        }

        //! MoveLocalCursor
        /*!
         *  This method is called to relocate the local machine's cursor to another point on the screen.
         */
        protected internal void MoveLocalCursor(int x, int y)
        {
            SetCursorPos(x, y);
        }

        //!  LocalXInBounds
        /*!
         * If the xOffset parameter were applied to the current local cursor position, and the local cursor
         * were still in the bounds of the multiple screens of the local machine, this method would return true.
         * Otherwise, it would return false.
         */
        public bool LocalXInBounds(int xOffset)
        {
            Point currPos = new Point();
            CursorModule.GetCursorPos(ref currPos);

            Rectangle bounds = Screen.GetBounds(currPos);

            int newX = currPos.X + xOffset;

            int currY = currPos.Y;

            if (newX < bounds.X || newX > bounds.Width || currY < bounds.Y || currY > bounds.Height)
            {
                bool returnVal = false;

                for (int i = 0; i < screenBounds.Length; i++)
                {
                    if (newX >= screenBounds[i].xMin &&
                        newX <= screenBounds[i].xMax &&
                        currY >= screenBounds[i].yMin &&
                        currY <= screenBounds[i].yMax)
                        returnVal = true;
                }
                return returnVal;
            }

            return true;
        }

        //!  LocalYInBounds
        /*!
         * If the yOffset parameter were applied to the current local cursor position, and the local cursor
         * were still in the bounds of the multiple screens of the local machine, this method would return true.
         * Otherwise, it would return false.
         */
        public bool LocalYInBounds(int yOffset)
        {
            Point currPos = new Point();
            CursorModule.GetCursorPos(ref currPos);

            Rectangle bounds = Screen.GetBounds(currPos);

            int newY = currPos.Y + yOffset;
            int currX = currPos.X;

            if (newY < bounds.Y || newY > bounds.Height || currX < bounds.X || currX > bounds.Width)
            {
                bool returnVal = false;

                for (int i = 0; i < screenBounds.Length; i++)
                {

                    if (newY >= screenBounds[i].yMin &&
                        newY <= screenBounds[i].yMax &&
                        currX >= screenBounds[i].xMin &&
                        currX <= screenBounds[i].xMax)
                        returnVal = true;
                }
                return returnVal;
            }

            return true;
        }

        //!  XInBounds
        /*!
         * If the xOffset parameter were applied to the current VirtualCursor position, and the VirtualCursor
         * were still the bounds of the multiple screens of the local machine, this method would return true.
         * Otherwise, it would return false.
         */
        public bool VirtualXInBounds(int index, int xOffset)
        {
            int newX = cursors[index].Location.X + xOffset;
            int currY = cursors[index].Location.Y;
            Rectangle bounds = Screen.GetBounds(cursors[index]);


            if (newX < bounds.X || newX > bounds.Width || currY < bounds.Y || currY > bounds.Height)
            {
                bool returnVal = false;

                for (int i = 0; i < screenBounds.Length; i++)
                {
                    if (newX >= screenBounds[i].xMin && newX <= screenBounds[i].xMax &&
                        currY >= screenBounds[i].yMin && currY <= screenBounds[i].yMax)
                        returnVal = true;
                }
                return returnVal;
            }

            return true;
        }

        //!  YInBounds
        /*!
         * If the yOffset parameter were applied to the current VirtualCursor position, and the VirtualCursor
         * were still the bounds of the multiple screens of the local machine, this method would return true.
         * Otherwise, it would return false.
         */
        public bool VirtualYInBounds(int index, int yOffset)
        {
            int newY = cursors[index].Location.Y + yOffset;
            int currX = cursors[index].Location.X;
            Rectangle bounds = Screen.GetBounds(cursors[index]);

            if (newY < bounds.Y || newY > bounds.Height || currX < bounds.X || currX > bounds.Width)
            {
                bool returnVal = false;

                for (int i = 0; i < screenBounds.Length; i++)
                {
                    if (newY >= screenBounds[i].yMin && newY <= screenBounds[i].yMax &&
                        currX >= screenBounds[i].xMin && currX <= screenBounds[i].xMax)
                        returnVal = true;
                }
                return returnVal;
            }
            return true;
        }

        //! CheckEdges
        /*!
         *  This method is called anytime a cursor (virtual or actual) moves, and it is used by the edge detection
         *  system to determine when the user is trying to move the cursor off the edge of the screen.
         */
        protected internal void CheckEdges(int index, int x, int y)
        {
            bool cursorIsLocal = false;

            bool checkLoiter = false;
            bool checkSpeed = false;

            //first, check to make sure that the index number is valid
            if (index != -1)
            {
                //If the cursor in question is the local cursor, then the methods to check whether it is in bounds
                //are different from the methods to check whether a virtual cursor is in bounds.  There is parallel
                //code below for the local cursor and for a virtual cursor.
                if (index == client.localIndex)
                    cursorIsLocal = true;

                if (cursorIsLocal)
                {
                    //this test will be false if loiter detection is currently turned off for this client
                    if (loiterDetectionActive)
                    {
                        //If a movement of two pixels to the left would take the cursor out of bounds, then
                        //the cursor must be loitering on the left side of the screen.  Increment the number
                        //of loiter movements to the left for this particular cursor.
                        if (!LocalXInBounds(-2))
                        {
                            cursors[index].numLoiterMovementsLeft++;
                            checkLoiter = true;
                        }
                        //Otherwise, reset the loiter count to zero.
                        else
                            cursors[index].numLoiterMovementsLeft = 0;

                        //parallel to the code above
                        if (!LocalXInBounds(2))
                        {
                            cursors[index].numLoiterMovementsRight++;
                            checkLoiter = true;
                        }
                        else
                            cursors[index].numLoiterMovementsRight = 0;

                        //parallel to the code above
                        if (!LocalYInBounds(-2))
                        {
                            cursors[index].numLoiterMovementsUp++;
                            checkLoiter = true;
                        }
                        else
                            cursors[index].numLoiterMovementsUp = 0;

                        //parallel to the code above
                        if (!LocalYInBounds(2))
                        {
                            cursors[index].numLoiterMovementsDown++;
                            checkLoiter = true;
                        }
                        else
                            cursors[index].numLoiterMovementsDown = 0;
                    }

                    //this test will be false if speed detection is currently turned off for this client
                    if (speedDetectionActive)
                    {
                        //Check to see if the X coordinate is within the left zone. 
                        if (!LocalXInBounds(-edgeDetectionParams.zoneSize))
                        {
                            //Check to see if the X coordinate change from the last point to the current point
                            //is enough for a miniJump.  If so, increment the number of miniJumpsLeft for the current
                            //cursor.  Otherwise, reset that same count to 0. 
                            if ((cursors[index].prevX - x) >= edgeDetectionParams.miniJumpSize)
                            {
                                cursors[index].numMiniJumpsLeft++;
                                checkSpeed = true;
                            }
                            else
                                cursors[index].numMiniJumpsLeft = 0;


                        }

                        //parallel to the code above
                        if (!LocalXInBounds(edgeDetectionParams.zoneSize))
                        {
                            if ((x - cursors[index].prevX) >= edgeDetectionParams.miniJumpSize)
                            {
                                cursors[index].numMiniJumpsRight++;
                                checkSpeed = true;
                            }
                            else
                                cursors[index].numMiniJumpsRight = 0;

                        }

                        //parallel to the code above
                        if (!LocalYInBounds(-edgeDetectionParams.zoneSize))
                        {
                            if ((cursors[index].prevY - y) >= edgeDetectionParams.miniJumpSize)
                            {
                                cursors[index].numMiniJumpsUp++;
                                checkSpeed = true;
                            }
                            else
                                cursors[index].numMiniJumpsUp = 0;

                        }

                        //parallel to the code above
                        if (!LocalYInBounds(edgeDetectionParams.zoneSize))
                        {
                            if ((y - cursors[index].prevY) >= edgeDetectionParams.miniJumpSize)
                            {
                                cursors[index].numMiniJumpsDown++;
                                checkSpeed = true;
                            }
                            else
                                cursors[index].numMiniJumpsDown = 0;

                        }
                    }


                }

                //This code in this else statement exactly parallels the code above, except that the code
                //below is for the case when the cursor in question is virtual, not local.  Please refer to
                //the comments above.
                else
                {
                    if (loiterDetectionActive)
                    {
                        if (!VirtualXInBounds(index, -2))
                        {
                            cursors[index].numLoiterMovementsLeft++;
                            checkLoiter = true;
                        }
                        else
                            cursors[index].numLoiterMovementsLeft = 0;

                        if (!VirtualXInBounds(index, 2))
                        {
                            cursors[index].numLoiterMovementsRight++;
                            checkLoiter = true;
                        }
                        else
                            cursors[index].numLoiterMovementsRight = 0;

                        if (!VirtualYInBounds(index, -2))
                        {
                            cursors[index].numLoiterMovementsUp++;
                            checkLoiter = true;
                        }
                        else
                            cursors[index].numLoiterMovementsUp = 0;

                        if (!VirtualYInBounds(index, 10))
                        {
                            cursors[index].numLoiterMovementsDown++;
                            checkLoiter = true;
                        }
                        else
                            cursors[index].numLoiterMovementsDown = 0;

                    }

                    if (speedDetectionActive)
                    {
                        if (!VirtualXInBounds(index, -edgeDetectionParams.zoneSize))
                        {
                            if ((cursors[index].prevX - x) >= edgeDetectionParams.miniJumpSize)
                            {
                                cursors[index].numMiniJumpsLeft++;
                                checkSpeed = true;
                            }
                            else
                                cursors[index].numMiniJumpsLeft = 0;

                        }


                        if (!VirtualXInBounds(index, edgeDetectionParams.zoneSize))
                        {
                            if ((x - cursors[index].prevX) >= edgeDetectionParams.miniJumpSize)
                            {
                                cursors[index].numMiniJumpsRight++;
                                checkSpeed = true;
                            }
                            else
                                cursors[index].numMiniJumpsRight = 0;

                        }

                        if (!VirtualYInBounds(index, -edgeDetectionParams.zoneSize))
                        {
                            if ((cursors[index].prevY - y) >= edgeDetectionParams.miniJumpSize)
                            {
                                cursors[index].numMiniJumpsUp++;
                                checkSpeed = true;
                            }
                            else
                                cursors[index].numMiniJumpsUp = 0;

                        }

                        if (!VirtualYInBounds(index, edgeDetectionParams.zoneSize))
                        {
                            if ((y - cursors[index].prevY) >= edgeDetectionParams.miniJumpSize)
                            {
                                cursors[index].numMiniJumpsDown++;
                                checkSpeed = true;
                            }
                            else
                                cursors[index].numMiniJumpsDown = 0;

                        }
                    }
                }

                //For this test to be true, loiter detection must be activated (as indicated by the loiterDetectionActive
                //flag, and at least one of the tests above must have detected a loiter (as indicated by the checkLoiter 
                //flag).
                if (loiterDetectionActive && checkLoiter)
                {
                    //If any of the cursors has a loiter movements parameter that has met the total loiter movements
                    //needed for Loiter Detection, an event is fired, with the appropriate information about the event
                    //passed as parameters.

                    if (cursors[index].numLoiterMovementsDown >= edgeDetectionParams.loiterMovementsNeeded)
                    {
                        cursors[index].numLoiterMovementsDown = 0;
                        Point currPoint = new Point(x, y);
                        CursorOnEdgeEventArgs e1 = new CursorOnEdgeEventArgs(index, client.clientUsernameList[index],
                            currPoint, ScreenEdges.Bottom, Screen.FromPoint(currPoint), 0);
                        OnCursorOnEdgeEvent(e1);
                    }

                    if (cursors[index].numLoiterMovementsUp >= edgeDetectionParams.loiterMovementsNeeded)
                    {
                        cursors[index].numLoiterMovementsUp = 0;
                        Point currPoint = new Point(x, y);
                        CursorOnEdgeEventArgs e1 = new CursorOnEdgeEventArgs(index, client.clientUsernameList[index],
                            currPoint, ScreenEdges.Top, Screen.FromPoint(currPoint), 0);
                        OnCursorOnEdgeEvent(e1);
                    }

                    if (cursors[index].numLoiterMovementsLeft >= edgeDetectionParams.loiterMovementsNeeded)
                    {
                        cursors[index].numLoiterMovementsLeft = 0;
                        Point currPoint = new Point(x, y);
                        CursorOnEdgeEventArgs e1 = new CursorOnEdgeEventArgs(index, client.clientUsernameList[index],
                            currPoint, ScreenEdges.Left, Screen.FromPoint(currPoint), 0);
                        OnCursorOnEdgeEvent(e1);
                    }

                    if (cursors[index].numLoiterMovementsRight >= edgeDetectionParams.loiterMovementsNeeded)
                    {
                        cursors[index].numLoiterMovementsRight = 0;
                        Point currPoint = new Point(x, y);
                        CursorOnEdgeEventArgs e1 = new CursorOnEdgeEventArgs(index, client.clientUsernameList[index],
                            currPoint, ScreenEdges.Right, Screen.FromPoint(currPoint), 0);
                        OnCursorOnEdgeEvent(e1);
                    }
                }

                //For this test to be true, speed detection must be activated (as indicated by the speedDetectionActive
                //flag, and at least one of the tests above must have detected a miniJump (as indicated by the checkSpeed 
                //flag).
                if (speedDetectionActive && checkSpeed)
                {
                    //Now, check to see if any of the current miniJumps counts (left, right, up, or down) for this 
                    //particular cursor is enough to fire an edge detection event.  If so, fire the event and pass 
                    //all pertinent information as parameters.

                    if (cursors[index].numMiniJumpsLeft >= edgeDetectionParams.miniJumpsNeeded)
                    {
                        cursors[index].numMiniJumpsLeft = 0;
                        Point currPoint = new Point(x, y);
                        CursorOnEdgeEventArgs e1 = new CursorOnEdgeEventArgs(index, client.clientUsernameList[index],
                            currPoint, ScreenEdges.Left, Screen.FromPoint(currPoint), 1);
                        OnCursorOnEdgeEvent(e1);
                    }

                    if (cursors[index].numMiniJumpsRight >= edgeDetectionParams.miniJumpsNeeded)
                    {
                        cursors[index].numMiniJumpsRight = 0;
                        Point currPoint = new Point(x, y);
                        CursorOnEdgeEventArgs e1 = new CursorOnEdgeEventArgs(index, client.clientUsernameList[index],
                            currPoint, ScreenEdges.Right, Screen.FromPoint(currPoint), 1);
                        OnCursorOnEdgeEvent(e1);
                    }

                    if (cursors[index].numMiniJumpsUp >= edgeDetectionParams.miniJumpsNeeded)
                    {
                        cursors[index].numMiniJumpsUp = 0;
                        Point currPoint = new Point(x, y);
                        CursorOnEdgeEventArgs e1 = new CursorOnEdgeEventArgs(index, client.clientUsernameList[index],
                            currPoint, ScreenEdges.Top, Screen.FromPoint(currPoint), 1);
                        OnCursorOnEdgeEvent(e1);
                    }

                    if (cursors[index].numMiniJumpsDown >= edgeDetectionParams.miniJumpsNeeded)
                    {
                        cursors[index].numMiniJumpsDown = 0;
                        Point currPoint = new Point(x, y);
                        CursorOnEdgeEventArgs e1 = new CursorOnEdgeEventArgs(index, client.clientUsernameList[index],
                            currPoint, ScreenEdges.Bottom, Screen.FromPoint(currPoint), 1);
                        OnCursorOnEdgeEvent(e1);
                    }
                }

                cursors[index].prevX = x;
                cursors[index].prevY = y;
            }

        }

        //! RemoveVirtualCursor
        /*!
         *  This method is called from the SocketModule when it receives a message to remove a virtual cursor from
         *  the local machine.  It simply hides the blocked cursor and virtual cursor.
         */
        protected internal void RemoveVirtualCursor(int index)
        {
            cursors[index].Hide();
        }

        //! BlockVirtualCursor
        /*!
         *  This method is called when a single virtual cursor is to be blocked on the local
         *  machine.  The virtual cursor is hidden, and the corresponding blocked cursor is moved to the same location
         *  and is made visible.
         */
        protected internal void BlockVirtualCursor(int index)
        {
            cursors[index].Hide();

            blockedCursors[index].SetDesktopLocation(cursors[index].Location.X, cursors[index].Location.Y);
            blockedCursors[index].Show();

        }

        //! UnblockVirtualCursor
        /*!
         *  This method is called when a single virtual cursor is to be unblocked on the local
         *  machine.  The cursor is unblocked and made visible, and the corresponding blocked cursor is hidden.
         */
        protected internal void UnblockVirtualCursor(int index)
        {
            cursors[index].Show();
            blockedCursors[index].Hide();
        }

        //! BlockAllVirtualCursors
        /*!
         *  This method is called when all virtual cursors are to be blocked on the local
         *  machine.  For each active virtual cursor, the virtual cursor is hidden, and the 
         *  corresponding blocked cursor is moved to the same location and is made visible.
         */
        protected internal void BlockAllVirtualCursors()
        {
            for (int i = 0; i < 10; i++)
            {
                if (cursors[i].Visible)
                {
                    blockedCursors[i].SetDesktopLocation(cursors[i].Location.X, cursors[i].Location.Y);
                    blockedCursors[i].Visible = true;
                    blockedCursors[i].BringToFront();
                }
            }
        }

        //! UnblockAllVirtualCursors
        /*!
         *  This method is called when all virtual cursors are to be unblocked on the local
         *  machine.  For each active virtual cursor, the cursor is unblocked and made visible, 
         *  and the corresponding blocked cursor is hidden.
         */
        protected internal void UnblockAllVirtualCursors()
        {
            for (int i = 0; i < 10; i++)
            {
                if (blockedCursors[i].Visible)
                    blockedCursors[i].Hide();
            }
        }

        //! BlockAllVirtualCursorsExceptOne
        /*!
         *  This method is called when all virtual cursors are to be blocked on the local
         *  machine, except for one.  For each active virtual cursor (except the one whose index is indicated
         *  by the parameter), the virtual cursor is hidden, and the corresponding blocked cursor is moved to the 
         *  same location and is made visible.
         */
        protected internal void BlockAllVirtualCursorsExceptOne(int index)
        {
            for (int i = 0; i < 10; i++)
            {
                if (cursors[i].Visible && i != index)
                {
                    blockedCursors[i].SetDesktopLocation(cursors[i].Location.X, cursors[i].Location.Y);
                    blockedCursors[i].Visible = true;
                    blockedCursors[i].BringToFront();
                }
            }
        }


        //! UnblockAllVirtualCursorsExceptOne
        /*!
         *  This method is called when all virtual cursors are to be unblocked on the local
         *  machine, except for one.  For each active virtual cursor (except the one whose index is indicated
         *  by the parameter), the cursor is unblocked and made visible, and the corresponding blocked cursor is hidden.
         */
        protected internal void UnblockAllVirtualCursorsExceptOne(int index)
        {
            for (int i = 0; i < 10; i++)
            {
                if (blockedCursors[i].Visible && i != index)
                    blockedCursors[i].Hide();
            }
        }


        //! MoveVirtualCursorRelative
        /*!
         *  This method is called when a message has arrived to perform a relative move on a virtual cursor
         *  that is currently being remoted to the local machine.  It is first determined whether the cursor
         *  would remain within the boundaries of the screen if the move were to take place.  If so, and if
         *  no other cursor (local or remote) is currently enacting a drag, then the move takes place.  The
         *  Edge Detection system also checks the move and increments the appopriate counters, if necessary.
         */
        public void MoveVirtualCursorRelative(int index, int xOffset, int yOffset)
        {
            bool xInBounds = VirtualXInBounds(index, xOffset);
            bool yInBounds = VirtualYInBounds(index, yOffset);
            if ((xInBounds || yInBounds) && !dragFromLocal &&
               (!dragFromVirtualCursor || draggingCursorIndex == index))
            {
                int prevX = cursors[index].Location.X;
                int prevY = cursors[index].Location.Y;

                client.cursorModule.cursors[index].Show();
                client.cursorModule.cursors[index].BringToFront();

                int newX = prevX;
                int newY = prevY;

                if (xInBounds)
                    newX += xOffset;

                if (yInBounds)
                    newY += yOffset;

                CheckEdges(index, newX, newY);
                client.cursorModule.cursors[index].SetDesktopLocation(newX, newY);

                if (cursors[index].drag == true)
                    MoveLocalCursor(newX - 1, newY - 1);

            }
        }

        //! MoveVirtualCursorAbsolute
        /*!
         *  This method is called when a message has arrived that a new virtual cursor has arrived on the
         *  local machine.  The virtual cursor is moved to a specific point on the local machine's screen.
         */
        protected internal void MoveVirtualCursorAbsolute(int index, int xCoord, int yCoord)
        {

            cursors[index].prevX = xCoord;
            cursors[index].prevY = yCoord;

            cursors[index].Show();
            cursors[index].BringToFront();

            cursors[index].SetDesktopLocation(xCoord, yCoord);

        }


        //! SimulateLeftClickDown
        /*!
         *  This method is called when a virtual cursor is trying to enact a left click down (and thus a potential drag)
         *  on the local machine.  The local cursor must be "borrowed" for the click/drag, and then returned to its normal
         *  position once the SimulateLeftClickUp method is called.
         */
        protected internal void SimulateLeftClickDown(int index)
        {
            //NOTE: the code for the SimulateRightClickDown method parallels this code exactly, so it will only be
            //commented here.

            //Before enacting a click down that could be a potential drag, first check that there is no virtual
            //cursor or local cursor that is currently dragging.
            if (!dragFromVirtualCursor && !dragFromLocal)
            {
                //a virtual cursor has now begun a drag
                dragFromVirtualCursor = true;

                //set the index of the currently dragging virtual cursor
                draggingCursorIndex = index;

                //all virtual cursors except the one that is currently dragging should be blocked
                BlockAllVirtualCursorsExceptOne(index);

                //set the drag flag within the currently dragging cursor to true
                cursors[index].drag = true;

                //get the local cursor's coordinates and store them to localX and localY
                Point defPnt = new Point();
                GetCursorPos(ref defPnt);
                localX = defPnt.X;
                localY = defPnt.Y;

                //get the currently dragging cursor's coordinates and store them to remoteX and remoteY
                int remoteX = cursors[index].Bounds.X;
                int remoteY = cursors[index].Bounds.Y;

                //check to see if the local cursor is currently being remoted
                if (localCursorIsRemote)
                {
                    //local cursor must be blocked on remote machine where it is currently virtual
                    client.socketModule.BlockVirtualCursor();

                    //click guard on local machine must be deactivated so that the virtual click on the local
                    //machine can take place
                    client.GUI.DeactivateClickGuard();
                }
                //local cursor is currently on local machine
                else
                {
                    //local blocked cursor is moved to current cursor's location and is set to visible
                    localBlockedCursor.SetDesktopLocation(localX, localY);
                    localBlockedCursor.Visible = true;
                    localBlockedCursor.SetDesktopLocation(localX, localY);
                }

                //local cursor movement is blocked to prevent interference with virtual cursor that is enacting the
                //click/drag
                BlockInput(true);

                //the local cursor is moved to the location of the virtual cursor enacting the drag (the location is
                //offset one pixel up and one pixel to the left to avoid clicking directly on top of the virtual cursor)
                SetCursorPos(remoteX - 1, remoteY - 1);

                //the actual left button down mouse event is enacted
                mouse_event(MOUSEEVENTF_LEFTDOWN, 0, 0, 0, 0);
                //Thread.Sleep(10);
            }
        }

        //! SimulateLeftClickUp
        /*!
         *  This method is called when a virtual cursor is trying to enact a left click up on the local machine,
         *  which must follow a left click down that happened previously.  The local cursor that was "borrowed" 
         *  for the click/drag is now returned to its normal position.
         */
        protected internal void SimulateLeftClickUp(int index)
        {
            //NOTE: the code for the SimulateRightClickUp method parallels this code exactly, so it will only be
            //commented here.

            //check to make sure that the virtual cursor that has just released its left mouse button is in
            //fact the same one that has been dragging
            if (draggingCursorIndex == index)
            {
                //the virtual cursor is no longer dragging
                dragFromVirtualCursor = false;

                //reset the draggingCursorIndex
                draggingCursorIndex = -1;

                //set the drag flag for this particular virtual cursor back to false
                cursors[index].drag = false;

                //the actual left button up mouse event is enacted
                mouse_event(MOUSEEVENTF_LEFTUP, 0, 0, 0, 0);
                Thread.Sleep(10);

                //all virtual cursors except the one that is currently finishing its drag should be unblocked
                UnblockAllVirtualCursorsExceptOne(index);

                //check to see if the local cursor is currently being remoted
                if (localCursorIsRemote)
                {
                    //local cursor's virtual representation should be unblocked on the appropriate remote client
                    client.socketModule.UnblockVirtualCursor();

                    //the click guard on the local machine should be reactivated
                    client.GUI.ActivateClickGuard();

                }
                else
                {
                    //if local cursor is currently local, simply hide the local blocked cursor
                    localBlockedCursor.Visible = false;
                }

                //remove block that was preventing local mouse activity
                BlockInput(false);

                //return local cursor to its previous position before the block began
                SetCursorPos(localX, localY);

            }
        }

        //! SimulateRightClickDown
        /*!
         *  This method is called when a virtual cursor is trying to enact a right click down (and thus a potential drag)
         *  on the local machine.  The local cursor must be "borrowed" for the click/drag, and then returned to its normal
         *  position once the SimulateRightClickUp method is called.
         */
        protected internal void SimulateRightClickDown(int index)
        {
            //NOTE: the code for the SimulateLeftClickDown method parallels this code exactly, so please see the
            //comments in that method.

            if (!dragFromVirtualCursor && !dragFromLocal)
            {
                dragFromVirtualCursor = true;
                draggingCursorIndex = index;

                BlockAllVirtualCursorsExceptOne(index);

                cursors[index].drag = true;

                Point defPnt = new Point();
                GetCursorPos(ref defPnt);
                localX = defPnt.X;
                localY = defPnt.Y;


                int remoteX = cursors[index].Bounds.X;
                int remoteY = cursors[index].Bounds.Y;


                if (localCursorIsRemote)
                {
                    client.socketModule.BlockVirtualCursor();
                    client.GUI.DeactivateClickGuard();
                }
                else
                {

                    localBlockedCursor.SetDesktopLocation(localX, localY);
                    localBlockedCursor.Visible = true;
                    localBlockedCursor.SetDesktopLocation(localX, localY);

                }

                BlockInput(true);
                SetCursorPos(remoteX - 1, remoteY - 1);


                Thread.Sleep(10);
                mouse_event(MOUSEEVENTF_RIGHTDOWN, 0, 0, 0, 0);
                Thread.Sleep(10);

            }


        }

        //! SimulateRightClickUp
        /*!
         *  This method is called when a virtual cursor is trying to enact a right click up on the local machine,
         *  which must follow a right click down that happened previously.  The local cursor that was "borrowed" 
         *  for the click/drag is now returned to its normal position.
         */
        protected internal void SimulateRightClickUp(int index)
        {
            //NOTE: the code for the SimulateLeftClickUp method parallels this code exactly, so please see the
            //comments in that method.

            if (draggingCursorIndex == index)
            {
                dragFromVirtualCursor = false;
                draggingCursorIndex = -1;
                cursors[index].drag = false;
                mouse_event(MOUSEEVENTF_RIGHTUP, 0, 0, 0, 0);
                Thread.Sleep(10);

                UnblockAllVirtualCursorsExceptOne(index);

                if (localCursorIsRemote)
                {
                    client.socketModule.UnblockVirtualCursor();
                    client.GUI.ActivateClickGuard();

                }
                else
                {

                    localBlockedCursor.Visible = false;
                }

                BlockInput(false);
                SetCursorPos(localX, localY);

            }

        }

        //! SimulateKeyPress
        /*!
         *  An incoming keypress has arrived from a machine whose input is currently being remoted to the local
         *  machine.  SendKeys actually enacts the keypress on the local machine.
         */
        protected internal void SimulateKeyPress(int index, String key)
        {
            SendKeys.SendWait(key);
        }
    }
}
