//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//                      Cursor Redirection Toolkit (CuRT)                       //
//               Written by Andy McGovern, Orchid Research Group                //
//                           andymcgovern@gmail.com                             //
//                          http://orchid.cs.uiuc.edu                           //
//  Department of Computer Science, University of Illinois at Urbana-Champaign  //
//                                July 4, 2006                                  //
//      Copyright 2006, The Board of Trustees of the University of Illinois     //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

using System;

namespace UIUC.DCS.ORCHID.CuRTClientLibrary
{
    //! This class encapsulates the parameters needed for both edge detection systems, Speed Detection and Loiter Detection
    public class EdgeDetectionParams
    {
        //! miniJumpsNeeded (Speed Detection)
        /*!
         *  This is the number of mini jumps in a row that are needed to trigger an edge detection.
         *  A mini jump occurs when all of the following are true:
         *  1) a change in cursor coordinate is detected
         *  2) the cursor is within one of the zones specified by the zoneSize parameter (see below)
         *  3) the cursor is moving toward the edge of the screen that corresponds to the particular zone
         *  4) the change of position, in pixels, of the cursor's position (the X coordinate in the case of movement
         *      toward the left or right edge of the screen, and the Y coordinate in the case of movement toward the top
         *      or bottom edge of the screen) is greater than or equal to the miniJumpSize parameter
         *
         *  Whenever a cursor movement is detected that does NOT trigger a minijump, the minijump count is reset to 0.
         *  If the miniJump count reaches the miniJumpsNeeded parameter, a Edge Detection event is fired.
         */
        public int miniJumpsNeeded;

        //! zoneSize (Speed Detection)
        /*!
         *  the size (in pixels) of the zone at the top, bottom, left, and right edge 
         *  of the screen where mini jumps can occur
         */
        public int zoneSize;

        //! miniJumpSize (Speed Detection)
        /*!
         *  the minimum size (in pixels) of one mini jump
         */
        public int miniJumpSize;

        //! loiterMovementsNeeded (Loiter Detection)
        /*!
         *  A loiter movement occurs when all of the following are true:
         *  1) either the X or the Y coordinate is one of the edges of the screen (top, bottom, left, or right) 
         *  2) a change in cursor coordinate is detected
         *  3) the coordinate that was on one of the edges of the screen remains on the edge 
         *      (i.e. only the other coordinate changes)
         * 
         *  Whenever a cursor movement is detected that is NOT a loiterMovement, the loiterMovement count is reset to 0.
         *  If the loiterMovement count reaches the loiterMovementsNeeded parameter, a Edge Detection event is fired.
         */
        public int loiterMovementsNeeded;

        //! The EdgeDetectionParams constructor 
        /*!
         *  This constructor just copies values over from the parameters to the object's member variables.
         *  It also makes appropriate adjustments to the values if the user has signified that he wants to
         *  disable one or both of the systems (by passing -1 as the parameter).  This constructor is to be used
         *  by developers using the CuRT system API in order to specify the parameters of the Edge Detection system.
         *  The newly created EdgeDetectionParams object is then passed into the CuRTClient constuctor.
         */
        public EdgeDetectionParams(int miniJumpsNeeded, int zoneSize, int miniJumpSize, int loiterMovementsNeeded)
        {

            this.miniJumpsNeeded = miniJumpsNeeded;
            this.zoneSize = zoneSize;
            this.miniJumpSize = miniJumpSize;
            this.loiterMovementsNeeded = loiterMovementsNeeded;

        }


    }
}
