using System;
using System.IO;
using System.Net;
using System.Text;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using System.Drawing;
using System.Reflection;
using System.Threading;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;


namespace UIUC.DCS.ORCHID.CuRTClientLibrary
{

    //! SocketModule class
    /*!
     *  This class is used for all network communication between the clients that make up the CuRT system.
     */
    public class SocketModule
    {
        //! this is used to connect to the CuRT server
        private TcpClient socketForServer;

        //! this is a reference to the CuRTClient object
        private CuRTClient client;

        //! this is the TCP network stream from which the reader and writer are generated
        private Stream networkStream;

        //! this is the TCP reader
        private System.IO.StreamReader streamReader;

        //! this is the TCP writer
        private System.IO.StreamWriter streamWriter;

        //! this is the server IP in String format
        private String IP;

        //! this is a flag used to ensure that the actions taken when the connection is closed only occur once
        private bool closingConnection;

        //! this the client-side UDP endpoint
        IPEndPoint UDPipep;

        //! this the server-side UDP endpoint
        EndPoint UDPServer;

        //! this the UDP client socket
        Socket UDPClient;

        #region Imported DLL functions

        //! SetCursorPos imported DLL function
        /*!
         *  This function provides the ability to send the local machine's cursor to a specific point on the screen.
         */
        [DllImport("user32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool SetCursorPos(int X, int Y);

        #endregion

        //! SocketModule constructor
        /*!
         *  The SocketModule constructor takies in an IP or host name and sets the class member IP address
         *  to the parameter IP address (in the case of a hostname, a lookup is performed first).  It also sets
         *  the class member client reference to be the same as the parameter.
         */
        public SocketModule(String IP, CuRTClient client)
        {
            try
            {
                this.client = client;
                IPHostEntry ipEntry = Dns.GetHostByName(IP);
                IPAddress[] addr = ipEntry.AddressList;
                this.IP = addr[0].ToString();
                streamReader = null;
                streamWriter = null;
                closingConnection = false;
            }

            catch (Exception ex)
            {
                Console.WriteLine("IP Error: " + ex.Message);
            }
        }

        public void Dispose()
        {
            if (streamReader != null)
            {
                streamReader.Close();
                streamReader.Dispose();
            }

            if (streamWriter != null)
            {
                streamWriter.Close();
                streamWriter.Dispose();
            }
        }

        //! InitializeListeners
        /*!
         *  In this method, the UDP and TCP listener threads are both created and started.
         */
        [STAThread]
        public void InitializeListeners()
        {
            Thread UDPMessageListenerThread = new Thread(new ThreadStart(ListenForUDPMessages));
            UDPMessageListenerThread.Name = "listenForUDPMessages";
            UDPMessageListenerThread.IsBackground = true;
            UDPMessageListenerThread.Start();

            Thread TCPDiconnectListenerThread = new Thread(new ThreadStart(ListenForTCPDisconnect));
            TCPDiconnectListenerThread.Name = "listenForTCPDisconnect";
            TCPDiconnectListenerThread.IsBackground = true;
            TCPDiconnectListenerThread.Start();

            Application.Run();
        }

        //! GetPort
        /*!
         *  This method connects to the server, using the given IP, at port 4415.  It reads a line of text 
         *  from the server's stream.  This line contains the port number that the client
         *  needs to connect back to.  It then returns that port as an integer.
         */
        private int GetPort()
        {
            int newPort = 0;
            try
            {
                socketForServer = new TcpClient();
                socketForServer.Connect(IP, 4415);
                networkStream = socketForServer.GetStream();
                streamReader = new System.IO.StreamReader(networkStream);

                String portAsString = streamReader.ReadLine();

                try { newPort = Convert.ToInt32(portAsString); }
                catch (Exception e)
                {
                    Console.WriteLine("Error..... " + e.StackTrace);
                }

            }

            catch (Exception e)
            {
                Console.WriteLine("Error..... " + e.StackTrace);
            }
            return newPort;
        }

        //! ConnectToServer
        /*!
         *  This method establishes all components of the UDP and TCP connections on the port number that is
         *  returned by the GetPort method.
         */
        public bool ConnectToServer(String username)
        {
            int port = GetPort();

            //this returns false when more than 10 users try to connect to the system
            if (port > 2009)
                return false;

            try
            {
                TcpClient socketForServer = new TcpClient();

                //server IP and new port
                socketForServer.Connect(IP, port);

                networkStream = socketForServer.GetStream();
                streamReader = new System.IO.StreamReader(networkStream);
                streamWriter = new System.IO.StreamWriter(networkStream);

                //the local client's username is sent to the server so that the master list can be updated
                SendUsername(username);

                //initialize UDP communication
                byte[] data = new byte[1024];
                UDPipep = new IPEndPoint(IPAddress.Parse(IP), port + 10);
                UDPClient = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                IPEndPoint UDPRemoteipep = new IPEndPoint(IPAddress.Any, 0);
                UDPServer = (EndPoint)UDPRemoteipep;
                String message = "UDP connect";
                data = Encoding.ASCII.GetBytes(message);
                Thread.Sleep(1000);
                UDPClient.SendTo(data, data.Length, SocketFlags.None, UDPipep);

                return true;
            }

            catch (Exception e)
            {
                Console.WriteLine("Error while connecting" + e);
                return false;
            }


        }

        //! SendUsername
        /*!
         *  This method sends the local client's username to the server so that the master list of usernames
         *  can be updated.
         */
        private void SendUsername(String username)
        {
            streamWriter.WriteLine(username);
            streamWriter.Flush();
        }

        //! SendUDPMessage
        /*!
         *  This method uses the UDP protocol to send a message from the local client to the server.
         */
        protected internal void SendUDPMessage(String message)
        {
            if (UDPClient != null)
            {
                byte[] data = new byte[1024];
                data = Encoding.ASCII.GetBytes(message);
                UDPClient.SendTo(data, data.Length, SocketFlags.None, UDPipep);
            }
        }

        //! SendRelativeCoordinates
        /*!
         *  This method uses the UDP protocol to send a relative coordinate change to a remote machine, through
         *  the server.  Both the X and Y coordinates are stored as strings, and they are formatted such that they
         *  are always four digits long, with leading zeros, if necessary.
         */
        protected internal void SendRelativeCoordinates(int x, int y)
        {
            String xAsString = "", yAsString = "";

            if (x < 0)
            {
                xAsString = "-";
                x = x * -1;
            }
            else
                xAsString = "0";

            if (x < 10)
                xAsString = xAsString + "00" + x;
            else if (x < 100)
                xAsString = xAsString + "0" + x;
            else
                xAsString = xAsString + x;

            if (y < 0)
            {
                yAsString = "-";
                y = y * -1;
            }
            else
                yAsString = "0";

            if (y < 10)
                yAsString = yAsString + "00" + y;
            else if (y < 100)
                yAsString = yAsString + "0" + y;
            else
                yAsString = yAsString + y;

            SendUDPMessage("MR" + xAsString + yAsString);
        }

        //! SendCursorToRemoteMachine
        /*!
         *  This method uses the UDP protocol to send a virtual cursor to a remote machine.  This can mean sending a
         *  cursor "home" to its own machine (meaning that the cursorIndex and the machineIndex are the same), or it
         *  can mean sending a virtual cursor to another machine (this is usually done when a virtual cursor reaches
         *  the edge of the local machine's screen and needs to be rerouted to another machine.  The initial
         *  location of the virtual cursor on the remote machine is also specified.  Both the X and Y coordinates 
         *  are reformatted as strings, and they are formatted such that they are always four digits long, with 
         *  leading zeros, if necessary.
         */
        public void SendCursorToRemoteMachine(int cursorIndex, int machineIndex, int x, int y)
        {
            String xAsString, yAsString;
            if (x < 0)
                xAsString = "0";
            else if (x < 10)
                xAsString = "000" + x;
            else if (x < 100)
                xAsString = "00" + x;
            else if (x < 1000)
                xAsString = "0" + x;
            else
                xAsString = "" + x;

            if (y < 0)
                yAsString = "0";
            else if (y < 10)
                yAsString = "000" + y;
            else if (y < 100)
                yAsString = "00" + y;
            else if (y < 1000)
                yAsString = "0" + y;
            else
                yAsString = "" + y;

            SendUDPMessage("SC" + cursorIndex + machineIndex + xAsString + yAsString);
        }

        //! RemoveLocalCursorFromRemote
        /*!
         *  This method uses the UDP protocol to send a message, through the server, to the client machine where
         *  the local machine's cursor is currently being remoted.  Upon receiving the message, the client removes
         *  the virtual cursor.  This method is called by the CuRTClient class.  Along with calling this method, the
         *  client also takes appropriate actions to re-receive the local cursor on the local machine.
         */
        public bool RemoveLocalCursorFromRemote()
        {
            SendUDPMessage("RC");
            return true;
        }

        //! ListenForTCPDisconnect
        /*!
         *  This method starts a TCP listener.  No TCP messages are sent, but an exception is generated if the
         *  connection is dropped between the client and the server.  This alerts the client that the connection
         *  has been dropped so that the appropriate actions can be taken.
         */
        [STAThread]
        public void ListenForTCPDisconnect()
        {
            try
            {
                while (true)
                {
                    String message = streamReader.ReadLine();
                    Console.WriteLine("LISTEN FOR TCP DISCONNECT");
                }
            }
            catch (Exception e)
            {
                if (!closingConnection)
                {
                    closingConnection = true;
                    client.GUI.DisconnectedFromServer();
                    Console.WriteLine("Connection dropped..." + e);
                }
            }
        }

        //! ListenForUDPMessages
        /*!
         *  This method starts a UDP listener.  All client-to-client communication is done through the UDP listener.
         *  When this method receives a message, the ProcessMessage method is called, and then appropriate action
         *  is taken.
         */
        [STAThread]
        public void ListenForUDPMessages()
        {
            byte[] data = new byte[1024];
            int recv;
            String message;

            try
            {
                while (true)
                {
                    data = new byte[1024];
                    recv = UDPClient.ReceiveFrom(data, ref UDPServer);
                    message = Encoding.ASCII.GetString(data, 0, recv);
                    ProcessMessage(message);

                    Console.WriteLine("LISTEN FOR UDP MESSAGES");

                }
            }
            catch (Exception e)
            {

                Console.WriteLine("UDP listen error!..... " + e.StackTrace);
            }

        }


        //! BlockVirtualCursor
        /*!
         *  This method sends a UDP message to the server when the local cursor needs to be blocked remotely.
         */
        public void BlockVirtualCursor()
        {
            SendUDPMessage("BRC");
        }

        //! UnblockVirtualCursor
        /*!
         *  This method sends a UDP message to the server when the local cursor needs to be unblocked remotely.
         */
        public void UnblockVirtualCursor()
        {
            SendUDPMessage("URC");
        }

        //! ProcessMessage
        /*!
         *  This method receives an incoming message from the server.  It takes appopriate action based on the
         *  contents of the message.
         */
        protected internal void ProcessMessage(String message)
        {
            try
            {
                //message: virtual cursor on local machine should be moved with relative coordinates
                if (message.Substring(0, 2) == "MR")
                {
                    int index = Convert.ToInt16(message.Substring(2, 1));
                    int xOffset = Convert.ToInt16(message.Substring(3, 4));
                    int yOffset = Convert.ToInt16(message.Substring(7, 4));
                    client.cursorModule.MoveVirtualCursorRelative(index, xOffset, yOffset);
                }

                //message: a keypress should be effected on the local machine
                else if (message.Length > 8 && message.Substring(0, 8) == "KEYPRESS")
                {
                    int index = Convert.ToInt16(message.Substring(8, 1));
                    String key = message.Substring(9);
                    //The index from where the keypress originates is passed into this method.
                    //It is currently not used, but it may be useful in the future.
                    client.cursorModule.SimulateKeyPress(index, key); 
                }

                //message: a cursor (possibly the local cursor) is arriving to this local machine
                else if (message.Substring(0, 2) == "SC")
                {

                    int cursorIndex = Convert.ToInt16(message.Substring(2, 1));
                    int x = Convert.ToInt16(message.Substring(3, 4));
                    int y = Convert.ToInt16(message.Substring(7, 4));
                    if (cursorIndex == client.localIndex)
                    {
                        client.cursorModule.localCursorIsRemote = false;
                        client.cursorModule.localCursorRemoteIndex = -1;
                        client.GUI.CleanUpToReleaseLocalCursor();
                        RemoveLocalCursorFromRemote();
                        SetCursorPos(x, y);                        
                    }
                    else
                    {
                        client.cursorModule.MoveVirtualCursorAbsolute(cursorIndex, x, y);
                    }
                }

                //message: a new client user list has arrived
                else if (message.Substring(0, 2) == "UL")
                {
                    String tempMsg = message.Substring(2, message.Length - 2);
                    client.GUI.ProcessClientUsernameList(tempMsg);
                }

                //message: a virtual cursor should be removed from the local machine
                else if (message.Substring(0, 2) == "RC")
                {
                    int index = Convert.ToInt16(message.Substring(2, 1));
                    client.cursorModule.RemoveVirtualCursor(index);
                }

                //message: a virtual cursor on the local machine should be blocked
                else if (message.Substring(0, 3) == "BRC")
                {
                    int index = Convert.ToInt16(message.Substring(3, 1));
                    client.cursorModule.BlockVirtualCursor(index);
                }

                //message: a virtual cursor on the local machine should be unblocked
                else if (message.Substring(0, 3) == "URC")
                {
                    int index = Convert.ToInt16(message.Substring(3, 1));
                    client.cursorModule.UnblockVirtualCursor(index);
                }

                //message: a left click down should be simulated on the local machine
                else if (message.Substring(0, 8) == "LFTCLKDN")
                {
                    int index = Convert.ToInt16(message.Substring(8, 1));
                    client.cursorModule.SimulateLeftClickDown(index);
                }

                //message: a left click up should be simulated on the local machine
                else if (message.Substring(0, 8) == "LFTCLKUP")
                {
                    int index = Convert.ToInt16(message.Substring(8, 1));
                    client.cursorModule.SimulateLeftClickUp(index);
                }

                //message: a right click down should be simulated on the local machine
                else if (message.Substring(0, 8) == "RGTCLKDN")
                {
                    int index = Convert.ToInt16(message.Substring(8, 1));
                    client.cursorModule.SimulateRightClickDown(index);
                }

                //message: a right click up should be simulated on the local machine
                else if (message.Substring(0, 8) == "RGTCLKUP")
                {
                    int index = Convert.ToInt16(message.Substring(8, 1));
                    client.cursorModule.SimulateRightClickUp(index);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Error in process message!..... " + e.StackTrace + ", message was: " + message);
            }
        }
    }
}
