//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//                      Cursor Redirection Toolkit (CuRT)                       //
//               Written by Andy McGovern, Orchid Research Group                //
//                           andymcgovern@gmail.com                             //
//                          http://orchid.cs.uiuc.edu                           //
//  Department of Computer Science, University of Illinois at Urbana-Champaign  //
//                                July 4, 2006                                  //
//      Copyright 2006, The Board of Trustees of the University of Illinois     //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

using System;
using System.Runtime.InteropServices;
using System.Reflection;
using System.Threading;

using System.Windows.Forms;
using System.Diagnostics;


namespace UIUC.DCS.ORCHID.CuRTClientLibrary
{

    //! UserActivityHook class
    /*!
     *  The UserActivityHook class is used to create global keyboard and mouse listeners that are used for
     *  CuRT system.
     */
	public class UserActivityHook : object {

        //! UserActivityHook constructor
		protected internal UserActivityHook() {
			Start();
		}

        //! UserActivityHook destructor
        ~UserActivityHook()
        { 
			Stop();
		}

        //! This event handler responds to the MouseUp event
        protected internal event MouseEventHandler MouseUp;

        //! This event handler responds to the MouseDown event
        protected internal event MouseEventHandler MouseDown;

        //! This event handler responds to the MouseMove event
        protected internal event MouseEventHandler MouseMove;

        //! This event handler responds to the KeyDown event
        protected internal event KeyEventHandler KeyDown;

        //! This event handler responds to the KeyPress event
        protected internal event KeyPressEventHandler KeyPress;

        //! This event handler responds to the KeyUp event
        protected internal event KeyEventHandler KeyUp;

        //! This is the keyboard/mouse hook delegate
        private delegate int HookProc(int nCode, Int32 wParam, IntPtr lParam);

        //! mouse hook handle as integer
		private static int hMouseHook = 0;

        //! keyboard hook handle as integer
        private static int hKeyboardHook = 0;

        //! global mouse hook constant
		private const int WH_MOUSE_LL = 14;
		
        //! global keyboard hook constant
        private const int WH_KEYBOARD_LL = 13;

        //! MouseHookProcedure as HookProc type
		private HookProc MouseHookProcedure;

        //! KeyboardHookProcedure as HookProc type
		private HookProc KeyboardHookProcedure;

		//! declaration of managed point class
		[StructLayout(LayoutKind.Sequential)]
		private class POINT 
		{
			public int x;
			public int y;
		}

		//! declaration of managed MouseHookStruct class
		[StructLayout(LayoutKind.Sequential)]
		private class MouseHookStruct 
		{
			public POINT pt;
			public int hwnd;
			public int wHitTestCode;
			public int dwExtraInfo;
		}

        //! declaration of managed KeyboardHookStruct class
		[StructLayout(LayoutKind.Sequential)]
		public class KeyboardHookStruct
		{
            //! virtual-key code, range 1-254
			public int vkCode;

            //! hardware scan code for the key
			public int scanCode;

            //! specifies the extended-key flag, event-injected flag, context code, and transition-state flag
			public int flags;

            //! specifies the time stamp for this message
            public int time;

            //! specifies extra information associated with the message
			public int dwExtraInfo;
        }

        #region Mouse/Keyboard event constants
        
        private const int WM_MOUSEMOVE = 0x200;
        private const int WM_LBUTTONDOWN = 0x201;
        private const int WM_RBUTTONDOWN = 0x204;
        private const int WM_MBUTTONDOWN = 0x207;
        private const int WM_LBUTTONUP = 0x202;
        private const int WM_RBUTTONUP = 0x205;
        private const int WM_MBUTTONUP = 0x208;
        private const int WM_LBUTTONDBLCLK = 0x203;
        private const int WM_RBUTTONDBLCLK = 0x206;
        private const int WM_MBUTTONDBLCLK = 0x209;
        private const int MOUSE_DOWN = 0;
        private const int MOUSE_UP = 1;
        private const int MOUSE_MOVE = 2;
        private const int WM_KEYDOWN = 0x100;
        private const int WM_KEYUP = 0x101;
        private const int WM_SYSKEYDOWN = 0x104;
        private const int WM_SYSKEYUP = 0x105;

        #endregion


        #region Imported DLL functions

        //! imported function to install a hook
		[DllImport("user32.dll",CharSet=CharSet.Auto,
			CallingConvention=CallingConvention.StdCall)]
		private static extern int SetWindowsHookEx(int idHook, HookProc lpfn, 
			IntPtr hInstance, int threadId);

		//! imported function to uninstall a hook
		[DllImport("user32.dll",CharSet=CharSet.Auto,
			 CallingConvention=CallingConvention.StdCall)]
		public static extern bool UnhookWindowsHookEx(int idHook);
		
		//! imported function to pass the hook information to next hook procedure in chain
		[DllImport("user32.dll",CharSet=CharSet.Auto,
			 CallingConvention=CallingConvention.StdCall)]
		public static extern int CallNextHookEx(int idHook, int nCode, 
			Int32 wParam, IntPtr lParam);

        //! imported function to translate the specified virtual-key code to the corresponding character
        [DllImport("user32")]
        public static extern int ToAscii(int uVirtKey, //[in] Specifies the virtual-key code to be translated. 
                                         int uScanCode, // [in] Specifies the hardware scan code of the key to be translated. The high-order bit of this value is set if the key is up (not pressed). 
                                         byte[] lpbKeyState, // [in] Pointer to a 256-byte array that contains the current keyboard state. Each element (byte) in the array contains the state of one key. If the high-order bit of a byte is set, the key is down (pressed). The low bit, if set, indicates that the key is toggled on. In this function, only the toggle bit of the CAPS LOCK key is relevant. The toggle state of the NUM LOCK and SCROLL LOCK keys is ignored.
                                         byte[] lpwTransKey, // [out] Pointer to the buffer that receives the translated character or characters. 
                                         int fuState); // [in] Specifies whether a menu is active. This parameter must be 1 if a menu is active, or 0 otherwise. 

        //! imported function to copy the status of the 256 virtual keys to the specified buffer. 
        [DllImport("user32")]
        public static extern int GetKeyboardState(byte[] pbKeyState);

        //! imported function to get the module handle
        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr GetModuleHandle(string lpModuleName);

        #endregion

        //! Start
        /*!
         *  This method installs the keyboard and mouse hooks.
         */
        protected internal void Start()
		{
			// mouse hook
			if(hMouseHook == 0)
			{


				// Create an instance of HookProc.
				MouseHookProcedure = new HookProc(MouseHookProc);
                using (Process curProcess = Process.GetCurrentProcess())
                using (ProcessModule curModule = curProcess.MainModule)
				hMouseHook = SetWindowsHookEx( WH_MOUSE_LL,
					MouseHookProcedure,
                    GetModuleHandle(curModule.ModuleName),
                    0);

				//If SetWindowsHookEx fails.
				if(hMouseHook == 0 )	{
					Stop();
					throw new Exception("SetWindowsHookEx failed.");
				}
			}
			
			//  keyboard hook 
			if(hKeyboardHook == 0)
			{

				
                KeyboardHookProcedure = new HookProc(KeyboardHookProc);

                using (Process curProcess = Process.GetCurrentProcess())
                using (ProcessModule curModule = curProcess.MainModule)

				hKeyboardHook = SetWindowsHookEx( WH_KEYBOARD_LL,
					KeyboardHookProcedure, 
                    GetModuleHandle(curModule.ModuleName),
					0);

				//If SetWindowsHookEx fails.
				if(hKeyboardHook == 0 )	{
					Stop();
					throw new Exception("SetWindowsHookEx ist failed.");
				}
			}
		}

        //! Stop
        /*!
         *  This method uninstalls the hook.
         */
        protected internal void Stop()
		{
			bool retMouse =true;
			bool retKeyboard = true;
			if(hMouseHook != 0)
			{
				retMouse = UnhookWindowsHookEx(hMouseHook);
				hMouseHook = 0;
			} 
			
			if(hKeyboardHook != 0)
			{
				retKeyboard = UnhookWindowsHookEx(hKeyboardHook);
				hKeyboardHook = 0;
			}

            try
            {
                //If UnhookWindowsHookEx fails.
                if (!(retMouse && retKeyboard)) throw new Exception("UnhookWindowsHookEx failed.");

            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }

            }

        //! MouseHookProc
		private int MouseHookProc(int nCode, Int32 wParam, IntPtr lParam)
		{
			int mouseStatus = -1;
			if ((nCode >= 0) && (MouseUp!=null || MouseDown!=null || MouseMove!=null)) 
			{
				
				MouseButtons button=MouseButtons.None;
				
				switch (wParam)
				{
					case WM_MOUSEMOVE:
						mouseStatus = MOUSE_MOVE;
						break;
					case WM_LBUTTONDOWN: 
						button=MouseButtons.Left;
						mouseStatus = MOUSE_DOWN;
						break;
					case WM_LBUTTONUP: 
						button=MouseButtons.Left;
						mouseStatus = MOUSE_UP;
						break;
					case WM_RBUTTONDOWN: 
						button=MouseButtons.Right;
						mouseStatus = MOUSE_DOWN;
						break;
					case WM_RBUTTONUP: 
						button=MouseButtons.Right; 
						mouseStatus = MOUSE_UP;
						break;
				}
				int clickCount=0;
				if (button!=MouseButtons.None)
					if (wParam==WM_LBUTTONDBLCLK || wParam==WM_RBUTTONDBLCLK) clickCount=2;
					else clickCount=1;
				
				//Marshall the data from callback.
				MouseHookStruct MyMouseHookStruct = (MouseHookStruct) Marshal.PtrToStructure(lParam, typeof(MouseHookStruct));
				MouseEventArgs e=new MouseEventArgs(
					button, 
					clickCount, 
					MyMouseHookStruct.pt.x, 
					MyMouseHookStruct.pt.y, 
					0 );
				if (mouseStatus==MOUSE_UP)
					MouseUp(this, e);
				else if (mouseStatus==MOUSE_DOWN)
					MouseDown(this, e);
				else
					MouseMove(this, e);

			}
			return CallNextHookEx(hMouseHook, nCode, wParam, lParam); 
		}

        //! KeyboardHookProc
		private int KeyboardHookProc(int nCode, Int32 wParam, IntPtr lParam)
		{
			if ((nCode >= 0) && (KeyDown!=null || KeyUp!=null || KeyPress!=null))
			{
				KeyboardHookStruct MyKeyboardHookStruct = (KeyboardHookStruct) Marshal.PtrToStructure(lParam, typeof(KeyboardHookStruct));
				// raise KeyDown

				if ( KeyDown!=null && ( wParam ==WM_KEYDOWN || wParam==WM_SYSKEYDOWN ))
				{
					Keys keyData=(Keys)MyKeyboardHookStruct.vkCode;
					KeyEventArgs e = new KeyEventArgs(keyData);
					KeyDown(this, e);
				}
				
				// raise KeyPress
				if ( KeyPress!=null &&  wParam ==WM_KEYDOWN )
				{
					byte[] keyState = new byte[256];
					GetKeyboardState(keyState);

					byte[] inBuffer= new byte[2];
					if (ToAscii(MyKeyboardHookStruct.vkCode,
				            MyKeyboardHookStruct.scanCode,
				            keyState,
				            inBuffer,
				            MyKeyboardHookStruct.flags)==1) 
				            {
				            	KeyPressEventArgs e = new KeyPressEventArgs((char)inBuffer[0]);
								KeyPress(this, e);
				            }
				}
				
				// raise KeyUp
				if ( KeyUp!=null && ( wParam ==WM_KEYUP || wParam==WM_SYSKEYUP ))
				{
					Keys keyData=(Keys)MyKeyboardHookStruct.vkCode;
					KeyEventArgs e = new KeyEventArgs(keyData);
					KeyUp(this, e);
				}

			}
			return CallNextHookEx(hKeyboardHook, nCode, wParam, lParam); 
		}
	}
}
