//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//                      Cursor Redirection Toolkit (CuRT)                       //
//               Written by Andy McGovern, Orchid Research Group                //
//                           andymcgovern@gmail.com                             //
//                          http://orchid.cs.uiuc.edu                           //
//  Department of Computer Science, University of Illinois at Urbana-Champaign  //
//                                July 4, 2006                                  //
//      Copyright 2006, The Board of Trustees of the University of Illinois     //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;

namespace CuRTClient
{
    //! ConnectionBox
    /*!
     *  The ConnetionBox form is used so that a CuRTClient can connect to the CuRTServer.  The user speficies
     *  the IP address or hostname, a name to identify his own machine on the server, and a hotkey to be used
     *  to return his machine's cursor to the local machine when being remoted.  Note that this ConnectionBox
     *  form is only used in the standalone mode of CuRT system.  When CuRT is used as an API, all of the above
     *  mentioned options are speficied programatically, in the CuRTClient object's constructor.
     */
    public partial class ConnectionBox : Form
    {

        //! this is a reference back to the parent form
        private CuRTGUI guiForm;

        //! The constructor simply adds items to the dropdown menus and copies the reference to the guiForm.
        protected internal ConnectionBox(CuRTGUI guiForm)
        {
            InitializeComponent();

            this.guiForm = guiForm;

            ModifierList.Items.Add("CTRL");
            ModifierList.Items.Add("ALT");
            ModifierList.Items.Add("WINKEY");
            ModifierList.SelectedIndex = 0;

            KeyList.Items.Add(Keys.F1);
            KeyList.Items.Add(Keys.F2);
            KeyList.Items.Add(Keys.F3);
            KeyList.Items.Add(Keys.F4);
            KeyList.Items.Add(Keys.F5);
            KeyList.Items.Add(Keys.F6);
            KeyList.Items.Add(Keys.F7);
            KeyList.Items.Add(Keys.F8);
            KeyList.Items.Add(Keys.F9);
            KeyList.Items.Add(Keys.F10);
            KeyList.Items.Add(Keys.F11);
            KeyList.Items.Add(Keys.F12);
            KeyList.Items.Add(Keys.A);
            KeyList.Items.Add(Keys.B);
            KeyList.Items.Add(Keys.C);
            KeyList.Items.Add(Keys.D);
            KeyList.Items.Add(Keys.E);
            KeyList.Items.Add(Keys.F);
            KeyList.Items.Add(Keys.G);
            KeyList.Items.Add(Keys.H);
            KeyList.Items.Add(Keys.I);
            KeyList.Items.Add(Keys.J);
            KeyList.Items.Add(Keys.K);
            KeyList.Items.Add(Keys.L);
            KeyList.Items.Add(Keys.M);
            KeyList.Items.Add(Keys.N);
            KeyList.Items.Add(Keys.O);
            KeyList.Items.Add(Keys.P);
            KeyList.Items.Add(Keys.Q);
            KeyList.Items.Add(Keys.R);
            KeyList.Items.Add(Keys.S);
            KeyList.Items.Add(Keys.T);
            KeyList.Items.Add(Keys.U);
            KeyList.Items.Add(Keys.V);
            KeyList.Items.Add(Keys.W);
            KeyList.Items.Add(Keys.X);
            KeyList.Items.Add(Keys.Y);
            KeyList.Items.Add(Keys.Z);
            KeyList.SelectedIndex = 0;

        }


        //! This is where the actual connection to the server is made
        /*!
         *  Several steps occur here as the connection is made.  At each step, a method in the guiForm
         *  is called, and a progress bar and status messages are displayed for the user.  The methods 
         *  return false if there was a problem, and that false return value causes an appropriate error 
         *  message to be displayed here.  Otherwise, if there are no problems during the initialization, 
         *  this form disappears, the guiForm is activated, and the cursor redirection can begin.
         */
        private void ConnectButton_Click(object sender, EventArgs e)
        {
            bool success = true;
            Status.ForeColor = Color.Black;

            //verify that the user has entered a username
            if (LocalUsername.Text == "")
            {
                ProgressBar.Value = 0;
                Status.ForeColor = Color.Red;
                Status.Text = "Status: blank username not allowed.  Please enter a username.";
                Status.Refresh();
                success = false;
            }

            //start connection process
            if (success)
            {
                Status.Text = "Status: starting to connect...";
                Status.Refresh();
                ProgressBar.Value = 20;
                Thread.Sleep(500);
                Status.Text = "Status: registering hotkey...";
                Status.Refresh();
            }

            //register hotkey
            if (success)
            {
                KeyModifiers modifierKey = KeyModifiers.Control;
                switch (ModifierList.SelectedIndex)
                {
                    case 0: modifierKey = KeyModifiers.Control; break;
                    case 1: modifierKey = KeyModifiers.Alt; break;
                    case 2: modifierKey = KeyModifiers.Windows; break;

                }
                    

                if (guiForm.TryHotkey(modifierKey, (Keys)KeyList.SelectedItem))
                {   
                    Thread.Sleep(500);
                    ProgressBar.Value = 40;
                    Status.Text = "Status: connecting to server...";
                    Status.Refresh();
                }
                else
                {   
                    ProgressBar.Value = 0;
                    Status.ForeColor = Color.Red;
                    Status.Text = "Status: hotkey registration failed.  Please choose another hotkey.";
                    Status.Refresh();
                    success = false;
                }
            }

            //connect to remote server
            if (success)
            {
                if (guiForm.InitializeSocketModule(ServerIP.Text, LocalUsername.Text))
                {
                    ProgressBar.Value = 60;
                    Status.Text = "Status: initializing cursors...";
                    Status.Refresh();
                }
                else
                {
                    ProgressBar.Value = 0;
                    Status.ForeColor = Color.Red;
                    Status.Text = "Status: connection to server failed.  Please check the IP address and try again.";
                    Status.Refresh();
                    success = false;
                }
            }

            //initialize cursors and listeners
            if (success)
            {
                if (guiForm.InitializeCursorModule())
                {
                    Thread.Sleep(500);
                    ProgressBar.Value = 80;
                    Status.Text = "Status: initializing mouse/keyboard listeners and main window...";
                    Status.Refresh();                
                }
                else
                {
                    ProgressBar.Value = 0;
                    Status.ForeColor = Color.Red;
                    Status.Text = "Status: cursor initialization failed.  Please retry.";
                    Status.Refresh();
                    success = false;
                }
            }

            //initialize GUI form
            if (success)
            {
                if (guiForm.InitializeGUIForm())
                {
                    Thread.Sleep(500);
                    ProgressBar.Value = 100;
                    Status.ForeColor = Color.Green;
                    Status.Text = "Status: all initializations successful.  Starting program.";
                    Status.Refresh();
                }
                else
                {
                    ProgressBar.Value = 0;
                    Status.ForeColor = Color.Red;
                    Status.Text = "Status: initialization of mouse/keyboard listeners failed.  Please retry.";
                    Status.Refresh();
                    success = false;
                }
            }

            Thread.Sleep(3000);

            //If initializations are successful, activate GUI form and hide this form.
            //Otherwise, release hotkey and let user try to connect again.
            if (success)
            {
                guiForm.Enabled = true;
                guiForm.Activate();
                guiForm.BringToFront();
                this.Dispose();
            }
            else
                guiForm.UnlockHotkey();
        
        }

        private void ExitButton_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }
    }
}