//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//                      Cursor Redirection Toolkit (CuRT)                       //
//               Written by Andy McGovern, Orchid Research Group                //
//                           andymcgovern@gmail.com                             //
//                          http://orchid.cs.uiuc.edu                           //
//  Department of Computer Science, University of Illinois at Urbana-Champaign  //
//                                July 4, 2006                                  //
//      Copyright 2006, The Board of Trustees of the University of Illinois     //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;
using System.Net;
using System.Text;
using System.Drawing;
using System.Net.Sockets;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Reflection;
using System.Threading;

namespace CuRTClient
{

    //! CuRTGUI
    /*!
     *  The CuRTGUI serves two main purposes.  First, it acts as the graphical user interface for the CuRT system.
     *  This GUI is available in both the standalone application and API modes, but it is primarily used in the
     *  standalone application mode.  The second purpose of the CuRTGUI is to be the intermediary between the operating
     *  system and the CuRT system.  It manages components of the system such as the mouse/keyboard event listeners (hooks)
     *  hotkeys, and the click guard that prevents the local machine's clicks from affecting the local machine when it is 
     *  currently being remoted.
     */
	public class CuRTGUI : System.Windows.Forms.Form
	{
		private System.ComponentModel.IContainer components;

        //! a reference to the CuRTClient object
        private static CuRTClient client;

        //! the hook that is used to capture all local keyboard and mouse events
		private UserActivityHook actHook;

        //! the invisible, screen-sized form that is used to prevent clicks by the local mouse on the local machine
        private ClickGuard clickGuard;

        //! used to store the cursor's previous X value 
        protected internal int prevX;

        //! used to store the cursor's previous Y value 
        protected internal int prevY;

        //! this is another Windows form that allows the user the specify connection parameters in the standalone application mode
        private ConnectionBox newConnection;

        /*!
         *  This flag is used to prevent race conditions when jumping the local cursor from the 
         *  edge of the screen back to the center (this is done when the local cursor is being remoted,
         *  as it must loop around the local screen, invisibly, to be able to send accurate coordinate
         *  changes to the remote machine).
         */
        private bool cursorJumpMutex;

        //! the local machine's primary screen width, in pixels
        private int screenWidth;

        //! the local machine's primary screen height, in pixels
        private int screenHeight;

        //! a string that represents the hotkey's key modifier (ALT, CTRL, or WINKEY)
        private String keyModifierAsString;

        //! the hotkey used with a key modifier to return the local cursor to the local machine
        private Keys hotkey;

        //! the unique ID of the CuRT hotkey for the Windows OS
        const int HOTKEY_ID = 31197;

        //! the hotkey identifier used with the WndProc message receiver
        const int WM_HOTKEY = 0x0312;

		#region Form Components
        

        private System.Windows.Forms.NotifyIcon systemTrayIcon;
        private ComboBox clientUsernameList;
        private Button button4;
        private Button button5;
        public bool connected;
        private TextBox Status;
        private Label label1;
        private Button SendButton;
        private MenuStrip menuBar;
        private ToolStripMenuItem fileToolStripMenuItem;
        private ToolStripMenuItem aboutToolStripMenuItem;
        private ToolStripMenuItem aboutCuRTToolStripMenuItem;
        private ToolStripMenuItem exitProgramToolStripMenuItem;

		#endregion

        #region Form Event Methods

        //! CuRTGUI_FormClosed
        /*!
         *  This method releases the hotkey when the CuRTGUI window closes.
         */
        private void CuRTGUI_FormClosed(object sender, FormClosedEventArgs e)
        {
            bool bcheck = UnregisterHotKey(Handle, HOTKEY_ID);
        }

        //! CuRTGUI_Resize
        /*!
         *  This method causes the CuRT application to disappear from the taskbar when minimized
         *  (it is still accessible from the system tray).
         */
        private void CuRTGUI_Resize(object sender, EventArgs e)
        {
            if (FormWindowState.Minimized == WindowState)
                Hide();
        }

        //! CuRTGUI_Resize
        /*!
         *  This method attempts to send the local cursor to a remote machine, as specified by the selected
         *  index of the ClientUsernameList dropdown box.  If the user tries to select his own machine,
         *  an error message is displayed.
         */
        private void SendButton_Click(object sender, EventArgs e)
        {
            int selectedIndex = clientUsernameList.SelectedIndex;

            if (selectedIndex == client.localIndex)
            {
                Status.ForeColor = Color.Red;
                Status.Text = "Error: you cannot redirect mouse/keyboard input to your own machine.  Please choose another machine.";
                Status.Refresh();
            }
            else
            {
                client.SendLocalCursorToRemoteMachine(selectedIndex, new Point(700, 700));
            }

        }

        //! UserList_Click
        /*!
         *  This method refreshes the clientUsernameList every time the dropdown list is clicked (note that all
         *  of the [EMPTY] entries are first removed.
         */
        private void clientUsernameList_Click(object sender, EventArgs e)
        {
            clientUsernameList.Items.Clear();
            
            int count = 9;

            while (count != 0 && client.clientUsernameList[count] == "[EMPTY]")
                count--;


            String[] newClientUsernameList = new String[count+1];

            if (client.clientUsernameList != null)
            {
                for (int i = 0; i < count+1; i++)
                {
                    newClientUsernameList[i] = client.clientUsernameList[i];
                }

                clientUsernameList.Items.AddRange(newClientUsernameList);
                clientUsernameList.MaxDropDownItems = newClientUsernameList.Length;
                clientUsernameList.Refresh();
            }
        }

        //! systemTrayIcon_DoubleClick
        /*!
         *  This method restores the CuRT application when its system tray icon is double-clicked.
         */
        private void systemTrayIcon_DoubleClick(object sender, EventArgs e)
        {
            Show();
            WindowState = FormWindowState.Normal;
        }

        //! exitProgramMenuBar_Click
        /*!
         *  This method exits the application when the user chooses to do so from the menu bar.
         */
        private void exitProgramMenuBar_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        #endregion

        #region Imported DLL functions

        //! SetCursorPos imported DLL function
        /*!
         *  This function provides the ability to send the local machine's cursor to a specific point on the screen.
         */
        [DllImport("user32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool SetCursorPos(int X, int Y);

        //! RegisterHotKey imported DLL function
        /*!
         *  This function provides the ability to register a hotkey on the Windows OS.
         */
        [DllImport("user32.dll", SetLastError = true)]
        public static extern bool RegisterHotKey(
            IntPtr hWnd, // handle to window    
            int id, // hot key identifier    
           KeyModifiers fsModifiers, // key-modifier options    
            Keys vk    // virtual-key code    
            );

        //! UnregisterHotKey imported DLL function
        /*!
         *  This function provides the ability to unregister a hotkey on the Windows OS.
         */
        [DllImport("user32.dll", SetLastError = true)]
        public static extern bool UnregisterHotKey(
            IntPtr hWnd, // handle to window    
            int id      // hot key identifier    
            );


		#endregion

        #region Windows Form Designer generated code

        private void InitializeComponent()
        {
            this.components = new System.ComponentModel.Container();
            System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(CuRTGUI));
            this.systemTrayIcon = new System.Windows.Forms.NotifyIcon(this.components);
            this.clientUsernameList = new System.Windows.Forms.ComboBox();
            this.button4 = new System.Windows.Forms.Button();
            this.button5 = new System.Windows.Forms.Button();
            this.Status = new System.Windows.Forms.TextBox();
            this.label1 = new System.Windows.Forms.Label();
            this.SendButton = new System.Windows.Forms.Button();
            this.menuBar = new System.Windows.Forms.MenuStrip();
            this.fileToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.exitProgramToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.aboutToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.aboutCuRTToolStripMenuItem = new System.Windows.Forms.ToolStripMenuItem();
            this.menuBar.SuspendLayout();
            this.SuspendLayout();
            // 
            // systemTrayIcon
            // 
            this.systemTrayIcon.Icon = ((System.Drawing.Icon)(resources.GetObject("systemTrayIcon.Icon")));
            this.systemTrayIcon.Text = "Input Redirection Toolkit";
            this.systemTrayIcon.Visible = true;
            this.systemTrayIcon.DoubleClick += new System.EventHandler(this.systemTrayIcon_DoubleClick);
            // 
            // clientUsernameList
            // 
            this.clientUsernameList.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
            this.clientUsernameList.Font = new System.Drawing.Font("Microsoft Sans Serif", 11.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
            this.clientUsernameList.FormattingEnabled = true;
            this.clientUsernameList.Location = new System.Drawing.Point(255, 24);
            this.clientUsernameList.Name = "clientUsernameList";
            this.clientUsernameList.Size = new System.Drawing.Size(362, 26);
            this.clientUsernameList.TabIndex = 35;
            this.clientUsernameList.Click += new System.EventHandler(this.clientUsernameList_Click);
            // 
            // button4
            // 
            this.button4.Location = new System.Drawing.Point(612, 132);
            this.button4.Name = "button4";
            this.button4.Size = new System.Drawing.Size(57, 36);
            this.button4.TabIndex = 36;
            this.button4.Text = "button4";
            this.button4.UseVisualStyleBackColor = true;
            // 
            // button5
            // 
            this.button5.Location = new System.Drawing.Point(692, 129);
            this.button5.Name = "button5";
            this.button5.Size = new System.Drawing.Size(55, 39);
            this.button5.TabIndex = 37;
            this.button5.Text = "button5";
            this.button5.UseVisualStyleBackColor = true;
            // 
            // Status
            // 
            this.Status.BackColor = System.Drawing.SystemColors.Menu;
            this.Status.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle;
            this.Status.Font = new System.Drawing.Font("Microsoft Sans Serif", 11.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
            this.Status.Location = new System.Drawing.Point(4, 56);
            this.Status.Multiline = true;
            this.Status.Name = "Status";
            this.Status.ReadOnly = true;
            this.Status.Size = new System.Drawing.Size(696, 50);
            this.Status.TabIndex = 38;
            // 
            // label1
            // 
            this.label1.AutoSize = true;
            this.label1.Font = new System.Drawing.Font("Microsoft Sans Serif", 11.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
            this.label1.Location = new System.Drawing.Point(1, 24);
            this.label1.Name = "label1";
            this.label1.Size = new System.Drawing.Size(248, 18);
            this.label1.TabIndex = 39;
            this.label1.Text = "Send local mouse/keyboard input to:";
            // 
            // SendButton
            // 
            this.SendButton.Location = new System.Drawing.Point(623, 24);
            this.SendButton.Name = "SendButton";
            this.SendButton.Size = new System.Drawing.Size(77, 23);
            this.SendButton.TabIndex = 40;
            this.SendButton.Text = "Send";
            this.SendButton.UseVisualStyleBackColor = true;
            this.SendButton.Click += new System.EventHandler(this.SendButton_Click);
            // 
            // menuBar
            // 
            this.menuBar.BackColor = System.Drawing.SystemColors.Control;
            this.menuBar.Items.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.fileToolStripMenuItem,
            this.aboutToolStripMenuItem});
            this.menuBar.Location = new System.Drawing.Point(0, 0);
            this.menuBar.Name = "menuBar";
            this.menuBar.Size = new System.Drawing.Size(709, 24);
            this.menuBar.TabIndex = 41;
            this.menuBar.Text = "menuBar";
            // 
            // fileToolStripMenuItem
            // 
            this.fileToolStripMenuItem.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.exitProgramToolStripMenuItem});
            this.fileToolStripMenuItem.Name = "fileToolStripMenuItem";
            this.fileToolStripMenuItem.Size = new System.Drawing.Size(35, 20);
            this.fileToolStripMenuItem.Text = "File";
            // 
            // exitProgramToolStripMenuItem
            // 
            this.exitProgramToolStripMenuItem.Name = "exitProgramToolStripMenuItem";
            this.exitProgramToolStripMenuItem.Size = new System.Drawing.Size(136, 22);
            this.exitProgramToolStripMenuItem.Text = "Exit Program";
            this.exitProgramToolStripMenuItem.Click += new System.EventHandler(this.exitProgramMenuBar_Click);
            // 
            // aboutToolStripMenuItem
            // 
            this.aboutToolStripMenuItem.DropDownItems.AddRange(new System.Windows.Forms.ToolStripItem[] {
            this.aboutCuRTToolStripMenuItem});
            this.aboutToolStripMenuItem.Name = "aboutToolStripMenuItem";
            this.aboutToolStripMenuItem.Size = new System.Drawing.Size(47, 20);
            this.aboutToolStripMenuItem.Text = "About";
            // 
            // aboutCuRTToolStripMenuItem
            // 
            this.aboutCuRTToolStripMenuItem.Name = "aboutCuRTToolStripMenuItem";
            this.aboutCuRTToolStripMenuItem.Size = new System.Drawing.Size(152, 22);
            this.aboutCuRTToolStripMenuItem.Text = "About CuRT";
            this.aboutCuRTToolStripMenuItem.Click += new System.EventHandler(this.aboutCuRTToolStripMenuItem_Click);
            // 
            // CuRTGUI
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(709, 118);
            this.Controls.Add(this.SendButton);
            this.Controls.Add(this.label1);
            this.Controls.Add(this.Status);
            this.Controls.Add(this.button5);
            this.Controls.Add(this.button4);
            this.Controls.Add(this.clientUsernameList);
            this.Controls.Add(this.menuBar);
            this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
            this.MainMenuStrip = this.menuBar;
            this.Name = "CuRTGUI";
            this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
            this.Text = "Cursor Redirection Toolkit (CuRT)";
            this.FormClosed += new System.Windows.Forms.FormClosedEventHandler(this.CuRTGUI_FormClosed);
            this.Resize += new System.EventHandler(this.CuRTGUI_Resize);
            this.menuBar.ResumeLayout(false);
            this.menuBar.PerformLayout();
            this.ResumeLayout(false);
            this.PerformLayout();

        }
        #endregion

        //! CuRTGUI constructor
        /*!
         *  This is one of two constructors for the CuRTGUI class.  This one is used in the standalone version of the
         *  application.  The only difference is that this constructor first displays the ConnectionBox form, which
         *  allows the user to manually specify the connection parameters before connecting.
         */
        protected internal CuRTGUI(CuRTClient newClient)
		{
            
            //this flag allows prevents a Windows Forms exception that occurs when different threads try to access
            //the same data member at the same time
            CheckForIllegalCrossThreadCalls = false;

            //this method initializes all of the components of the Windows form
            InitializeComponent();

            //this copies the client reference to the member variable of this class
            client = newClient;

            //this copies this object to the client's reference to the GUI
            client.GUI = this;

            //this form is disabled until a successful connection has been made
            this.Enabled = false;

            //the text of the CuRTGUI form status box is set, to be seen once the connection is made
            Status.ForeColor = Color.Black;
            Status.Text = "Status: waiting to send local mouse/keyboard input to remote machine.";
            Status.Refresh();

            //a ConnectionBox object is instantiated and made active
            newConnection = new ConnectionBox(this);
            newConnection.Show();
            newConnection.Activate();
            this.SendToBack();
            newConnection.BringToFront();


       
		}

        //! CuRTGUI constructor
        /*!
         *  This is one of two constructors for the CuRTGUI class.  This one is used in the API version of the
         *  application.  The only difference is that this constructor does not display the ConnectionBox form;
         *  instead, the connection parameters are specified programatically through the API.
         */
        protected internal CuRTGUI(CuRTClient newClient, String clientName, String IP, KeyModifiers modkey, 
            Keys hotkey, CursorColors[] newColorScheme, 
            int newCursorMovementResolution, EdgeDetectionParams newEdgeDetectionParams)
		{
            //this flag allows prevents a Windows Forms exception that occurs when different threads try to access
            //the same data member at the same time
            CheckForIllegalCrossThreadCalls = false;
            //this method initializes all of the components of the Windows form
            InitializeComponent();

            //this copies the client reference to the member variable of this class
            client = newClient;

            //this copies this object to the client's reference to the GUI
            client.GUI = this;

            //The following lines are used to initialize the various components of the CuRT system, and also
            //to provide feedback in the case of a failure of one of the components to initialize.  This code
            //is very similiar to the code in the ConnectButton_Click method of the ConnectionBox class.  Please
            //see that method for comments on this section.
            bool success = true;
            Status.ForeColor = Color.Black;
            Status.Text = "Status: waiting to send local mouse/keyboard input to remote machine.";
            Status.Refresh();

            if (TryHotkey(modkey, hotkey))
            {   
                Status.Text = "Status: connecting to server...";
                Status.Refresh();
            }
            else
            {   
                Status.ForeColor = Color.Red;
                Status.Text = "Status: hotkey registration failed.  Please choose another hotkey.";
                Status.Refresh();
                success = false;
            }
            if (success)
            {
                if (InitializeSocketModule(IP, clientName))
                {
                    Status.Text = "Status: initializing cursors...";
                    Status.Refresh();
                }
                else
                {
                    Status.ForeColor = Color.Red;
                    Status.Text = "Status: connection to server failed.  Please check the IP address and try again.";
                    Status.Refresh();
                    success = false;
                }
            }
            if (success)
            {
                if (InitializeCursorModule(newColorScheme, newCursorMovementResolution, newEdgeDetectionParams))
                {
                    Thread.Sleep(500);
                    Status.Text = "Status: initializing mouse/keyboard listeners and main window...";
                    Status.Refresh();                
                }
                else
                {
                    Status.ForeColor = Color.Red;
                    Status.Text = "Status: cursor initialization failed.  Please retry.";
                    Status.Refresh();
                    success = false;
                }
            }
            if (success)
            {
                if (InitializeGUIForm())
                {
                    Status.ForeColor = Color.Green;
                    Status.Text = "Status: all initializations successful.  Starting program.";
                    Status.Refresh();
                }
                else
                {
                    Status.ForeColor = Color.Red;
                    Status.Text = "Status: initialization of mouse/keyboard listeners failed.  Please retry.";
                    Status.Refresh();
                    success = false;
                }
            }
            Thread.Sleep(3000);

            if (success)
            {  
                //all initializations are successful, so the text of the CuRTGUI form status box is set
                Status.ForeColor = Color.Black;
                Status.Text = "Status: waiting to send local mouse/keyboard input to remote machine.";
                Status.Refresh();
            }
            else
                UnlockHotkey();
            }

        //! Overridden Dispose method
        /*!
         *  This method overrides the CuRTGUI Dispose method so that the hotkey is released in the case of the
         *  window closing.
         */
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
                {
                    bool bcheck = UnregisterHotKey(Handle, HOTKEY_ID);
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

        //! TryHotKey
        /*!
         *  This method attempts to initialize a hotkey.  It returns true if the initialization is successful,
         *  and false otherwise.
         */
        protected internal bool TryHotkey(KeyModifiers modifierKey, Keys hotkey)
        {   bool returnVal = RegisterHotKey(Handle, HOTKEY_ID, modifierKey, hotkey);
            if (returnVal)
            {
                keyModifierAsString= "";
                this.hotkey = hotkey;
                switch (modifierKey)
                {
                    case KeyModifiers.Alt: keyModifierAsString = "ALT"; break;
                    case KeyModifiers.Control: keyModifierAsString = "CTRL"; break;
                    case KeyModifiers.Windows: keyModifierAsString = "WINKEY"; break;
                }

                return true;
            }
            else
                return false;
        }

        //! UnlockHotkey
        /*!
         *  This method releases the hotkey from the Windows OS
         */
        protected internal void UnlockHotkey()
        {
            UnregisterHotKey(Handle, HOTKEY_ID);
        }

        //! InitializeSocketModule
        /*!
         *  This method calls the appropriate CuRTClient method to initialize the socket module.
         */
        protected internal bool InitializeSocketModule(String serverIP, String username)
        {
            return client.InitializeSocketModule(serverIP, username);
        }

        //! InitializeCursorModule
        /*!
         *  This method calls the appropriate CuRTClient method to initialize the cursor module.  This is
         *  the standalone application version of this method.
         */
        protected internal bool InitializeCursorModule()
        {
            return client.InitializeCursorModule();
        }

        //! InitializeCursorModule
        /*!
         *  This method calls the appropriate CuRTClient method to initialize the cursor module.  This is
         *  the API version of this method.
         */
        protected internal bool InitializeCursorModule(CursorColors[] newColorScheme, 
            int newCursorMovementResolution, EdgeDetectionParams newEdgeDetectionParams)
        {
            return client.InitializeCursorModule(newColorScheme, newCursorMovementResolution, newEdgeDetectionParams);
        }

        

        //! InitializeGUIForm
        /*!
         *  This method initializes the components of the GUI form.
         */
        protected internal bool InitializeGUIForm()
        {
            clickGuard = new ClickGuard();
            clickGuard.Visible = false;

            cursorJumpMutex = true;

            screenWidth = Screen.PrimaryScreen.Bounds.Width;
            screenHeight = Screen.PrimaryScreen.Bounds.Height;
            InitializeHook();

            return true;
        }

        //! InitializeHook
        /*!
         *  This method initializes all of the necessary keyboard and mouse hooks.
         */
        protected internal void InitializeHook()
		{
			actHook = new UserActivityHook();
			actHook.MouseUp += new MouseEventHandler(MyMouseUp);
			actHook.MouseDown += new MouseEventHandler(MyMouseDown);
			actHook.MouseMove += new MouseEventHandler(MyMouseMove);
			actHook.KeyDown += new KeyEventHandler(MyKeyDown);
			actHook.KeyPress += new KeyPressEventHandler(MyKeyPress);
			actHook.Start();
		}

        //! MyMouseDown
        /*!
         *  This method is called whenever a local MouseDown event is fired.  It first must be determined if
         *  the MouseDown is coming from the local machine itself or from a virtual cursor on the machine that
         *  is simulating the MouseDown.  If it is coming from the local machine itself, then appropriate actions
         *  are taken.
         */
        protected internal void MyMouseDown(object sender, MouseEventArgs e)
        {   
            //There must be no other drag in progress when this MouseDown event occurs.  If this is the beginning of a virtual
            //drag, then the client.cursorModule.dragFromVirtualCursor flag will have just been set to true in the
            //SimulateLeftClickDown method of the CursorModule, so the test will be false.  If the test is true, then
            //we know that a local drag is beginning.
            if (!client.cursorModule.dragFromVirtualCursor && !client.cursorModule.dragFromLocal)
            {
                //At this point, we know that either the local cursor is being remoted and is enacting a drag on the remote
                //machine, or the local cursor is local and is enacting a drag on the local machine.  In either case, we must
                //block all virtual cursors.
                client.cursorModule.BlockAllVirtualCursors();

                //Regardless of whether the local cursor is being remoted or not, we know that the drag is originating from
                //the local machine.
                client.cursorModule.dragFromLocal = true;

                //check to see if the local cursor (that is undergoing a drag) is currently being remoted
                if (client.cursorModule.localCursorIsRemote)
                {
                    //if the local cursor is being remoted, then the MouseDown should be passed to the remote machine
                    if (e.Button == MouseButtons.Left)
                        client.socketModule.SendUDPMessage("LFTCLKDN");
                    else if (e.Button == MouseButtons.Right)
                        client.socketModule.SendUDPMessage("RGTCLKDN");
                }
            }
        }

        //! MyMouseUp
        /*!
         *  This method is called whenever a local MouseUp event is fired.  It first must be determined if
         *  the MouseUp is coming from the local machine itself or from a virtual cursor on the machine that
         *  is simulating the MouseUp.  If it is coming from the local machine itself, then appropriate actions
         *  are taken.
         */
        protected internal void MyMouseUp(object sender, MouseEventArgs e)
		{
            //Check to make sure that this MouseUp event is coming from the local machine's mouse, and not from
            //a simulated virtual cursor MouseUp on the local machine.
            if (client.cursorModule.dragFromLocal)
            {
                //drag is coming to an end, so set flag to false and unblock virtual cursors
                client.cursorModule.dragFromLocal = false;
                client.cursorModule.UnblockAllVirtualCursors();

                //check to see if the local cursor (that is undergoing a drag) is currently being remoted
                if (client.cursorModule.localCursorIsRemote)
                {
                    //if the local cursor is being remoted, then the MouseUp should be passed to the remote machine
                    if (e.Button == MouseButtons.Left)
                        client.socketModule.SendUDPMessage("LFTCLKUP");
                    else if (e.Button == MouseButtons.Right)
                        client.socketModule.SendUDPMessage("RGTCLKUP");
                }
            }
		}


        //! MyMouseMove
        /*!
         *  This method is called whenever a local MouseMove event is fired.
         */
        protected internal void MyMouseMove(object sender, MouseEventArgs e)
        {
            //increment currCursorMovementCount for use with cursor movement resolution parameter
            client.cursorModule.currCursorMovementCount++;

            //if local cursor is currently on local machine, then do edge detection checks
            if (!client.cursorModule.localCursorIsRemote)
                client.cursorModule.CheckEdges(client.localIndex, e.X, e.Y);

            //The cursorJumpMutex is used when the local cursor is being remoted and the invisible local cursor 
            //on the local machine is being jumped from the edge of the screen back to the center.  If the cursorJumpMutex
            //is currently true, then no edge-to-center jump is currently taking place.  If no edge-to-center jump is currently
            //taking place, and the local cursor is being remoted, and the currCursorMovementCount matches the
            //cursorMovementResolution parameter, then the cursor's relative coordinates (i.e. its change in coordinates since
            //the last time this section of code was entered) is sent to the remote machine where the local cursor is currently
            //being remoted.
            if (cursorJumpMutex == true && client.cursorModule.localCursorIsRemote 
                && client.cursorModule.currCursorMovementCount >= client.cursorModule.cursorMovementResolution)
            {   
                client.cursorModule.currCursorMovementCount = 0;
                //determine the relative coordinate change for both the X and Y coordinates
                int xDiff = e.X - prevX;
                int yDiff = e.Y - prevY;

                //this test prevents the edge-to-center jump from being registered as a coordinate change that should be sent
                //to the remote machine
                if (xDiff < 100 && yDiff < 100)
                    client.socketModule.SendRelativeCoordinates(xDiff, yDiff);
                prevX = e.X;
                prevY = e.Y;
            }

            //This test checks to see whether the local cursor on the local machine (which is invisible) is approaching an
            //edge of the screen (with a buffer of 10 pixels on each edge).  If so, the cursor gets sent to the middle of
            //the screen.
            if (cursorJumpMutex && ((e.X < 10 || e.X > (screenWidth - 10) || e.Y < 10 || e.Y > (screenHeight - 10))
                && client.cursorModule.localCursorIsRemote))
            {
                //the thread is used to help with efficiency and to ensure correct timing of the jump
                Thread JumpLocalCursorToCenterThread = new Thread(new ThreadStart(JumpLocalCursorToCenter));
                JumpLocalCursorToCenterThread.Name = "JumpLocalCursorToCenterThread";
                cursorJumpMutex = false;
                JumpLocalCursorToCenterThread.Start();
            }
            

            
		}

        //! JumpLocalCursorToCenter
        /*!
         *  This threaded method is used to jump the local (invisible) cursor from the edge of the screen back to the center
         *  when it is being remoted.
         */
        private void JumpLocalCursorToCenter()
        {
          Thread.Sleep(10);
          SetCursorPos(screenWidth/2, screenHeight/2);
          prevX = screenWidth/2;
          prevY = screenHeight/2;
          Thread.Sleep(25);
          cursorJumpMutex = true;
        }

        //! MyKeyDown
        /*!
         *  This method is called as a reaction to the event when special non-character keys are pressed on the local machine.
         *  If the key pressed matches one of the possible keys in the switch statement, the string represntation of the key
         *  is converted into the format that can later be reproduced using the SendKeys method.  If the local cursor is currently
         *  being remoted, then the appropriate message is sent to reproduce the keypress.
         */
		public void MyKeyDown(object sender, KeyEventArgs e)
		{
            String key = e.KeyData.ToString();
            String keyToPress = "";
            switch (key)
            {
                case "Up": keyToPress = "{UP}";
                    break;
                case "Down": keyToPress = "{DOWN}";
                    break;
                case "Left": keyToPress = "{LEFT}";
                    break;
                case "Right": keyToPress = "{RIGHT}";
                    break;
                case "Home": keyToPress = "{HOME}";
                    break;
                case "End": keyToPress = "{END}";
                    break;
                case "PageUp": keyToPress = "{PGUP}";
                    break;
                case "PageDown": keyToPress = "{PGDN}";
                    break;
                case "Next": keyToPress = "{PGDN}";
                    break;
                case "Delete": keyToPress = "{DEL}";
                    break;
                case "Insert": keyToPress = "{INS}";
                    break;
                case "Tab": keyToPress = "{TAB}";
                    break;
                case "Escape": keyToPress = "{ESC}";
                    break;
                case "Capital": keyToPress = "{CAPSLOCK}";
                    break;
                case "NumLock": keyToPress = "{NUMLOCK}";
                    break;
                case "Scroll": keyToPress = "{SCROLLLOCK}";
                    break;
                case "F1": keyToPress = "{F1}";
                    break;
                case "F2": keyToPress = "{F2}";
                    break;
                case "F3": keyToPress = "{F3}";
                    break;
                case "F4": keyToPress = "{F4}";
                    break;
                case "F5": keyToPress = "{F5}";
                    break;
                case "F6": keyToPress = "{F6}";
                    break;
                case "F7": keyToPress = "{F7}";
                    break;
                case "F8": keyToPress = "{F8}";
                    break;
                case "F9": keyToPress = "{F9}";
                    break;
                case "F10": keyToPress = "{F10}";
                    break;
                case "F11": keyToPress = "{F11}";
                    break;
                case "F12": keyToPress = "{F12}";
                    break;
            }
            if (keyToPress != "" && client.cursorModule.localCursorIsRemote)
            {
                client.socketModule.SendUDPMessage("KEYPRESS" + keyToPress);
            }
		}

        //! MyKeyDown
        /*!
         *  This method is called as a reaction to the event when a character key is pressed on the local machine.
         *  Several keys need to be escaped, as they have other meanings to the SendKeys method.  Their format is changed
         *  using the switch statement.  If the local cursor is currently being remoted, then the appropriate message 
         *  is sent to reproduce the keypress.
         */
		public void MyKeyPress(object sender, KeyPressEventArgs e)
		{
            String keyToPress = e.KeyChar.ToString();
            switch (keyToPress)
            {
                case "+": keyToPress = "{+}";
                    break;
                case "^": keyToPress = "{^}";
                    break;
                case "~": keyToPress = "{~}";
                    break;
                case "%": keyToPress = "{%}";
                    break;
                case "{": keyToPress = "{{}";
                    break;
                case "}": keyToPress = "{}}";
                    break;
                case "(": keyToPress = "{(}";
                    break;
                case ")": keyToPress = "{)}";
                    break;
                case "[": keyToPress = "{[}";
                    break;
                case "]": keyToPress = "{]}";
                    break;
            }

            if (client.cursorModule.localCursorIsRemote)
            {
                client.socketModule.SendUDPMessage("KEYPRESS" + keyToPress);
            }
		}

        //! SendCoordinates
        /*!
         *  This method is used to call the SocketModule method that sends coordinates to a remote machine.
         */
		private static void SendCoordinates(int x, int y)
		{
			client.socketModule.SendRelativeCoordinates(x,y);
		}

        //! ActivateClickGuard
        /*!
         *  This method is used to activate the ClickGuard that prevents local clicks from being registered.
         */
        public void ActivateClickGuard()
        {
            clickGuard.Show();
            clickGuard.BringToFront();
            clickGuard.Activate();
            Cursor.Hide();
            //clickGuard.ActivateClickGuard();
        }

        //! DeactivateClickGuard
        /*!
         *  This method is used to deactivate the ClickGuard that prevents local clicks from being registered.
         */
        public void DeactivateClickGuard()
        {
            clickGuard.Hide();
            this.Activate();
            Thread.Sleep(1000);
            Cursor.Show();
        }

        //! ProcessClientUsernameList
        /*!
         *  This method receives a new, unformatted list of semicolon-separated client usernames.  It extracts 
         *  the usernames and then updates the client.clientUsernameList String array with the new usernames.
         */
        public void ProcessClientUsernameList(String list)
        {
            String[] users = new String[10];
            char[] listAsCharArray = list.ToCharArray();
            int wordCount = 0;
            char[] tempWord = null;
            client.localIndex = Convert.ToInt16(list.Substring(0,1));
            int i = 1;
            while (i < listAsCharArray.Length)
            {
                tempWord = new char[100];
                int j = 0;
                while (listAsCharArray[i] != ';')
                {
                    tempWord[j] = listAsCharArray[i];
                    i++; j++;
                }
                String wordToAdd = new String(tempWord, 0, j);
                if (wordToAdd != "")
                {
                    users[wordCount] = wordToAdd;
                }
                else
                {
                    users[wordCount] = "[EMPTY]";
                }

                wordCount++;

                i++;
            }

            client.clientUsernameList = new String[wordCount];

            for (int k = 0; k < wordCount; k++)
            {
                client.clientUsernameList[k] = users[k];
            }

           
        }

        //! WndProc
        /*!
         *  This method is executed every time a new OS message is received.  It is overridden and used in this
         *  class to detect when the hotkey is pressed.
         */
        protected override void WndProc(ref Message msg)
        {
            //an OS message has arrived
            switch (msg.Msg)
            {
                case WM_HOTKEY:
                    //the hotkey has been pressed
                    if (client.cursorModule.localCursorIsRemote)
                    {
                        client.RemoveLocalCursorFromRemote(Point.Empty);
                        Status.ForeColor = Color.Black;
                        Status.Text = "Status: waiting to send local mouse/keyboard input to remote machine.";
                        Status.Refresh();
                    }
                    break;
            }
            base.WndProc(ref msg);
        }

        //! DisconnectedFromServer
        /*!
         *  This method is called when the SocketModule detects that the client has been disconnected 
         *  from the server.  The hotkey is released, an error message is displayed, and other cleanup
         *  actions are taken.
         */
        public void DisconnectedFromServer()
        {
            bool bcheck = UnregisterHotKey(Handle, HOTKEY_ID);
            CleanUpToReleaseLocalCursor();
            Status.ForeColor = Color.Red;
            Status.Text = "Status: connection dropped.  Please exit program and restart.";
            Status.Refresh();
            clientUsernameList.Enabled = false;
            SendButton.Enabled = false;
            Cursor.Show();
            Cursor.Show();
            actHook.Stop();
        }

        //! ReleaseLocalCursor
        /*!
         *  This method is called to return the local cursor from a remote machine to the local machine.
         */
        public void ReleaseLocalCursor()
        {
            client.RemoveLocalCursorFromRemote(Point.Empty);
        }

        //! PrepareToSendLocalCursor
        /*!
         *  This method is called to to prepare the local cursor to be remoted.  The appropriate status message is
         *  dispalyed, the click guard is activated, and the (now invisible) local mouse cursor is relocated to
         *  the position (50, 50).
         */
        public void PrepareToSendLocalCursor()
        {

            Status.ForeColor = Color.Black;
            Status.Text = "Status: mouse/keyboard input is currently being redirected.  To send mouse/keyboard back to local machine, press " + keyModifierAsString + " + " + hotkey;
            Status.Refresh();
            ActivateClickGuard();
            prevX = Screen.PrimaryScreen.Bounds.X + 50;
            prevY = Screen.PrimaryScreen.Bounds.Y + 50;
            SetCursorPos(prevX, prevY);
        }

        //! CleanUpToReleaseLocalCursor
        /*!
         *  This method is called to finish the process of returning the local cursor from a remote machine
         *  back to the local machine.
         */
        public void CleanUpToReleaseLocalCursor()
        {
            Status.Text = "Status: waiting to send local mouse/keyboard input to remote machine.";
            Status.Refresh();
            DeactivateClickGuard();
        }

        private void aboutCuRTToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutCuRT aboutBox = new AboutCuRT(this);
            aboutBox.Activate();
            aboutBox.Show();
        }

    }
}
