//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//                      Cursor Redirection Toolkit (CuRT)                       //
//               Written by Andy McGovern, Orchid Research Group                //
//                           andymcgovern@gmail.com                             //
//                          http://orchid.cs.uiuc.edu                           //
//  Department of Computer Science, University of Illinois at Urbana-Champaign  //
//                                July 4, 2006                                  //
//      Copyright 2006, The Board of Trustees of the University of Illinois     //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;


namespace CuRTServer
{
    //! ClientCommunicator
    /*!
     *  A new CuRTServer ClientCommunicator object is created each time that a new client connects to the CuRTServer.
     *  This ClientCommunicator object is responsible for all interactions with the client machine.
    */
    public class ClientCommunicator
	{

        //!necessary network objects
		private TcpListener userThreadTCPListener;
        private System.IO.StreamReader streamReader;
		private Socket userThreadSocket = null;
		private Stream networkStream;
        private UdpClient UDPClientSock;
        private IPEndPoint UDPSender;

        //!the private port where the ClientCommunicator will connect
		private int port;

        //!the index number (0-9) of this particular client machine
		private int index;

        //!the index number of the machine where this client's cursor is currently located
        //!(this value is -1 if this client's cursor is currently located on its own machine)
		protected internal int cursorLocationIndex;

        //!this is a copy of the instantiated CuRTServer object
        private CuRTServer server;

        //this is the local IP address of the server's machine
		private String IP;

        //this is name of the machine that is given by the user during the initialization process
        protected internal String name;


        public IPEndPoint Sender
        {
            get
            {
                return UDPSender;
            }
        }


        protected internal ClientCommunicator(int index, String IP, int port, CuRTServer server)
		{
			this.index = index;
			this.port = port;
			this.server = server;
			this.IP = IP;

            //a value of -1 indicates that the client's cursor is currently on its own machine
            cursorLocationIndex = -1;

		}

        //!  This method initializes the ClientThread object
        /*!
         * This method initialies both the TCP and the UDP connections on the new, private port.
         * It also calls a method in the server to broadcast the updated list of clients.  Then,
         * it starts the UDP message listener.
         */
		[STAThread]
        protected internal void Initialize()
		{
			try
			{

                IPAddress localIPAddress = IPAddress.Parse(IP);
                userThreadTCPListener = new TcpListener(localIPAddress, this.port);
				userThreadTCPListener.Start();
				userThreadSocket = userThreadTCPListener.AcceptSocket();
				networkStream = new NetworkStream(userThreadSocket);
				System.IO.StreamWriter streamWriter = new System.IO.StreamWriter(networkStream);
                streamReader = new System.IO.StreamReader(networkStream);

                //the user-defined name of this client machine is recevied via the TCP connection
                name = streamReader.ReadLine();

                Console.WriteLine("New client connected at index " + index + " with user name " + name + ".");

				byte[] data = new byte[1024];
				IPEndPoint ipep = new IPEndPoint(IPAddress.Any, this.port+10);
				UDPClientSock = new UdpClient(ipep);

				UDPSender = new IPEndPoint(IPAddress.Any, 0);

				data = UDPClientSock.Receive(ref UDPSender);

                Thread.Sleep(2000);
                server.BroadcastUserList();

                ListenForUDPMessages();
			}

			catch (Exception e) 
			{
				Console.WriteLine("Error..... " + e.StackTrace);
			}	



		}


        protected internal void Dispose()
        {
            if (userThreadTCPListener != null)
            {
                userThreadTCPListener.Stop();
                userThreadTCPListener = null;
            }

            if (streamReader != null)
            {
                streamReader.Close();
                streamReader.Dispose();
                streamReader = null;
            }

            if (networkStream != null)
            {
                networkStream.Close();
                networkStream.Dispose();
                networkStream = null;
            }

            if (userThreadSocket != null)
            {
                userThreadSocket.Disconnect(true);
                userThreadSocket.Close();
                userThreadSocket = null;
            }

            System.Windows.Forms.Application.ExitThread();
        }



        //!  This method listens for incoming UDP messages, calling ProcessMessage whenever a new message arrives.
		private void ListenForUDPMessages()
		{	byte[] data = new byte[1024];
			String message = null;
            while (true)
            {
                try
                {
                    data = UDPClientSock.Receive(ref UDPSender);
                    message = Encoding.ASCII.GetString(data, 0, data.Length);

                    processMessage(message);
                }
                catch (Exception e)
                {
                    //Console.WriteLine("Error..... " + e.StackTrace + ", message was: " + message);
                }

            }

		}

        //!  This method processes incoming messages as they arrive, using a predetermined messaging protocol.
		private void processMessage(String message)
		{
			try
			{
                if (message.Length < 2)
                    return;


                //message arrived to move a virtual cursor with relative coordinates
                else if (message.Substring(0, 2) == "MR")
                {
                    int x = Convert.ToInt16(message.Substring(2, 4));
                    int y = Convert.ToInt16(message.Substring(6, 4));
                    server.BroadcastRelativeMove(index, x, y);
                }


                //message arrived that a new cursor is being sent to the local machine
                else if (message.Substring(0, 2) == "SC")
                {
                    int cursorIndex = Convert.ToInt16(message.Substring(2, 1));
                    int machineIndex = Convert.ToInt16(message.Substring(3, 1));
                    int startX = Convert.ToInt16(message.Substring(4, 4));
                    int startY = Convert.ToInt16(message.Substring(8, 4));
                    server.BroadcastSendCursor(cursorIndex, machineIndex, startX, startY);
                }

                //message arrived to remove a virtual cursor from a machine
                else if (message.Substring(0, 2) == "RC")
                {
                    int cursorToRemoveIndex = index;
                    int machineIndex = this.cursorLocationIndex;
                    this.cursorLocationIndex = -1;
                    server.BroadcastRemoveCursor(machineIndex, cursorToRemoveIndex);
                }

                //message arrived to block a cursor on a machine
                else if (message.Substring(0, 3) == "BRC")
                {
                    server.BlockRemoteCursor(this.index);
                }

                //message arrived to unblock a cursor on a machine
                else if (message.Substring(0, 3) == "URC")
                {
                    server.UnblockRemoteCursor(this.index);
                }

                else if (message.Length < 4)
                    return;

                else if (message.Substring(0, 4) == "EXIT")
                {
                    server.DisconnectClient(index);
                }

                else if (message.Length < 8)
                    return;


                    //message arrived for a keypress
                else if (message.Substring(0, 8) == "KEYPRESS")
                {
                    String key = message.Substring(8);
                    server.BroadcastKeyPress(index, key);
                }


                //message arrived to perform a left click down
                else if (message.Substring(0, 8) == "LFTCLKDN")
                {
                    server.BroadcastLeftClickDown(this.index);
                }

                //message arrived to perform a left click up
                else if (message.Substring(0, 8) == "LFTCLKUP")
                {
                    server.BroadcastLeftClickUp(this.index);
                }

                //message arrived to perform a right click down
                else if (message.Substring(0, 8) == "RGTCLKDN")
                {
                    server.BroadcastRightClickDown(this.index);
                }

                //message arrived to perform a right click up
                else if (message.Substring(0, 8) == "RGTCLKUP")
                {
                    server.BroadcastRightClickUp(this.index);
                }

			}
			catch (Exception e) 
			{
				Console.WriteLine("Error..... " + e.StackTrace + ", message was "+ message);
			}	
		}

        //!  This method sends a message over the UDP protocol to the corresponding client.
		public void sendUDPMessage(String message)
		{
			byte[] data = new byte[1024];
			data = Encoding.ASCII.GetBytes(message);
			UDPClientSock.Send(data, data.Length, UDPSender);
		}
	}
}
