//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//                      Cursor Redirection Toolkit (CuRT)                       //
//               Written by Andy McGovern, Orchid Research Group                //
//                           andymcgovern@gmail.com                             //
//                          http://orchid.cs.uiuc.edu                           //
//  Department of Computer Science, University of Illinois at Urbana-Champaign  //
//                                July 4, 2006                                  //
//      Copyright 2006, The Board of Trustees of the University of Illinois     //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

using System;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace CuRTServer
{
    //! CuRTServer
    /*!
     *  The CuRTServer runs in a console window as a standalone application.  It must be run first before any
     *  CuRTClients can be run.
     */
	public class CuRTServer
	{
        //!server's IP address
		private static IPAddress serverIP;

        //!server's TCP listener
		private static TcpListener serverTCPListener;

        //!array of ClientThread objects representing clients currently connected to the system
		internal protected ClientCommunicator[] clients;

        //!the next available client ID number (0-9)
        private int nextID;

        //!the next available IP port number (2000-2009)
        private int nextPort;

        //!the local (server) IP address
		private String IP;

        private bool logging = true;
        private Logging.Data data;

        //!  CuRTServer constructor
        /*! 
         *  This constructor displays the welcome screen, starts the TCP listener, and then begins to listen
         *  for incoming connections.
         */
		public CuRTServer()
        {
            Console.WriteLine("Welcome to CuRT, the Cursor Redirection Toolkit");
            Console.WriteLine("Written by Andy McGovern, Orchid Research Group");
            Console.WriteLine("andymcgovern@gmail.com");
            Console.WriteLine("http://orchid.cs.uiuc.edu");
            Console.WriteLine("Department of Computer Science, University of Illinois at Urbana-Champaign");
            Console.WriteLine("July 4, 2006");
            Console.WriteLine("Copyright 2006, The Board of Trustees of the University of Illinois.");

            if(logging){
                data = new Logging.Data();
                Console.WriteLine("Starting logging support...");

            }

            Console.Write("Checking for network connection...");

            clients = new ClientCommunicator[10];
			nextID = 100;
			nextPort = 2000;
            IP = GetLocalIP();
			try 
			{
                Console.WriteLine("");
                Console.WriteLine("");
                Console.WriteLine("Waiting for clients to connect to "+ IP+"...");

                //all users must first connect to server at port 4415, and then they reconnect on their own private port
                serverIP = IPAddress.Parse(IP);
                serverTCPListener = new TcpListener(serverIP, 4415);
                serverTCPListener.Start();

                ListenForConnections();

			}
			catch (Exception e) 
			{
				Console.WriteLine("Error..... " + e.StackTrace);
			}	
		}

        //!  Returns a String with the local machine's IP address
        private String GetLocalIP()
        {
            String strHostName = Dns.GetHostName();
            IPHostEntry ipEntry = Dns.GetHostEntry(strHostName);
            IPAddress[] addr = ipEntry.AddressList;

            for (int i = 0; i < addr.Length; i++)
            {
                if (addr[i].AddressFamily == AddressFamily.InterNetwork)
                    return addr[i].ToString();
            }

            return null;

        }

        //!  Listens for incoming connections from clients trying to connect
        /*! 
         *  This method listens for new connections in an infinite loop.  As each new connection is made, a UserThread
         *  object is created, and a new thread is created for that object.  That UserThread object begins to perform 
         *  all communications directly with the client.
         */
		private void ListenForConnections()
		{

			Socket socketForClient = null;
			ClientCommunicator tempUserThread = null;

			while (true) 
			{ 
				if (serverTCPListener == null)
                {
                    Console.WriteLine("Error: TCPListener did not initialize correctly.");
					return;
                }

				try 
				{   
					socketForClient = serverTCPListener.AcceptSocket();

					NetworkStream networkStream = new NetworkStream(socketForClient);
					System.IO.StreamWriter streamWriter = new System.IO.StreamWriter(networkStream);

                    //the for loop finds the next available index to place the incoming user
                    for (int nextIndex=0; nextIndex<clients.Length; nextIndex++)
						if (clients[nextIndex]==null)
						//new UserThread object is created for the user to be added
                        {	
							tempUserThread = new ClientCommunicator(nextIndex, IP, this.nextPort, this);
							clients[nextIndex] = tempUserThread;
							nextIndex=clients.Length;
                        }

                    //the client's private port is sent as a message so that it can immediately reconnect
					streamWriter.WriteLine(this.nextPort);
					streamWriter.Flush();
					socketForClient.Close();

                    //new user is initialized in its own new thread
					Thread NewUserThread = new Thread(new ThreadStart(tempUserThread.Initialize));
					NewUserThread.Name = "NewUserThread";
					NewUserThread.Start();

                    //nextID and nextPort are incremented for the next incoming user
					this.nextID++;
					this.nextPort++;

				}
				catch (Exception e) 
				{
					Console.WriteLine("Error..... " + e.StackTrace);
				}	
			}
		}

        //!  Broadcasts list of clients to all currently connected clients
        protected internal void BroadcastUserList()
        {
            String nameList = "";

            for (int i = 0; i < 10; i++)
            {
                if (clients[i] != null)
                {
                    nameList = nameList + clients[i].name;
                }
                nameList = nameList + ";";
            }

            for (int j = 0; j < 10; j++)
            {
                if (clients[j] != null)
                    clients[j].sendUDPMessage("UL" + j + nameList);
            }
        }

        //!  Broadcasts list of clients to all currently connected clients
        /*!
         *  This method generates strings representing the X and Y coordinates and then
         *  calls the appropriate ClientThread method to send the message.
         * 
         */
        protected internal void BroadcastRelativeMove(int index, int x, int y)
		{

			int remoteIndex = clients[index].cursorLocationIndex;

			if (remoteIndex != -1)

			{
				String xAsString="", yAsString="";

                if (x < 0)
                {
                    xAsString = "-";
                    x = x * -1;
                }
                else
                    xAsString = "0";

                if (x < 10)
                    xAsString = xAsString + "00" + x;
                else if (x < 100)
                    xAsString = xAsString + "0" + x;
                else
                    xAsString = xAsString + x;

                if (y < 0)
                {
                    yAsString = "-";
                    y = y * -1;
                }
                else
                    yAsString = "0";

                if (y < 10)
                    yAsString = yAsString + "00" + y;
                else if (y < 100)
                    yAsString = yAsString + "0" + y;
                else
                    yAsString = yAsString + y;

				clients[remoteIndex].sendUDPMessage("MR"+index+xAsString+yAsString);

			}

		}

        //!  BroadcastSendCursor
        /*!
         *   This method is called when the local machine intends to send a virtual cursor (currently
         *   remoted to the local machine) to another machine, possibly its own "home" machine.  The
         *   x and y coordinates are reformatted as four-digit Strings, with leading zeros, if necessary.
         */
        internal protected void BroadcastSendCursor(int cursorIndex, int machineIndex, int x, int y)
		{
            String xAsString, yAsString;
            
            //if cursorIndex and machineIndex are the same, then the purpose of the message was to send
            //a cursor back to its own machine
            if (cursorIndex != machineIndex)
                clients[cursorIndex].cursorLocationIndex = machineIndex;

            if (x < 0)
                xAsString = "0";
            else if (x < 10)
                xAsString = "000" + x;
            else if (x < 100)
                xAsString = "00" + x;
            else if (x < 1000)
                xAsString = "0" + x;
            else
                xAsString = "" + x;

            if (y < 0)
                yAsString = "0";
            else if (y < 10)
                yAsString = "000" + y;
            else if (y < 100)
                yAsString = "00" + y;
            else if (y < 1000)
                yAsString = "0" + y;
            else
                yAsString = "" + y;

            clients[machineIndex].sendUDPMessage("SC" + cursorIndex + xAsString + yAsString);
            BroadcastUserList();

            if(logging)
                data.InsertIntoShadowEventRedirectionsTable(DateTime.Now, Dns.GetHostName(), Dns.GetHostEntry(clients[machineIndex].Sender.Address.ToString()).HostName);
            

		}

        //!  This method removes a virtual cursor from a machine
        /*!
         * This method acts as a switchboard between the machine that wants its cursor back and the machine where
         * the cursor is present virtually.  Once the second machine receives the message, the virtual cursor is removed.
         */
        internal protected void BroadcastRemoveCursor(int machineIndex, int cursorToRemoveIndex)
        {
            clients[machineIndex].sendUDPMessage("RC" + cursorToRemoveIndex);

            if (logging)
                data.UpdateShadowEventRedirectionsTable(DateTime.Now);
        }

        //!  This method broadcasts a left click down to a machine
        /*!
         * This method acts as a switchboard between the machine that performed the click and the machine where the click
         * should take place.  Once the second machine receives the message, the click occurs.
         */
        internal protected void BroadcastLeftClickDown(int index)
        {
            int remoteIndex = clients[index].cursorLocationIndex;
            if (remoteIndex != -1)
                clients[remoteIndex].sendUDPMessage("LFTCLKDN" + index);
        }

        //!  This method broadcasts a left click up to a machine
        /*!
         * This method acts as a switchboard between the machine that performed the click and the machine where the click
         * should take place.  Once the second machine receives the message, the click occurs.
         */
        internal protected void BroadcastLeftClickUp(int index)
		{				
            int remoteIndex = clients[index].cursorLocationIndex;
			if (remoteIndex != -1)
				clients[remoteIndex].sendUDPMessage("LFTCLKUP"+index);
		}

        //!  This method broadcasts a right click down to a machine
        /*!
         * This method acts as a switchboard between the machine that performed the click and the machine where the click
         * should take place.  Once the second machine receives the message, the click occurs.
         */
        internal protected void BroadcastRightClickDown(int index)
        {
            int remoteIndex = clients[index].cursorLocationIndex;
            clients[remoteIndex].sendUDPMessage("RGTCLKDN" + index);
        }

        //!  This method broadcasts a right click up to a machine
        /*!
         * This method acts as a switchboard between the machine that performed the click and the machine where the click
         * should take place.  Once the second machine receives the message, the click occurs.
         */
        internal protected void BroadcastRightClickUp(int index)
		{
			int remoteIndex = clients[index].cursorLocationIndex;
            clients[remoteIndex].sendUDPMessage("RGTCLKUP" + index);
		}

        //!  This method broadcasts a key press to a machine
        /*!
         * This method acts as a switchboard between the machine that performed the keypress and the machine where the keypress
         * should take place.  Once the second machine receives the message, the keypress occurs.
         */
        internal protected void BroadcastKeyPress(int index, String key)
		{
			int remoteIndex = clients[index].cursorLocationIndex;
            if (remoteIndex!=-1)
                clients[remoteIndex].sendUDPMessage("KEYPRESS" + index + key);
		}

        //!  to add
        /*!
         * 
         */
        internal protected void BroadcastMoveToCoord(String message)
        {

        }

        //!  This method blocks a remote cursor
        /*!
         * This method acts as a switchboard between the machine that needs to block the cursor and the machine where the
         * to-be-blocked cursor is currently located.  Once the second machine receives the message, the cursor is blocked.
         */
        internal protected void BlockRemoteCursor(int index)
        {
            int remoteIndex = clients[index].cursorLocationIndex;
            clients[remoteIndex].sendUDPMessage("BRC" + index);
        }

        //!  This method unblocks a remote cursor
        /*!
         * This method acts as a switchboard between the machine that needs to unblock the cursor and the machine where the
         * to-be-unblocked cursor is currently located.  Once the second machine receives the message, the cursor is blocked.
         */
        internal protected void UnblockRemoteCursor(int index)
        {
            int remoteIndex = clients[index].cursorLocationIndex;
            clients[remoteIndex].sendUDPMessage("URC" + index);
        }

        internal protected void DisconnectClient(int index)
        {
            clients[index].Dispose();
            clients[index] = null;

        }


        //!  This is the CuRTServer main method
        /*!
         * This method simply creates a CuRTServer object.  The CuRTServer constructor takes care of the rest.
         */
        internal protected static void Main(String[] arguments) 
		{
            CuRTServer server = new CuRTServer();
		}


	}
}
